// This file is generated by scripts/release.py
// Do not edit it manually.

// Include compatibility layer for cross-platform support
#include "compat.h"

#if defined(_WIN32) || defined(__APPLE__)
// Stub implementations - instrumentation not supported on Windows/macOS
#include <stdbool.h>
#include <stdint.h>

typedef struct InstrumentHooks {
  char reserved;
} InstrumentHooks;

InstrumentHooks* instrument_hooks_init() {
  static InstrumentHooks instance = {};
  return &instance;
}

void instrument_hooks_deinit(InstrumentHooks* hooks) {}

bool instrument_hooks_is_instrumented(InstrumentHooks* hooks) { return false; }

uint8_t instrument_hooks_start_benchmark(InstrumentHooks* hooks) { return 0; }

uint8_t instrument_hooks_stop_benchmark(InstrumentHooks* hooks) { return 0; }

uint8_t instrument_hooks_set_executed_benchmark(InstrumentHooks* hooks,
                                                uint32_t pid, const char* uri) {
  return 0;
}

// Deprecated: use instrument_hooks_set_executed_benchmark instead
uint8_t instrument_hooks_executed_benchmark(InstrumentHooks* hooks,
                                            uint32_t pid, const char* uri) {
  return 0;
}

uint8_t instrument_hooks_set_integration(InstrumentHooks* hooks,
                                         const char* name,
                                         const char* version) {
  return 0;
}

void instrument_hooks_set_feature(InstrumentHooks* hooks, uint64_t feature,
                                  bool enabled) {}

uint64_t instrument_hooks_current_timestamp() { return 0; }

uint8_t instrument_hooks_add_marker(InstrumentHooks* hooks, uint32_t pid,
                                    uint8_t marker_type, uint64_t timestamp) {
  return 0;
}

#else
#define ZIG_TARGET_MAX_INT_ALIGNMENT 16
#include "zig.h"
struct anon__lazy_57 {
 uint8_t const *ptr;
 uintptr_t len;
};
struct instruments_valgrind_ValgrindInstrument__554;
typedef struct anon__lazy_83 nav__739_39;
struct mem_Allocator__589;
struct mem_Allocator_VTable__592;
struct mem_Allocator__589 {
 void *ptr;
 struct mem_Allocator_VTable__592 const *vtable;
};
struct instruments_valgrind_ValgrindInstrument__554 {
 struct mem_Allocator__589 allocator;
};
struct anon__lazy_83 {
 struct instruments_valgrind_ValgrindInstrument__554 payload;
 uint16_t error;
};
struct fs_File__636;
typedef struct anon__lazy_94 nav__1115_39;
typedef struct anon__lazy_57 nav__1115_41;
struct fs_File__636 {
 int32_t handle;
};
struct anon__lazy_94 {
 struct fs_File__636 payload;
 uint16_t error;
};
struct fs_File_OpenFlags__1954;
struct fs_File_OpenFlags__1954 {
 uint8_t mode;
 uint8_t lock;
 bool lock_nonblocking;
 bool allow_ctty;
};
struct fifo_UnixPipe_Writer__628;
typedef struct anon__lazy_102 nav__1117_39;
typedef struct anon__lazy_57 nav__1117_42;
struct array_list_ArrayListAligned_28u8_2cnull_29__648;
typedef struct anon__lazy_105 nav__1117_50;
struct anon__lazy_105 {
 uint8_t *ptr;
 uintptr_t len;
};
struct array_list_ArrayListAligned_28u8_2cnull_29__648 {
 struct anon__lazy_105 items;
 uintptr_t capacity;
 struct mem_Allocator__589 allocator;
};
struct fifo_UnixPipe_Writer__628 {
 struct mem_Allocator__589 allocator;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 buffer;
 struct fs_File__636 file;
};
struct anon__lazy_102 {
 struct fifo_UnixPipe_Writer__628 payload;
 uint16_t error;
};
typedef struct anon__lazy_94 nav__1117_59;
struct fifo_UnixPipe_Reader__630;
typedef struct anon__lazy_112 nav__1116_39;
typedef struct anon__lazy_57 nav__1116_42;
typedef struct anon__lazy_105 nav__1116_50;
struct fifo_UnixPipe_Reader__630 {
 struct mem_Allocator__589 allocator;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 buffer;
 struct fs_File__636 file;
};
struct anon__lazy_112 {
 struct fifo_UnixPipe_Reader__630 payload;
 uint16_t error;
};
typedef struct anon__lazy_94 nav__1116_59;
struct runner_fifo_RunnerFifo__621;
typedef struct anon__lazy_118 nav__1082_39;
typedef struct anon__lazy_105 nav__1082_49;
struct runner_fifo_RunnerFifo__621 {
 struct mem_Allocator__589 allocator;
 struct fifo_UnixPipe_Writer__628 writer;
 struct fifo_UnixPipe_Reader__630 reader;
};
struct anon__lazy_118 {
 struct runner_fifo_RunnerFifo__621 payload;
 uint16_t error;
};
typedef struct anon__lazy_57 nav__1082_64;
typedef struct anon__lazy_102 nav__1082_66;
typedef struct anon__lazy_112 nav__1082_71;
struct shared_Command__2038;
struct shared_Command__struct_2041__2041;
typedef struct anon__lazy_57 nav__1121_44;
struct shared_Command__struct_2041__2041 {
 struct anon__lazy_57 uri;
 uint32_t pid;
};
struct shared_Command__struct_2042__2042;
struct shared_Command__struct_2042__2042 {
 struct anon__lazy_57 name;
 struct anon__lazy_57 version;
};
struct shared_Command__struct_2043__2043;
struct shared_MarkerType__2045;
struct shared_MarkerType__2045 {
 union {
  uint64_t SampleStart;
  uint64_t SampleEnd;
  uint64_t BenchmarkStart;
  uint64_t BenchmarkEnd;
 } payload;
 uint8_t tag;
};
struct shared_Command__struct_2043__2043 {
 struct shared_MarkerType__2045 marker;
 uint32_t pid;
};
struct shared_Command__2038 {
 union {
  struct shared_Command__struct_2041__2041 ExecutedBenchmark;
  struct shared_Command__struct_2042__2042 SetIntegration;
  struct shared_Command__struct_2043__2043 AddMarker;
  uint64_t SetVersion;
  uint8_t IntegrationModeResponse;
 } payload;
 uint8_t tag;
};
typedef struct anon__lazy_105 nav__1121_65;
struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080;
struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *context;
};
typedef struct anon__lazy_149 nav__4211_38;
struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263;
struct anon__lazy_149 {
 uint32_t payload;
 uint16_t error;
};
struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239;
struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *context;
};
struct io_Reader__2409;
typedef struct anon__lazy_161 nav__4211_50;
typedef struct anon__lazy_105 nav__4211_52;
struct io_Reader__2409 {
 void const *context;
 struct anon__lazy_161 (*readFn)(void const *, struct anon__lazy_105);
};
struct anon__lazy_161 {
 uintptr_t payload;
 uint16_t error;
};
typedef struct anon__lazy_171 nav__4211_64;
struct anon__lazy_171 {
 uint16_t error;
 uint8_t payload[4];
};
typedef struct anon__lazy_149 nav__4210_38;
typedef struct anon__lazy_105 nav__4253_39;
typedef struct anon__lazy_175 nav__4253_40;
struct anon__lazy_175 {
 struct anon__lazy_105 payload;
 uint16_t error;
};
typedef struct anon__lazy_161 nav__4253_59;
typedef struct anon__lazy_180 nav__4253_69;
struct anon__lazy_180 {
 uint64_t payload;
 uint16_t error;
};
typedef struct anon__lazy_182 nav__4253_72;
struct anon__lazy_182 {
 uint16_t error;
 uint8_t payload[8];
};
typedef struct anon__lazy_57 nav__4252_39;
typedef struct anon__lazy_185 nav__4252_40;
struct anon__lazy_185 {
 struct anon__lazy_57 payload;
 uint16_t error;
};
typedef struct anon__lazy_105 nav__4252_55;
typedef struct anon__lazy_175 nav__4252_57;
typedef struct anon__lazy_189 nav__4251_39;
typedef struct anon__lazy_57 nav__4251_44;
struct anon__lazy_189 {
 struct shared_Command__struct_2041__2041 payload;
 uint16_t error;
};
typedef struct anon__lazy_149 nav__4251_57;
typedef struct anon__lazy_185 nav__4251_60;
typedef struct anon__lazy_189 nav__4250_39;
typedef struct anon__lazy_57 nav__4250_44;
typedef struct anon__lazy_197 nav__4259_39;
typedef struct anon__lazy_57 nav__4259_44;
struct anon__lazy_197 {
 struct shared_Command__struct_2042__2042 payload;
 uint16_t error;
};
typedef struct anon__lazy_185 nav__4259_57;
typedef struct anon__lazy_197 nav__4258_39;
typedef struct anon__lazy_57 nav__4258_44;
typedef struct anon__lazy_180 nav__4265_38;
typedef struct anon__lazy_161 nav__4265_50;
typedef struct anon__lazy_105 nav__4265_52;
typedef struct anon__lazy_182 nav__4265_64;
typedef struct anon__lazy_180 nav__4264_38;
typedef struct anon__lazy_205 nav__4263_39;
struct anon__lazy_205 {
 struct shared_MarkerType__2045 payload;
 uint16_t error;
};
typedef struct anon__lazy_149 nav__4263_54;
typedef struct anon__lazy_180 nav__4263_56;
typedef struct anon__lazy_205 nav__4262_39;
typedef struct anon__lazy_209 nav__4261_39;
struct anon__lazy_209 {
 struct shared_Command__struct_2043__2043 payload;
 uint16_t error;
};
typedef struct anon__lazy_149 nav__4261_57;
typedef struct anon__lazy_205 nav__4261_60;
typedef struct anon__lazy_209 nav__4260_39;
typedef struct anon__lazy_214 nav__4267_38;
struct anon__lazy_214 {
 uint16_t error;
 uint8_t payload;
};
typedef struct anon__lazy_149 nav__4267_46;
typedef struct anon__lazy_214 nav__4266_38;
typedef struct anon__lazy_220 nav__4209_39;
typedef struct anon__lazy_57 nav__4209_45;
struct anon__lazy_220 {
 struct shared_Command__2038 payload;
 uint16_t error;
};
typedef struct anon__lazy_149 nav__4209_66;
typedef struct anon__lazy_189 nav__4209_68;
typedef struct anon__lazy_197 nav__4209_70;
typedef struct anon__lazy_209 nav__4209_72;
typedef struct anon__lazy_180 nav__4209_74;
typedef struct anon__lazy_214 nav__4209_76;
typedef struct anon__lazy_220 nav__4208_39;
typedef struct anon__lazy_57 nav__4208_45;
typedef struct anon__lazy_220 nav__1126_39;
typedef struct anon__lazy_57 nav__1126_45;
typedef struct anon__lazy_105 nav__1126_68;
typedef struct anon__lazy_161 nav__1126_76;
struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 {
 struct anon__lazy_105 buffer;
 uintptr_t pos;
};
typedef struct anon__lazy_220 nav__1127_39;
typedef struct anon__lazy_229 nav__1127_42;
typedef struct anon__lazy_57 nav__1127_46;
struct anon__lazy_229 {
 uint64_t payload;
 bool is_null;
};
typedef struct anon__lazy_229 nav__1128_40;
typedef struct anon__lazy_57 nav__1128_48;
typedef struct anon__lazy_220 nav__1128_60;
typedef struct anon__lazy_105 nav__1128_70;
typedef struct anon__lazy_236 nav__1128_77;
struct anon__lazy_236 {
 struct shared_Command__2038 f0;
};
typedef struct anon__lazy_105 nav__1092_50;
typedef struct anon__lazy_57 nav__1092_63;
typedef struct anon__lazy_229 nav__1092_76;
typedef struct anon__lazy_57 nav__1083_43;
typedef struct anon__lazy_241 nav__1083_45;
struct anon__lazy_241 {
 struct anon__lazy_57 f0;
};
typedef struct anon__lazy_214 nav__1093_38;
typedef struct anon__lazy_105 nav__1093_53;
typedef struct anon__lazy_57 nav__1093_66;
typedef struct anon__lazy_229 nav__1093_79;
typedef struct anon__lazy_220 nav__1093_81;
struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581;
typedef struct anon__lazy_246 nav__773_39;
typedef struct anon__lazy_105 nav__773_50;
struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 {
 struct runner_fifo_RunnerFifo__621 fifo;
};
struct anon__lazy_246 {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 payload;
 uint16_t error;
};
typedef struct anon__lazy_118 nav__773_62;
typedef struct anon__lazy_214 nav__773_65;
struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572;
typedef struct anon__lazy_252 nav__760_39;
typedef struct anon__lazy_105 nav__760_50;
struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 {
 struct runner_fifo_RunnerFifo__621 fifo;
};
struct anon__lazy_252 {
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 payload;
 uint16_t error;
};
typedef struct anon__lazy_118 nav__760_62;
typedef struct anon__lazy_214 nav__760_65;
struct instruments_root_InstrumentHooks__547;
typedef struct anon__lazy_258 nav__726_39;
typedef struct anon__lazy_105 nav__726_53;
struct instruments_root_InstrumentHooks__547 {
 union {
  struct instruments_valgrind_ValgrindInstrument__554 valgrind;
  struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 perf;
  struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 analysis;
 } payload;
 uint8_t tag;
};
struct anon__lazy_258 {
 struct instruments_root_InstrumentHooks__547 payload;
 uint16_t error;
};
typedef struct anon__lazy_83 nav__726_69;
typedef struct anon__lazy_246 nav__726_71;
typedef struct anon__lazy_252 nav__726_73;
typedef struct anon__lazy_266 nav__237_46;
struct anon__lazy_266 {
 struct instruments_root_InstrumentHooks__547 *payload;
 uint16_t error;
};
typedef struct anon__lazy_105 nav__237_51;
struct mem_Allocator_VTable__592 {
 uint8_t *(*alloc)(void *, uintptr_t, uint8_t, uintptr_t);
 bool (*resize)(void *, struct anon__lazy_105, uint8_t, uintptr_t, uintptr_t);
 uint8_t *(*remap)(void *, struct anon__lazy_105, uint8_t, uintptr_t, uintptr_t);
 void (*free)(void *, struct anon__lazy_105, uint8_t, uintptr_t);
};
typedef struct anon__lazy_258 nav__237_78;
typedef struct anon__lazy_266 nav__3717_40;
typedef struct anon__lazy_279 nav__3717_51;
struct anon__lazy_279 {
 uint8_t *payload;
 uint16_t error;
};
typedef struct anon__lazy_57 nav__1183_39;
struct fs_Dir__1956;
struct fs_Dir__1956 {
 int32_t fd;
};
typedef struct anon__lazy_94 nav__1180_39;
typedef struct anon__lazy_57 nav__1180_41;
typedef struct anon__lazy_105 nav__1118_48;
typedef struct anon__lazy_105 nav__1123_48;
typedef struct anon__lazy_105 nav__1397_43;
typedef struct anon__lazy_57 nav__4101_46;
typedef struct anon__lazy_57 nav__1274_40;
typedef struct anon__lazy_161 nav__1274_47;
struct os_linux_timespec__struct_3100__3100;
struct os_linux_timespec__struct_3100__3100 {
 intptr_t sec;
 intptr_t nsec;
};
typedef struct anon__lazy_310 nav__4115_41;
struct anon__lazy_310 {
 struct os_linux_timespec__struct_3100__3100 payload;
 uint16_t error;
};
typedef struct anon__lazy_161 nav__1264_38;
typedef struct anon__lazy_105 nav__1264_41;
typedef struct anon__lazy_105 nav__1394_43;
typedef struct anon__lazy_105 nav__4174_40;
typedef struct anon__lazy_161 nav__4207_38;
typedef struct anon__lazy_105 nav__4207_41;
typedef struct anon__lazy_171 nav__4249_39;
typedef struct anon__lazy_161 nav__4249_44;
typedef struct anon__lazy_105 nav__4249_46;
typedef struct anon__lazy_182 nav__4254_39;
typedef struct anon__lazy_161 nav__4254_44;
typedef struct anon__lazy_105 nav__4254_46;
typedef struct anon__lazy_105 nav__4255_39;
typedef struct anon__lazy_175 nav__4255_40;
typedef struct anon__lazy_279 nav__4255_51;
typedef struct anon__lazy_161 nav__4214_38;
typedef struct anon__lazy_105 nav__4214_41;
struct cimport_struct_timespec__3171;
struct cimport_struct_timespec__3171 {
 long tv_sec;
 long tv_nsec;
};
typedef struct anon__lazy_336 nav__3969_42;
struct anon__lazy_336 {
 long payload;
 bool is_null;
};
typedef struct anon__lazy_57 nav__3974_43;
typedef struct anon__lazy_236 nav__4277_39;
typedef struct anon__lazy_57 nav__4277_43;
typedef struct anon__lazy_57 nav__4280_39;
typedef struct anon__lazy_241 nav__4280_40;
typedef struct anon__lazy_105 nav__1084_50;
typedef struct anon__lazy_279 nav__4281_39;
typedef struct anon__lazy_105 nav__4281_52;
typedef struct anon__lazy_57 nav__4316_39;
typedef struct anon__lazy_57 nav__3918_40;
typedef struct anon__lazy_359 nav__3918_49;
struct anon__lazy_359 {
 uint16_t error;
 uint8_t payload[4096];
};
typedef struct anon__lazy_94 nav__3854_39;
typedef struct anon__lazy_57 nav__3854_42;
typedef struct anon__lazy_359 nav__3854_53;
typedef struct anon__lazy_105 nav__1362_42;
typedef struct anon__lazy_105 nav__1399_43;
typedef struct anon__lazy_57 nav__5311_46;
typedef struct anon__lazy_161 nav__1273_38;
typedef struct anon__lazy_57 nav__1273_41;
typedef struct anon__lazy_310 nav__1773_39;
typedef struct anon__lazy_161 nav__1263_38;
typedef struct anon__lazy_105 nav__1263_41;
typedef struct anon__lazy_161 nav__4166_38;
typedef struct anon__lazy_105 nav__4166_42;
typedef struct anon__lazy_105 nav__4216_40;
typedef struct anon__lazy_161 nav__4216_43;
typedef struct anon__lazy_279 nav__5312_39;
typedef struct anon__lazy_161 nav__5312_49;
typedef struct anon__lazy_161 nav__4215_38;
typedef struct anon__lazy_105 nav__4215_41;
typedef struct anon__lazy_336 nav__6247_38;
typedef struct anon__lazy_57 nav__7028_40;
typedef struct anon__lazy_105 nav__7028_51;
typedef struct anon__lazy_236 nav__7029_39;
typedef struct anon__lazy_57 nav__7029_43;
typedef struct anon__lazy_105 nav__7029_59;
typedef struct anon__lazy_175 nav__7029_62;
typedef struct anon__lazy_105 nav__7030_41;
typedef struct anon__lazy_175 nav__7030_44;
typedef struct anon__lazy_105 nav__7031_41;
typedef struct anon__lazy_175 nav__7031_44;
typedef struct anon__lazy_57 nav__7033_39;
typedef struct anon__lazy_241 nav__7033_40;
typedef struct anon__lazy_105 nav__7033_47;
typedef struct anon__lazy_175 nav__7033_50;
typedef struct anon__lazy_105 nav__1122_49;
typedef struct anon__lazy_105 nav__1129_50;
typedef struct anon__lazy_161 nav__1129_58;
typedef struct anon__lazy_57 nav__4322_39;
typedef struct anon__lazy_359 nav__1857_39;
typedef struct anon__lazy_57 nav__1857_41;
typedef struct anon__lazy_105 nav__1857_47;
typedef struct anon__lazy_94 nav__3855_39;
typedef struct anon__lazy_398 nav__3855_50;
struct anon__lazy_398 {
 int32_t payload;
 uint16_t error;
};
typedef struct anon__lazy_105 nav__1400_43;
typedef struct anon__lazy_175 nav__1400_55;
typedef struct anon__lazy_57 nav__7106_45;
typedef struct anon__lazy_57 nav__7108_45;
typedef struct anon__lazy_161 nav__1592_38;
typedef struct anon__lazy_57 nav__1592_40;
typedef struct anon__lazy_161 nav__1584_38;
typedef struct anon__lazy_105 nav__1584_40;
typedef struct anon__lazy_161 nav__7179_38;
typedef struct anon__lazy_426 nav__7179_42;
struct anon__lazy_426 {
 uintptr_t f0;
 uint8_t f1;
};
typedef struct anon__lazy_279 nav__7180_39;
typedef struct anon__lazy_105 nav__7180_52;
typedef struct anon__lazy_161 nav__4213_38;
typedef struct anon__lazy_105 nav__4213_41;
typedef struct anon__lazy_57 nav__7182_39;
struct io_Writer__4171;
typedef struct anon__lazy_236 nav__7476_40;
typedef struct anon__lazy_161 nav__7476_43;
typedef struct anon__lazy_57 nav__7476_45;
struct io_Writer__4171 {
 void const *context;
 struct anon__lazy_161 (*writeFn)(void const *, struct anon__lazy_57);
};
struct fmt_FormatOptions__4577;
typedef struct anon__lazy_441 nav__7476_67;
struct anon__lazy_441 {
 uintptr_t payload;
 bool is_null;
};
struct fmt_FormatOptions__4577 {
 struct anon__lazy_441 precision;
 struct anon__lazy_441 width;
 uint32_t fill;
 uint8_t alignment;
};
typedef struct anon__lazy_105 nav__7342_39;
typedef struct anon__lazy_175 nav__7342_40;
typedef struct anon__lazy_236 nav__7342_42;
typedef struct anon__lazy_57 nav__7342_48;
struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161;
struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *context;
};
typedef struct anon__lazy_161 nav__7342_71;
typedef struct anon__lazy_161 nav__7844_41;
typedef struct anon__lazy_57 nav__7844_43;
typedef struct anon__lazy_441 nav__7844_53;
typedef struct anon__lazy_105 nav__7344_39;
typedef struct anon__lazy_175 nav__7344_40;
typedef struct anon__lazy_161 nav__7344_54;
typedef struct anon__lazy_57 nav__7344_56;
typedef struct anon__lazy_161 nav__7846_41;
typedef struct anon__lazy_57 nav__7846_43;
typedef struct anon__lazy_105 nav__7345_39;
typedef struct anon__lazy_175 nav__7345_40;
typedef struct anon__lazy_161 nav__7345_54;
typedef struct anon__lazy_57 nav__7345_56;
typedef struct anon__lazy_57 nav__7847_40;
typedef struct anon__lazy_241 nav__7847_41;
typedef struct anon__lazy_161 nav__7847_44;
typedef struct anon__lazy_441 nav__7847_55;
typedef struct anon__lazy_105 nav__7346_39;
typedef struct anon__lazy_175 nav__7346_40;
typedef struct anon__lazy_57 nav__7346_42;
typedef struct anon__lazy_241 nav__7346_43;
typedef struct anon__lazy_161 nav__7346_59;
typedef struct anon__lazy_105 nav__1364_41;
typedef struct anon__lazy_398 nav__1604_38;
typedef struct anon__lazy_105 nav__1410_39;
typedef struct anon__lazy_105 nav__7434_39;
typedef struct anon__lazy_161 nav__7434_50;
typedef struct anon__lazy_105 nav__7435_39;
typedef struct anon__lazy_175 nav__7435_40;
typedef struct anon__lazy_279 nav__7435_51;
typedef struct anon__lazy_105 nav__7436_40;
typedef struct anon__lazy_161 nav__7437_48;
typedef struct anon__lazy_57 nav__7437_50;
typedef struct anon__lazy_57 nav__7438_45;
typedef struct anon__lazy_57 nav__7439_45;
typedef struct anon__lazy_161 nav__7441_48;
typedef struct anon__lazy_57 nav__7441_50;
typedef struct anon__lazy_161 nav__7459_38;
typedef struct anon__lazy_57 nav__7459_41;
typedef struct anon__lazy_57 nav__7467_40;
typedef struct anon__lazy_161 nav__7467_43;
typedef struct anon__lazy_57 nav__7856_44;
typedef struct anon__lazy_441 nav__7856_56;
typedef struct anon__lazy_161 nav__7856_60;
typedef struct anon__lazy_491 nav__7856_66;
struct anon__lazy_491 {
 uint32_t f0;
 struct anon__lazy_57 f1;
};
typedef struct anon__lazy_496 nav__7856_76;
struct anon__lazy_496 {
 struct anon__lazy_57 f0;
 struct anon__lazy_57 f1;
};
typedef struct anon__lazy_498 nav__7856_78;
struct anon__lazy_498 {
 uint32_t f0;
 struct shared_MarkerType__2045 f1;
};
typedef struct anon__lazy_500 nav__7856_80;
struct anon__lazy_500 {
 uint64_t f0;
};
typedef struct anon__lazy_503 nav__7856_84;
struct anon__lazy_503 {
 uint8_t f0;
};
typedef struct anon__lazy_57 nav__7843_44;
typedef struct anon__lazy_441 nav__7843_56;
typedef struct anon__lazy_161 nav__7843_60;
typedef struct anon__lazy_105 nav__4172_39;
typedef struct anon__lazy_441 nav__7845_41;
typedef struct anon__lazy_161 nav__7845_45;
typedef struct anon__lazy_57 nav__7845_47;
typedef struct anon__lazy_57 nav__7848_39;
typedef struct anon__lazy_441 nav__7848_44;
typedef struct anon__lazy_161 nav__7848_48;
typedef struct anon__lazy_105 nav__7852_39;
typedef struct anon__lazy_105 nav__7853_39;
typedef struct anon__lazy_161 nav__4100_38;
typedef struct anon__lazy_57 nav__4100_41;
typedef struct anon__lazy_57 nav__7854_40;
typedef struct anon__lazy_161 nav__4167_38;
typedef struct anon__lazy_57 nav__4167_42;
typedef struct anon__lazy_105 nav__4167_50;
typedef struct anon__lazy_161 nav__7466_38;
typedef struct anon__lazy_57 nav__7466_41;
typedef struct anon__lazy_57 nav__7857_40;
typedef struct anon__lazy_491 nav__7857_41;
typedef struct anon__lazy_161 nav__7857_44;
typedef struct anon__lazy_57 nav__7858_40;
typedef struct anon__lazy_496 nav__7858_41;
typedef struct anon__lazy_161 nav__7858_44;
typedef struct anon__lazy_498 nav__7859_40;
typedef struct anon__lazy_161 nav__7859_43;
typedef struct anon__lazy_57 nav__7859_45;
typedef struct anon__lazy_500 nav__7860_39;
typedef struct anon__lazy_161 nav__7860_42;
typedef struct anon__lazy_57 nav__7860_44;
typedef struct anon__lazy_503 nav__7861_39;
typedef struct anon__lazy_161 nav__7861_42;
typedef struct anon__lazy_57 nav__7861_44;
typedef struct anon__lazy_214 nav__7488_38;
typedef struct anon__lazy_161 nav__7504_38;
typedef struct anon__lazy_57 nav__7504_40;
typedef struct anon__lazy_214 nav__7504_48;
typedef struct anon__lazy_149 nav__7504_50;
typedef struct anon__lazy_57 nav__7863_39;
typedef struct anon__lazy_441 nav__7863_44;
typedef struct anon__lazy_161 nav__7863_48;
typedef struct anon__lazy_105 nav__7863_57;
typedef struct anon__lazy_214 nav__7863_60;
typedef struct anon__lazy_441 nav__7872_41;
typedef struct anon__lazy_161 nav__7872_45;
typedef struct anon__lazy_57 nav__7872_47;
typedef struct anon__lazy_546 nav__7872_56;
struct anon__lazy_546 {
 uint8_t array[2];
};
typedef struct anon__lazy_105 nav__7872_58;
typedef struct anon__lazy_441 nav__7871_41;
typedef struct anon__lazy_161 nav__7871_45;
typedef struct anon__lazy_57 nav__7871_47;
typedef struct anon__lazy_441 nav__7862_41;
typedef struct anon__lazy_161 nav__7862_45;
typedef struct anon__lazy_57 nav__7862_47;
typedef struct anon__lazy_161 nav__1388_38;
typedef struct anon__lazy_57 nav__1388_42;
typedef struct anon__lazy_57 nav__7864_40;
typedef struct anon__lazy_161 nav__7864_51;
typedef struct anon__lazy_57 nav__7866_40;
typedef struct anon__lazy_491 nav__7866_41;
typedef struct anon__lazy_161 nav__7866_44;
typedef struct anon__lazy_441 nav__7866_55;
typedef struct anon__lazy_57 nav__7867_40;
typedef struct anon__lazy_496 nav__7867_41;
typedef struct anon__lazy_161 nav__7867_44;
typedef struct anon__lazy_441 nav__7867_55;
typedef struct anon__lazy_498 nav__7868_40;
typedef struct anon__lazy_161 nav__7868_43;
typedef struct anon__lazy_57 nav__7868_45;
typedef struct anon__lazy_441 nav__7868_58;
typedef struct anon__lazy_500 nav__7869_39;
typedef struct anon__lazy_161 nav__7869_42;
typedef struct anon__lazy_57 nav__7869_44;
typedef struct anon__lazy_441 nav__7869_55;
typedef struct anon__lazy_503 nav__7870_39;
typedef struct anon__lazy_161 nav__7870_42;
typedef struct anon__lazy_57 nav__7870_44;
typedef struct anon__lazy_441 nav__7870_55;
typedef struct anon__lazy_546 nav__7243_39;
typedef struct anon__lazy_149 nav__7494_38;
typedef struct anon__lazy_57 nav__7494_40;
typedef struct anon__lazy_546 nav__7494_48;
typedef struct anon__lazy_566 nav__7494_52;
struct anon__lazy_566 {
 uint8_t array[3];
};
typedef struct anon__lazy_568 nav__7494_56;
struct anon__lazy_568 {
 uint8_t array[4];
};
typedef struct anon__lazy_214 nav__7487_38;
typedef struct anon__lazy_214 nav__7878_38;
typedef struct anon__lazy_105 nav__7878_40;
typedef struct anon__lazy_214 nav__7489_38;
typedef struct anon__lazy_105 nav__7489_40;
typedef struct anon__lazy_57 nav__7471_40;
typedef struct anon__lazy_161 nav__7471_43;
typedef struct anon__lazy_57 nav__1382_41;
typedef struct anon__lazy_441 nav__7873_41;
typedef struct anon__lazy_161 nav__7873_45;
typedef struct anon__lazy_57 nav__7873_47;
typedef struct anon__lazy_57 nav__7874_39;
typedef struct anon__lazy_441 nav__7874_44;
typedef struct anon__lazy_161 nav__7874_48;
typedef struct anon__lazy_441 nav__7880_44;
typedef struct anon__lazy_161 nav__7880_48;
typedef struct anon__lazy_57 nav__7880_50;
typedef struct anon__lazy_500 nav__7880_56;
typedef struct anon__lazy_441 nav__7875_44;
typedef struct anon__lazy_161 nav__7875_48;
typedef struct anon__lazy_57 nav__7875_50;
typedef struct anon__lazy_441 nav__7876_41;
typedef struct anon__lazy_161 nav__7876_45;
typedef struct anon__lazy_57 nav__7876_47;
typedef struct anon__lazy_441 nav__7877_41;
typedef struct anon__lazy_161 nav__7877_45;
typedef struct anon__lazy_57 nav__7877_47;
typedef struct anon__lazy_149 nav__7496_38;
typedef struct anon__lazy_546 nav__7496_40;
typedef struct anon__lazy_149 nav__7498_38;
typedef struct anon__lazy_566 nav__7498_40;
typedef struct anon__lazy_149 nav__7502_38;
typedef struct anon__lazy_568 nav__7502_40;
typedef struct anon__lazy_105 nav__1401_43;
typedef struct anon__lazy_161 nav__1401_52;
typedef struct anon__lazy_57 nav__1383_41;
typedef struct anon__lazy_105 nav__1383_47;
typedef struct anon__lazy_441 nav__7887_41;
typedef struct anon__lazy_161 nav__7887_45;
typedef struct anon__lazy_57 nav__7887_47;
typedef struct anon__lazy_546 nav__7887_56;
typedef struct anon__lazy_105 nav__7887_58;
typedef struct anon__lazy_441 nav__7886_41;
typedef struct anon__lazy_161 nav__7886_45;
typedef struct anon__lazy_57 nav__7886_47;
typedef struct anon__lazy_441 nav__7879_41;
typedef struct anon__lazy_161 nav__7879_45;
typedef struct anon__lazy_57 nav__7879_47;
typedef struct anon__lazy_500 nav__7881_39;
typedef struct anon__lazy_161 nav__7881_42;
typedef struct anon__lazy_57 nav__7881_44;
typedef struct anon__lazy_500 nav__7882_39;
typedef struct anon__lazy_161 nav__7882_42;
typedef struct anon__lazy_57 nav__7882_44;
typedef struct anon__lazy_500 nav__7883_39;
typedef struct anon__lazy_161 nav__7883_42;
typedef struct anon__lazy_57 nav__7883_44;
typedef struct anon__lazy_500 nav__7884_39;
typedef struct anon__lazy_161 nav__7884_42;
typedef struct anon__lazy_57 nav__7884_44;
typedef struct anon__lazy_441 nav__7892_41;
typedef struct anon__lazy_161 nav__7892_45;
typedef struct anon__lazy_57 nav__7892_47;
typedef struct anon__lazy_441 nav__7885_41;
typedef struct anon__lazy_161 nav__7885_45;
typedef struct anon__lazy_57 nav__7885_47;
typedef struct anon__lazy_149 nav__7500_38;
typedef struct anon__lazy_566 nav__7500_40;
typedef struct anon__lazy_161 nav__1326_38;
typedef struct anon__lazy_426 nav__1326_42;
typedef struct anon__lazy_500 nav__7888_39;
typedef struct anon__lazy_161 nav__7888_42;
typedef struct anon__lazy_57 nav__7888_44;
typedef struct anon__lazy_441 nav__7888_55;
typedef struct anon__lazy_500 nav__7889_39;
typedef struct anon__lazy_161 nav__7889_42;
typedef struct anon__lazy_57 nav__7889_44;
typedef struct anon__lazy_441 nav__7889_55;
typedef struct anon__lazy_500 nav__7890_39;
typedef struct anon__lazy_161 nav__7890_42;
typedef struct anon__lazy_57 nav__7890_44;
typedef struct anon__lazy_441 nav__7890_55;
typedef struct anon__lazy_500 nav__7891_39;
typedef struct anon__lazy_161 nav__7891_42;
typedef struct anon__lazy_57 nav__7891_44;
typedef struct anon__lazy_441 nav__7891_55;
typedef struct anon__lazy_105 nav__238_54;
typedef struct anon__lazy_105 nav__761_51;
typedef struct anon__lazy_105 nav__774_51;
typedef struct anon__lazy_105 nav__7893_52;
typedef struct anon__lazy_105 nav__239_54;
typedef struct anon__lazy_105 nav__1087_50;
typedef struct anon__lazy_57 nav__1087_63;
typedef struct anon__lazy_229 nav__1087_76;
typedef struct anon__lazy_105 nav__763_51;
typedef struct anon__lazy_105 nav__776_51;
typedef struct anon__lazy_105 nav__240_54;
typedef struct anon__lazy_105 nav__1088_50;
typedef struct anon__lazy_57 nav__1088_63;
typedef struct anon__lazy_229 nav__1088_76;
typedef struct anon__lazy_105 nav__764_51;
typedef struct anon__lazy_105 nav__777_51;
typedef struct anon__lazy_105 nav__241_54;
typedef struct anon__lazy_105 nav__1089_51;
typedef struct anon__lazy_57 nav__1089_61;
typedef struct anon__lazy_229 nav__1089_76;
typedef struct anon__lazy_105 nav__765_52;
typedef struct anon__lazy_105 nav__778_52;
typedef struct anon__lazy_105 nav__242_55;
typedef struct anon__lazy_57 nav__7898_39;
typedef struct anon__lazy_633 nav__7900_45;
struct anon__lazy_633 {
 bool array[64];
};
typedef struct anon__lazy_635 nav__7900_47;
struct anon__lazy_635 {
 bool is_null;
 uint8_t payload;
};
typedef struct anon__lazy_635 nav__7908_38;
typedef struct anon__lazy_633 nav__7908_40;
typedef struct anon__lazy_105 nav__1090_51;
typedef struct anon__lazy_57 nav__1090_61;
typedef struct anon__lazy_229 nav__1090_76;
typedef struct anon__lazy_105 nav__766_52;
typedef struct anon__lazy_105 nav__779_52;
typedef struct anon__lazy_105 nav__244_55;
typedef struct anon__lazy_650 nav__244_71;
struct anon__lazy_650 {
 uint8_t const *f0;
 uint8_t const *f1;
};
typedef struct anon__lazy_175 nav__244_73;
typedef struct anon__lazy_105 nav__7916_39;
typedef struct anon__lazy_175 nav__7916_40;
typedef struct anon__lazy_650 nav__7916_43;
typedef struct anon__lazy_105 nav__7917_40;
typedef struct anon__lazy_105 nav__7918_39;
typedef struct anon__lazy_175 nav__7918_40;
typedef struct anon__lazy_650 nav__7918_43;
typedef struct anon__lazy_441 nav__7918_54;
typedef struct anon__lazy_105 nav__7919_39;
typedef struct anon__lazy_650 nav__7961_40;
typedef struct anon__lazy_161 nav__7961_43;
typedef struct anon__lazy_57 nav__7961_44;
typedef struct anon__lazy_441 nav__7961_55;
typedef struct anon__lazy_650 nav__7920_39;
struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790;
struct io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29__6782;
struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 {
 uint64_t bytes_written;
};
struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804;
struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *context;
};
typedef struct anon__lazy_161 nav__7920_53;
typedef struct anon__lazy_57 nav__7920_54;
typedef struct anon__lazy_441 nav__7921_38;
typedef struct anon__lazy_105 nav__7922_39;
typedef struct anon__lazy_175 nav__7922_40;
typedef struct anon__lazy_650 nav__7922_42;
typedef struct anon__lazy_161 nav__7922_57;
typedef struct anon__lazy_57 nav__7922_58;
typedef struct anon__lazy_161 nav__7960_38;
typedef struct anon__lazy_57 nav__7960_41;
typedef struct anon__lazy_441 nav__7962_42;
typedef struct anon__lazy_161 nav__7962_46;
typedef struct anon__lazy_57 nav__7962_47;
typedef struct anon__lazy_161 nav__7945_38;
typedef struct anon__lazy_57 nav__7945_42;
typedef struct anon__lazy_161 nav__4078_38;
typedef struct anon__lazy_57 nav__4078_40;
typedef struct anon__lazy_105 nav__1091_54;
typedef struct anon__lazy_57 nav__1091_69;
typedef struct anon__lazy_229 nav__1091_77;
typedef struct anon__lazy_105 nav__767_55;
typedef struct anon__lazy_105 nav__780_55;
typedef struct anon__lazy_105 nav__249_57;
typedef struct anon__lazy_310 nav__3971_41;
struct Target_Os__665;
union Target_Os_VersionRange__721;
struct SemanticVersion_Range__726;
struct SemanticVersion__724;
typedef struct anon__lazy_57 nav__263_43;
struct SemanticVersion__724 {
 uintptr_t major;
 uintptr_t minor;
 uintptr_t patch;
 struct anon__lazy_57 pre;
 struct anon__lazy_57 build;
};
struct SemanticVersion_Range__726 {
 struct SemanticVersion__724 zig_e_min;
 struct SemanticVersion__724 zig_e_max;
};
struct Target_Os_HurdVersionRange__728;
struct Target_Os_HurdVersionRange__728 {
 struct SemanticVersion_Range__726 range;
 struct SemanticVersion__724 glibc;
};
struct Target_Os_LinuxVersionRange__730;
struct Target_Os_LinuxVersionRange__730 {
 struct SemanticVersion_Range__726 range;
 struct SemanticVersion__724 glibc;
 uint32_t android;
};
struct Target_Os_WindowsVersion_Range__786;
struct Target_Os_WindowsVersion_Range__786 {
 uint32_t zig_e_min;
 uint32_t zig_e_max;
};
union Target_Os_VersionRange__721 {
 struct SemanticVersion_Range__726 semver;
 struct Target_Os_HurdVersionRange__728 hurd;
 struct Target_Os_LinuxVersionRange__730 linux;
 struct Target_Os_WindowsVersion_Range__786 windows;
};
struct Target_Os__665 {
 union Target_Os_VersionRange__721 version_range;
 uint8_t tag;
};
struct Target_DynamicLinker__844;
struct Target_DynamicLinker__844 {
 uint8_t buffer[255];
 uint8_t len;
};
typedef struct anon__lazy_105 nav__3714_40;
typedef struct anon__lazy_105 nav__3715_40;
typedef struct anon__lazy_105 nav__3716_40;
typedef struct anon__lazy_105 nav__3678_46;
typedef struct anon__lazy_105 nav__235_46;
struct Target_Cpu_Feature_Set__861;
struct Target_Cpu_Feature_Set__861 {
 uintptr_t ints[5];
};
struct Target_Cpu__829;
struct Target_Cpu_Model__856;
struct Target_Cpu__829 {
 struct Target_Cpu_Model__856 const *model;
 struct Target_Cpu_Feature_Set__861 features;
 uint8_t arch;
};
typedef struct anon__lazy_57 nav__262_46;
struct Target_Cpu_Model__856 {
 struct anon__lazy_57 name;
 struct anon__lazy_57 llvm_name;
 struct Target_Cpu_Feature_Set__861 features;
};
struct Target__663;
typedef struct anon__lazy_57 nav__264_51;
struct Target__663 {
 struct Target_Cpu__829 cpu;
 struct Target_Os__665 os;
 uint8_t abi;
 uint8_t ofmt;
 struct Target_DynamicLinker__844 dynamic_linker;
};
struct builtin_CallingConvention__266;
struct builtin_CallingConvention_CommonOptions__268;
typedef struct anon__lazy_229 nav__463_40;
struct builtin_CallingConvention_CommonOptions__268 {
 struct anon__lazy_229 incoming_stack_alignment;
};
struct builtin_CallingConvention_X86RegparmOptions__270;
struct builtin_CallingConvention_X86RegparmOptions__270 {
 struct anon__lazy_229 incoming_stack_alignment;
 uint8_t register_params;
};
struct builtin_CallingConvention_ArmInterruptOptions__272;
struct builtin_CallingConvention_ArmInterruptOptions__272 {
 struct anon__lazy_229 incoming_stack_alignment;
 uint8_t type;
};
struct builtin_CallingConvention_MipsInterruptOptions__274;
struct builtin_CallingConvention_MipsInterruptOptions__274 {
 struct anon__lazy_229 incoming_stack_alignment;
 uint8_t mode;
};
struct builtin_CallingConvention_RiscvInterruptOptions__276;
struct builtin_CallingConvention_RiscvInterruptOptions__276 {
 struct anon__lazy_229 incoming_stack_alignment;
 uint8_t mode;
};
struct builtin_CallingConvention__266 {
 union {
  struct builtin_CallingConvention_CommonOptions__268 x86_64_sysv;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_win;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_regcall_v3_sysv;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_regcall_v4_win;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_vectorcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_interrupt;
  struct builtin_CallingConvention_X86RegparmOptions__270 x86_sysv;
  struct builtin_CallingConvention_X86RegparmOptions__270 x86_win;
  struct builtin_CallingConvention_X86RegparmOptions__270 x86_stdcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_fastcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_thiscall;
  struct builtin_CallingConvention_CommonOptions__268 x86_thiscall_mingw;
  struct builtin_CallingConvention_CommonOptions__268 x86_regcall_v3;
  struct builtin_CallingConvention_CommonOptions__268 x86_regcall_v4_win;
  struct builtin_CallingConvention_CommonOptions__268 x86_vectorcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_aapcs;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_aapcs_darwin;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_aapcs_win;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_vfabi;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_vfabi_sve;
  struct builtin_CallingConvention_CommonOptions__268 arm_aapcs;
  struct builtin_CallingConvention_CommonOptions__268 arm_aapcs_vfp;
  struct builtin_CallingConvention_ArmInterruptOptions__272 arm_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 mips64_n64;
  struct builtin_CallingConvention_CommonOptions__268 mips64_n32;
  struct builtin_CallingConvention_MipsInterruptOptions__274 mips64_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 mips_o32;
  struct builtin_CallingConvention_MipsInterruptOptions__274 mips_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 riscv64_lp64;
  struct builtin_CallingConvention_CommonOptions__268 riscv64_lp64_v;
  struct builtin_CallingConvention_RiscvInterruptOptions__276 riscv64_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 riscv32_ilp32;
  struct builtin_CallingConvention_CommonOptions__268 riscv32_ilp32_v;
  struct builtin_CallingConvention_RiscvInterruptOptions__276 riscv32_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 sparc64_sysv;
  struct builtin_CallingConvention_CommonOptions__268 sparc_sysv;
  struct builtin_CallingConvention_CommonOptions__268 powerpc64_elf;
  struct builtin_CallingConvention_CommonOptions__268 powerpc64_elf_altivec;
  struct builtin_CallingConvention_CommonOptions__268 powerpc64_elf_v2;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_sysv;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_sysv_altivec;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_aix;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_aix_altivec;
  struct builtin_CallingConvention_CommonOptions__268 wasm_mvp;
  struct builtin_CallingConvention_CommonOptions__268 arc_sysv;
  struct builtin_CallingConvention_CommonOptions__268 bpf_std;
  struct builtin_CallingConvention_CommonOptions__268 csky_sysv;
  struct builtin_CallingConvention_CommonOptions__268 csky_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 hexagon_sysv;
  struct builtin_CallingConvention_CommonOptions__268 hexagon_sysv_hvx;
  struct builtin_CallingConvention_CommonOptions__268 lanai_sysv;
  struct builtin_CallingConvention_CommonOptions__268 loongarch64_lp64;
  struct builtin_CallingConvention_CommonOptions__268 loongarch32_ilp32;
  struct builtin_CallingConvention_CommonOptions__268 m68k_sysv;
  struct builtin_CallingConvention_CommonOptions__268 m68k_gnu;
  struct builtin_CallingConvention_CommonOptions__268 m68k_rtd;
  struct builtin_CallingConvention_CommonOptions__268 m68k_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 msp430_eabi;
  struct builtin_CallingConvention_CommonOptions__268 propeller_sysv;
  struct builtin_CallingConvention_CommonOptions__268 s390x_sysv;
  struct builtin_CallingConvention_CommonOptions__268 s390x_sysv_vx;
  struct builtin_CallingConvention_CommonOptions__268 ve_sysv;
  struct builtin_CallingConvention_CommonOptions__268 xcore_xs1;
  struct builtin_CallingConvention_CommonOptions__268 xcore_xs2;
  struct builtin_CallingConvention_CommonOptions__268 xtensa_call0;
  struct builtin_CallingConvention_CommonOptions__268 xtensa_windowed;
  struct builtin_CallingConvention_CommonOptions__268 amdgcn_device;
  struct builtin_CallingConvention_CommonOptions__268 amdgcn_cs;
 } payload;
 uint8_t tag;
};
struct std_Options__4590;
typedef struct anon__lazy_441 nav__97_39;
struct std_Options__4590 {
 struct anon__lazy_441 page_size_min;
 struct anon__lazy_441 page_size_max;
 uintptr_t fmt_max_depth;
 bool enable_segfault_handler;
 uint8_t log_level;
 bool crypto_always_getrandom;
 bool crypto_fork_safety;
 bool keep_sigpipe;
 bool http_disable_tls;
 bool http_enable_ssl_key_log_file;
 uint8_t side_channels_mitigations;
};
typedef struct anon__lazy_441 nav__3694_38;
typedef struct anon__lazy_105 nav__3705_43;
typedef struct anon__lazy_57 nav__3833_40;
static uint8_t const __anon_1941[21];
static uint8_t const __anon_2012[21];
static char const __anon_3681[33];
static uint8_t const __anon_2558[62];
static uint8_t const __anon_3768[34];
static uint8_t const __anon_4279[49];
static uint8_t const __anon_4659[1];
static uint8_t const __anon_4761[68];
static uint8_t const __anon_4872[82];
static uint8_t const __anon_4926[32];
static uint8_t const __anon_5059[22];
static uint8_t const __anon_2281[15];
static uint8_t const __anon_2286[14];
static uint8_t const __anon_2291[4];
static uint8_t const __anon_2296[9];
static uint8_t const __anon_2304[4];
static uint8_t const __anon_2319[19];
static uint8_t const __anon_5389[8];
static uint8_t const __anon_5566[3];
static uint8_t const __anon_5638[25];
static uint8_t const __anon_5704[7];
static uint8_t const __anon_5729[2];
static uint8_t const __anon_5782[23];
static uint8_t const __anon_5807[11];
static uint8_t const __anon_5863[17];
static uint8_t const __anon_5885[10];
static uint8_t const __anon_5955[31];
static uint8_t const __anon_6010[24];
static uint8_t const __anon_6125[23];
static uint8_t const __anon_6130[2];
static uint8_t const __anon_6221[12];
static uint8_t const __anon_6231[1];
static uint8_t const __anon_6250[10];
static uint8_t const __anon_6282[15];
static uint8_t const __anon_6310[13];
static uint8_t const __anon_6848[10];
static uint8_t const __anon_6863[1];
static uint8_t const __anon_6876[1];
static uint8_t const __anon_5243[4];
static uint8_t const __anon_6988[7];
#define c_instrument_hooks_set_feature__236 instrument_hooks_set_feature
zig_extern void instrument_hooks_set_feature(uint64_t, bool);
static void features_set_feature__323(uint64_t, bool);
static void bit_set_IntegerBitSet_2864_29_set__362(uint64_t *, uintptr_t);
static void bit_set_IntegerBitSet_2864_29_unset__364(uint64_t *, uintptr_t);
static void debug_assert__180(bool);
static uint64_t bit_set_IntegerBitSet_2864_29_maskBit__383(uintptr_t);
static nav__739_39 instruments_valgrind_ValgrindInstrument_init__739(struct mem_Allocator__589);
static nav__1115_39 fifo_UnixPipe_openPipe__1115(nav__1115_41);
static nav__1117_39 fifo_UnixPipe_openWrite__1117(struct mem_Allocator__589, nav__1117_42);
static nav__1116_39 fifo_UnixPipe_openRead__1116(struct mem_Allocator__589, nav__1116_42);
static nav__1082_39 runner_fifo_RunnerFifo_init__1082(struct mem_Allocator__589);
static uint16_t fifo_UnixPipe_Writer_sendCmd__1121(struct fifo_UnixPipe_Writer__628 *, struct shared_Command__2038);
static nav__4211_38 bincode_deserializeInt__anon_2394__4211(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263);
static nav__4210_38 bincode_deserializeAlloc__anon_2383__4210(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4253_40 bincode_deserializePointerAlloc__anon_2520__4253(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4252_40 bincode_deserializeAlloc__anon_2507__4252(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4251_39 bincode_deserializeStructAlloc__anon_2482__4251(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4250_39 bincode_deserializeAlloc__anon_2453__4250(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static uint16_t bincode_deserializeAlloc__anon_2579__4257(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4259_39 bincode_deserializeStructAlloc__anon_2643__4259(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4258_39 bincode_deserializeAlloc__anon_2624__4258(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4265_38 bincode_deserializeInt__anon_2766__4265(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263);
static nav__4264_38 bincode_deserializeAlloc__anon_2761__4264(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4263_39 bincode_deserializeUnionAlloc__anon_2748__4263(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4262_39 bincode_deserializeAlloc__anon_2714__4262(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4261_39 bincode_deserializeStructAlloc__anon_2696__4261(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4260_39 bincode_deserializeAlloc__anon_2682__4260(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4267_38 bincode_deserializeEnum__anon_2852__4267(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263);
static nav__4266_38 bincode_deserializeAlloc__anon_2823__4266(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4209_39 bincode_deserializeUnionAlloc__anon_2378__4209(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__4208_39 bincode_deserializeAlloc__anon_2271__4208(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263, struct mem_Allocator__589);
static nav__1126_39 fifo_UnixPipe_Reader_recvCmd__1126(struct fifo_UnixPipe_Reader__630 *);
static nav__1127_39 fifo_UnixPipe_Reader_waitForResponse__1127(struct fifo_UnixPipe_Reader__630 *, nav__1127_42);
static uint16_t fifo_UnixPipe_Reader_waitForAck__1128(struct fifo_UnixPipe_Reader__630 *, nav__1128_40);
static uint16_t runner_fifo_RunnerFifo_send_version__1092(struct runner_fifo_RunnerFifo__621 *, uint64_t);
static uint16_t runner_fifo_RunnerFifo_validate_protocol_version__1083(struct runner_fifo_RunnerFifo__621 *);
static nav__1093_38 runner_fifo_RunnerFifo_get_integration_mode__1093(struct runner_fifo_RunnerFifo__621 *);
static nav__773_39 instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_init__773(struct mem_Allocator__589);
static nav__760_39 instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_init__760(struct mem_Allocator__589);
static nav__726_39 instruments_root_InstrumentHooks_init__726(struct mem_Allocator__589);
#define c_instrument_hooks_init__237 instrument_hooks_init
zig_extern struct instruments_root_InstrumentHooks__547 *instrument_hooks_init(void);
static nav__3717_40 mem_Allocator_create__anon_901__3717(struct mem_Allocator__589);
static uint16_t fs_accessAbsolute__1183(nav__1183_39, struct fs_File_OpenFlags__1954);
static nav__1180_39 fs_openFileAbsolute__1180(nav__1180_41, struct fs_File_OpenFlags__1954);
static void utils_setNonBlocking__3970(int32_t);
static struct fifo_UnixPipe_Writer__628 fifo_UnixPipe_Writer_init__1118(struct fs_File__636, struct mem_Allocator__589);
static struct fifo_UnixPipe_Reader__630 fifo_UnixPipe_Reader_init__1123(struct fs_File__636, struct mem_Allocator__589);
static void array_list_ArrayListAligned_28u8_2cnull_29_clearRetainingCapacity__1397(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *);
static struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 array_list_ArrayListAligned_28u8_2cnull_29_writer__1387(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *);
static uint16_t bincode_serialize__anon_2090__4101(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_Command__2038);
static uint8_t const (*mem_asBytes__anon_2128__4102(uint32_t const *))[4];
static uint16_t fs_File_writeAll__1274(struct fs_File__636, nav__1274_40);
static zig_i128 time_nanoTimestamp__4115(void);
static nav__1264_38 fs_File_readAll__1264(struct fs_File__636, nav__1264_41);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_resize__1394(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *, uintptr_t);
static struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 io_fixed_buffer_stream_fixedBufferStream__anon_2242__4174(nav__4174_40);
static struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_reader__4163(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *);
static nav__4207_38 io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4207(void const *, nav__4207_41);
static nav__4249_39 io_Reader_readBytesNoEof__anon_2440__4249(struct io_Reader__2409);
static nav__4254_39 io_Reader_readBytesNoEof__anon_2533__4254(struct io_Reader__2409);
static nav__4255_40 mem_Allocator_alloc__anon_2542__4255(struct mem_Allocator__589, uintptr_t);
static nav__4214_38 io_Reader_readAll__4214(struct io_Reader__2409, nav__4214_41);
static zig_noreturn void bincode_invalidProtocol__anon_2564__4256(void);
static void utils_sleep__3969(uint64_t);
static void shared_Command_deinit__3974(struct shared_Command__2038, struct mem_Allocator__589);
static void logger_debug__anon_2894__4277(nav__4277_39);
static void logger_err__anon_2917__4278(void);
static void logger_err__anon_2925__4279(void);
static zig_noreturn void posix_exit__1582(uint8_t);
static void logger_err__anon_2938__4280(nav__4280_40);
static void runner_fifo_RunnerFifo_deinit__1084(struct runner_fifo_RunnerFifo__621 *);
static nav__4281_39 mem_Allocator_allocBytesWithAlignment__anon_2980__4281(struct mem_Allocator__589, uintptr_t, uintptr_t);
static bool fs_path_isAbsolute__4316(nav__4316_39);
static struct fs_Dir__1956 fs_cwd__1175(void);
static uint16_t fs_Dir_access__3918(struct fs_Dir__1956, nav__3918_40, struct fs_File_OpenFlags__1954);
static nav__3854_39 fs_Dir_openFile__3854(struct fs_Dir__1956, nav__3854_42, struct fs_File_OpenFlags__1954);
static struct array_list_ArrayListAligned_28u8_2cnull_29__648 array_list_ArrayListAligned_28u8_2cnull_29_init__1362(struct mem_Allocator__589);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1399(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *, uintptr_t);
static uint16_t bincode_serializeUnion__anon_3043__5311(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_Command__2038);
static nav__1273_38 fs_File_write__1273(struct fs_File__636, nav__1273_41);
static nav__1773_39 posix_clock_gettime__1773(uint32_t);
static nav__1263_38 fs_File_read__1263(struct fs_File__636, nav__1263_41);
static nav__4166_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_read__4166(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *, nav__4166_42);
static uint16_t io_Reader_readNoEof__4216(struct io_Reader__2409, nav__4216_40);
static nav__5312_39 mem_Allocator_allocWithSizeAndAlignment__anon_3137__5312(struct mem_Allocator__589, uintptr_t, uintptr_t);
static nav__4215_38 io_Reader_readAtLeast__4215(struct io_Reader__2409, nav__4215_41, uintptr_t);
static void utils_print__anon_3157__5313(void);
static nav__6247_38 math_cast__anon_3184__6247(uint64_t);
static void mem_Allocator_free__anon_3261__7028(struct mem_Allocator__589, nav__7028_40);
static void logger_logWithPrefix__anon_3269__7029(nav__7029_39);
static void logger_logWithPrefix__anon_3271__7030(void);
static void logger_logWithPrefix__anon_3272__7031(void);
static void logger_logWithPrefix__anon_3277__7033(nav__7033_40);
static void fifo_UnixPipe_Writer_deinit__1122(struct fifo_UnixPipe_Writer__628 *);
static void fifo_UnixPipe_Reader_deinit__1129(struct fifo_UnixPipe_Reader__630 *);
static uint8_t mem_Alignment_fromByteUnits__1066(uintptr_t);
static bool fs_path_isAbsolutePosix__4322(nav__4322_39);
static nav__1857_39 posix_toPosixPath__1857(nav__1857_41);
static uint16_t fs_Dir_accessZ__3919(struct fs_Dir__1956, uint8_t const *, struct fs_File_OpenFlags__1954);
static nav__3855_39 fs_Dir_openFileZ__3855(struct fs_Dir__1956, uint8_t const *, struct fs_File_OpenFlags__1954);
static uintptr_t array_list_ArrayListAlignedUnmanaged_28u8_2cnull_29_growCapacity__7104(uintptr_t, uintptr_t);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacityPrecise__1400(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *, uintptr_t);
static uint16_t bincode_serialize__anon_3381__7105(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, uint32_t);
static uint16_t bincode_serialize__anon_3383__7106(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_Command__struct_2041__2041);
static uint16_t bincode_serialize__anon_3385__7107(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080);
static uint16_t bincode_serialize__anon_3387__7108(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_Command__struct_2042__2042);
static uint16_t bincode_serialize__anon_3389__7109(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_Command__struct_2043__2043);
static uint16_t bincode_serialize__anon_3391__7110(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, uint64_t);
static uint16_t bincode_serialize__anon_3393__7111(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, uint8_t);
static nav__1592_38 posix_write__1592(int32_t, nav__1592_40);
static uint16_t posix_errno__anon_3559__7178(int);
static uint16_t posix_unexpectedErrno__1856(uint16_t);
static nav__1584_38 posix_read__1584(int32_t, nav__1584_40);
static nav__7179_38 math_mul__anon_3583__7179(uintptr_t, uintptr_t);
static nav__7180_39 mem_Allocator_allocBytesWithAlignment__anon_3585__7180(struct mem_Allocator__589, uintptr_t, uintptr_t);
static nav__4213_38 io_Reader_read__4213(struct io_Reader__2409, nav__4213_41);
static nav__7182_39 mem_sliceAsBytes__anon_3715__7182(nav__7182_39);
static uint16_t fmt_format__anon_4185__7476(struct io_Writer__4171, nav__7476_40);
static nav__7342_40 fmt_bufPrint__anon_3764__7342(nav__7342_39, nav__7342_42);
static uint16_t fmt_format__anon_4674__7844(struct io_Writer__4171);
static nav__7344_40 fmt_bufPrint__anon_3785__7344(nav__7344_39);
static uint16_t fmt_format__anon_4777__7846(struct io_Writer__4171);
static nav__7345_40 fmt_bufPrint__anon_3795__7345(nav__7345_39);
static uint16_t fmt_format__anon_4878__7847(struct io_Writer__4171, nav__7847_41);
static nav__7346_40 fmt_bufPrint__anon_3807__7346(nav__7346_39, nav__7346_43);
static void array_list_ArrayListAligned_28u8_2cnull_29_deinit__1364(struct array_list_ArrayListAligned_28u8_2cnull_29__648);
static void fs_File_close__1220(struct fs_File__636);
static bool math_isPowerOfTwo__anon_3820__7347(uintptr_t);
static uint16_t posix_faccessatZ__1748(int32_t, uint8_t const *, uint32_t, uint32_t);
static nav__1604_38 posix_openatZ__1604(int32_t, uint8_t const *, uint32_t, uintptr_t);
static uint16_t posix_flock__1765(int32_t, int32_t);
static void posix_close__1562(int32_t);
static nav__1410_39 array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1410(struct array_list_ArrayListAligned_28u8_2cnull_29__648);
static nav__7434_39 mem_Allocator_remap__anon_3973__7434(struct mem_Allocator__589, nav__7434_39, uintptr_t);
static nav__7435_40 mem_Allocator_alignedAlloc__anon_3978__7435(struct mem_Allocator__589, uintptr_t);
static void mem_Allocator_free__anon_3980__7436(struct mem_Allocator__589, nav__7436_40);
static uint16_t bincode_serializeInt__anon_3985__7437(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, uint32_t);
static uint16_t bincode_serializeStruct__anon_3990__7438(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_Command__struct_2041__2041);
static uint16_t bincode_serializeStruct__anon_3991__7439(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_Command__struct_2042__2042);
static uint16_t bincode_serializeStruct__anon_3992__7440(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_Command__struct_2043__2043);
static uint16_t bincode_serializeInt__anon_3993__7441(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, uint64_t);
static uint16_t bincode_serializeEnum__anon_3996__7442(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, uint8_t);
static uint16_t posix_errno__anon_4007__7444(intptr_t);
static struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4164(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *);
static nav__7459_38 io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__7459(void const *, nav__7459_41);
static uint16_t io_Writer_writeAll__7467(struct io_Writer__4171, nav__7467_40);
static uint16_t shared_Command_format__anon_5290__7856(struct shared_Command__2038, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static uint16_t fmt_formatType__anon_4653__7843(struct shared_Command__2038, struct fmt_FormatOptions__4577, struct io_Writer__4171, uintptr_t);
static nav__4172_39 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4172(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239);
static uint16_t fmt_formatType__anon_4769__7845(uint64_t, struct fmt_FormatOptions__4577, struct io_Writer__4171, uintptr_t);
static uint16_t fmt_formatType__anon_5030__7848(nav__7848_39, struct fmt_FormatOptions__4577, struct io_Writer__4171, uintptr_t);
static nav__7852_39 mem_sliceAsBytes__anon_5184__7852(nav__7852_39);
static nav__7853_39 mem_bytesAsSlice__anon_5197__7853(nav__7853_39);
static nav__4100_38 io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4100(void const *, nav__4100_41);
static uint16_t bincode_serialize__anon_5228__7854(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, nav__7854_40);
static uint16_t bincode_serialize__anon_5230__7855(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_MarkerType__2045);
static nav__4167_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_write__4167(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *, nav__4167_42);
static nav__7466_38 io_Writer_write__7466(struct io_Writer__4171, nav__7466_41);
static uint16_t io_Writer_print__anon_5310__7857(struct io_Writer__4171, nav__7857_41);
static uint16_t io_Writer_print__anon_5331__7858(struct io_Writer__4171, nav__7858_41);
static uint16_t io_Writer_print__anon_5343__7859(struct io_Writer__4171, nav__7859_40);
static uint16_t io_Writer_print__anon_5352__7860(struct io_Writer__4171, nav__7860_39);
static uint16_t io_Writer_print__anon_5365__7861(struct io_Writer__4171, nav__7861_39);
static nav__7488_38 unicode_utf8ByteSequenceLength__7488(uint8_t);
static nav__7504_38 unicode_utf8CountCodepoints__7504(nav__7504_40);
static uint16_t fmt_formatBuf__anon_5399__7863(nav__7863_39, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static uint16_t fmt_formatInt__anon_5459__7872(uint64_t, uint8_t, uint8_t, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static uint16_t fmt_formatIntValue__anon_5442__7871(uint64_t, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static uint16_t fmt_formatValue__anon_5374__7862(uint64_t, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static nav__1388_38 array_list_ArrayListAligned_28u8_2cnull_29_appendWrite__1388(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *, nav__1388_42);
static uint16_t bincode_serializePointer__anon_5411__7864(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, nav__7864_40);
static uint16_t bincode_serializeUnion__anon_5412__7865(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080, struct shared_MarkerType__2045);
static uint16_t fmt_format__anon_5416__7866(struct io_Writer__4171, nav__7866_41);
static uint16_t fmt_format__anon_5420__7867(struct io_Writer__4171, nav__7867_41);
static uint16_t fmt_format__anon_5424__7868(struct io_Writer__4171, nav__7868_40);
static uint16_t fmt_format__anon_5428__7869(struct io_Writer__4171, nav__7869_39);
static uint16_t fmt_format__anon_5432__7870(struct io_Writer__4171, nav__7870_39);
static nav__7243_39 fmt_digits2__7243(uint8_t);
static uint8_t fmt_digitToChar__7259(uint8_t, uint8_t);
static nav__7494_38 unicode_utf8Decode__7494(nav__7494_40);
static nav__7487_38 unicode_utf8CodepointSequenceLength__7487(uint32_t);
static nav__7878_38 unicode_utf8EncodeImpl__anon_6041__7878(uint32_t, nav__7878_40);
static nav__7489_38 unicode_utf8Encode__7489(uint32_t, nav__7489_40);
static uint16_t io_Writer_writeBytesNTimes__7471(struct io_Writer__4171, nav__7471_40, uintptr_t);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_appendSlice__1382(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *, nav__1382_41);
static uint16_t fmt_formatType__anon_5690__7873(uint32_t, struct fmt_FormatOptions__4577, struct io_Writer__4171, uintptr_t);
static uint16_t fmt_formatType__anon_5720__7874(nav__7874_39, struct fmt_FormatOptions__4577, struct io_Writer__4171, uintptr_t);
static uint16_t shared_MarkerType_format__anon_6084__7880(struct shared_MarkerType__2045, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static uint16_t fmt_formatType__anon_5895__7875(struct shared_MarkerType__2045, struct fmt_FormatOptions__4577, struct io_Writer__4171, uintptr_t);
static uint16_t fmt_formatType__anon_5964__7876(uint64_t, struct fmt_FormatOptions__4577, struct io_Writer__4171, uintptr_t);
static uint16_t fmt_formatType__anon_6018__7877(uint8_t, struct fmt_FormatOptions__4577, struct io_Writer__4171, uintptr_t);
static nav__7496_38 unicode_utf8Decode2__7496(nav__7496_40);
static nav__7498_38 unicode_utf8Decode3__7498(nav__7498_40);
static nav__7502_38 unicode_utf8Decode4__7502(nav__7502_40);
static bool unicode_isSurrogateCodepoint__7567(uint32_t);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureUnusedCapacity__1401(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *, uintptr_t);
static void array_list_ArrayListAligned_28u8_2cnull_29_appendSliceAssumeCapacity__1383(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *, nav__1383_41);
static uint16_t fmt_formatInt__anon_6172__7887(uint32_t, uint8_t, uint8_t, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static uint16_t fmt_formatIntValue__anon_6161__7886(uint32_t, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static uint16_t fmt_formatValue__anon_6065__7879(uint32_t, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static uint16_t io_Writer_print__anon_6098__7881(struct io_Writer__4171, nav__7881_39);
static uint16_t io_Writer_print__anon_6103__7882(struct io_Writer__4171, nav__7882_39);
static uint16_t io_Writer_print__anon_6110__7883(struct io_Writer__4171, nav__7883_39);
static uint16_t io_Writer_print__anon_6116__7884(struct io_Writer__4171, nav__7884_39);
static uint16_t fmt_formatIntValue__anon_6198__7892(uint64_t, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static uint16_t fmt_formatValue__anon_6118__7885(uint64_t, struct fmt_FormatOptions__4577, struct io_Writer__4171);
static nav__7500_38 unicode_utf8Decode3AllowSurrogateHalf__7500(nav__7500_40);
static nav__1326_38 array_list_addOrOom__1326(uintptr_t, uintptr_t);
static uint16_t fmt_format__anon_6182__7888(struct io_Writer__4171, nav__7888_39);
static uint16_t fmt_format__anon_6186__7889(struct io_Writer__4171, nav__7889_39);
static uint16_t fmt_format__anon_6190__7890(struct io_Writer__4171, nav__7890_39);
static uint16_t fmt_format__anon_6194__7891(struct io_Writer__4171, nav__7891_39);
#define c_instrument_hooks_deinit__238 instrument_hooks_deinit
zig_extern void instrument_hooks_deinit(struct instruments_root_InstrumentHooks__547 *);
static void instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_deinit__761(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *);
static void instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_deinit__774(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *);
static void mem_Allocator_destroy__anon_6345__7893(struct mem_Allocator__589, struct instruments_root_InstrumentHooks__547 *);
#define c_instrument_hooks_is_instrumented__239 instrument_hooks_is_instrumented
zig_extern bool instrument_hooks_is_instrumented(struct instruments_root_InstrumentHooks__547 *);
static zig_cold uint16_t runner_fifo_RunnerFifo_start_benchmark__1087(struct runner_fifo_RunnerFifo__621 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_start_benchmark__763(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_start_benchmark__776(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *);
#define c_instrument_hooks_start_benchmark__240 instrument_hooks_start_benchmark
zig_extern uint8_t instrument_hooks_start_benchmark(struct instruments_root_InstrumentHooks__547 *);
static bool features_is_feature_enabled__324(uint64_t);
static bool bit_set_IntegerBitSet_2864_29_isSet__359(uint64_t, uintptr_t);
static zig_cold uint16_t runner_fifo_RunnerFifo_stop_benchmark__1088(struct runner_fifo_RunnerFifo__621 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_stop_benchmark__764(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_stop_benchmark__777(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *);
#define c_instrument_hooks_stop_benchmark__241 instrument_hooks_stop_benchmark
zig_extern uint8_t instrument_hooks_stop_benchmark(struct instruments_root_InstrumentHooks__547 *);
static uint16_t runner_fifo_RunnerFifo_set_executed_benchmark__1089(struct runner_fifo_RunnerFifo__621 *, uint32_t, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_set_executed_benchmark__765(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *, uint32_t, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_executed_benchmark__778(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *, uint32_t, uint8_t const *);
#define c_instrument_hooks_set_executed_benchmark__242 instrument_hooks_set_executed_benchmark
zig_extern uint8_t instrument_hooks_set_executed_benchmark(struct instruments_root_InstrumentHooks__547 *, uint32_t, uint8_t const *);
static nav__7898_39 mem_span__anon_6487__7898(uint8_t const *);
static uintptr_t mem_len__anon_6504__7899(uint8_t const *);
static uintptr_t mem_indexOfSentinel__anon_6511__7900(uint8_t const *);
static bool math_isPowerOfTwo__anon_6522__7901(void);
static nav__7908_38 simd_firstTrue__anon_6579__7908(nav__7908_40);
static uintptr_t mem_alignForward__anon_6582__7909(uintptr_t, uintptr_t);
static bool mem_isAligned__935(uintptr_t, uintptr_t);
static bool mem_isValidAlignGeneric__anon_6671__7910(uintptr_t);
static uintptr_t mem_alignBackward__anon_6672__7911(uintptr_t, uintptr_t);
static bool mem_isAlignedGeneric__anon_6676__7912(uint64_t, uint64_t);
static uint64_t mem_alignBackward__anon_6678__7913(uint64_t, uint64_t);
static bool mem_isValidAlignGeneric__anon_6680__7914(uint64_t);
static bool math_isPowerOfTwo__anon_6681__7915(uint64_t);
#define c_instrument_hooks_executed_benchmark__243 instrument_hooks_executed_benchmark
zig_extern uint8_t instrument_hooks_executed_benchmark(struct instruments_root_InstrumentHooks__547 *, uint32_t, uint8_t const *);
static uint16_t runner_fifo_RunnerFifo_set_integration__1090(struct runner_fifo_RunnerFifo__621 *, uint8_t const *, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_set_integration__766(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *, uint8_t const *, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_integration__779(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *, uint8_t const *, uint8_t const *);
#define c_instrument_hooks_set_integration__244 instrument_hooks_set_integration
zig_extern uint8_t instrument_hooks_set_integration(struct instruments_root_InstrumentHooks__547 *, uint8_t const *, uint8_t const *);
static nav__7916_40 fmt_allocPrintZ__anon_6716__7916(struct mem_Allocator__589, nav__7916_43);
static void mem_Allocator_free__anon_6718__7917(struct mem_Allocator__589, nav__7917_40);
static nav__7918_40 fmt_allocPrint__anon_6753__7918(struct mem_Allocator__589, nav__7918_43);
static nav__7919_39 mem_sliceAsBytes__anon_6760__7919(nav__7919_39);
static uint16_t fmt_format__anon_6823__7961(struct io_Writer__4171, nav__7961_40);
static uint64_t fmt_count__anon_6767__7920(nav__7920_39);
static nav__7921_38 math_cast__anon_6769__7921(uint64_t);
static nav__7922_40 fmt_bufPrint__anon_6772__7922(nav__7922_39, nav__7922_42);
static struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 io_counting_writer_countingWriter__anon_6793__7947(void);
static struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_writer__7946(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *);
static nav__7960_38 io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWr__7960(void const *, nav__7960_41);
static uint16_t fmt_formatType__anon_6857__7962(uint8_t const *, struct fmt_FormatOptions__4577, struct io_Writer__4171, uintptr_t);
static nav__7945_38 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_write__7945(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *, nav__7945_42);
static nav__4078_38 io_dummyWrite__4078(nav__4078_40);
static uint16_t runner_fifo_RunnerFifo_add_marker__1091(struct runner_fifo_RunnerFifo__621 *, uint32_t, struct shared_MarkerType__2045);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_add_marker__767(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *, uint32_t, struct shared_MarkerType__2045);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_add_marker__780(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *, uint32_t, struct shared_MarkerType__2045);
#define c_instrument_hooks_add_marker__249 instrument_hooks_add_marker
zig_extern uint8_t instrument_hooks_add_marker(struct instruments_root_InstrumentHooks__547 *, uint32_t, uint8_t, uint64_t);
#define c_instrument_hooks_current_timestamp__250 instrument_hooks_current_timestamp
zig_extern uint64_t instrument_hooks_current_timestamp(void);
static uint64_t utils_clock_gettime_monotonic__3971(void);
static uint64_t const builtin_zig_backend__255;
static bool const start_simplified_logic__109;
static uint8_t const builtin_output_mode__256;
static uint8_t const builtin_link_mode__257;
static uint64_t features_features__322;
static uintptr_t const bit_set_IntegerBitSet_2864_29_bit_length__353;
static bool const builtin_link_libc__267;
static bool const posix_use_libc__1425;
static struct Target_Os__665 const builtin_os__263;
static uint8_t const c_native_os__1867;
static struct Target_DynamicLinker__844 const Target_DynamicLinker_none__3589;
static bool const builtin_is_test__259;
static uint8_t *heap_CAllocator_alloc__3713(void *, uintptr_t, uint8_t, uintptr_t);
static bool heap_CAllocator_resize__3714(void *, nav__3714_40, uint8_t, uintptr_t, uintptr_t);
static uint8_t *heap_CAllocator_remap__3715(void *, nav__3715_40, uint8_t, uintptr_t, uintptr_t);
static void heap_CAllocator_free__3716(void *, nav__3716_40, uint8_t, uintptr_t);
static uint8_t *heap_CAllocator_alignedAlloc__3710(uintptr_t, uint8_t);
static uintptr_t heap_CAllocator_alignedAllocSize__3712(uint8_t *);
static void heap_CAllocator_alignedFree__3711(uint8_t *);
static uintptr_t mem_Alignment_toByteUnits__1065(uint8_t);
static struct mem_Allocator__589 const heap_c_allocator__3678;
static struct mem_Allocator__589 const c_allocator__235;
static struct Target_Cpu_Feature_Set__861 const Target_Cpu_Feature_Set_empty__3634;
static struct Target_Cpu__829 const builtin_cpu__262;
static uint8_t const builtin_abi__261;
static uint8_t const builtin_object_format__265;
static struct Target__663 const builtin_target__264;
static struct builtin_CallingConvention__266 const builtin_CallingConvention_c__463;
zig_extern uint8_t running_on_valgrind(void);
static uint8_t const (*const shared_RUNNER_CTL_FIFO__754)[21];
static uint8_t const (*const shared_RUNNER_ACK_FIFO__755)[21];
static uint64_t const runner_fifo_PROTOCOL_VERSION__1078;
static uint8_t const mem_native_endian__789;
zig_extern int fcntl(int, int, ...);
static int const cimport_F_GETFL__5193;
static int const cimport_O_NONBLOCK__5169;
static int const cimport_F_SETFL__5194;
static uint8_t const os_linux_native_arch__2773;
zig_extern int nanosleep(struct cimport_struct_timespec__3171 const *, struct cimport_struct_timespec__3171 *);
static int const cimport_EINTR__6896;
zig_extern zig_noreturn void exit(int);
static uint8_t const fs_path_native_os__4295;
static uint8_t const fs_native_os__1138;
static uint8_t const fs_Dir_native_os__3960;
static bool const fs_File_is_windows__1314;
static uint8_t const posix_native_os__1423;
zig_extern int clock_gettime(uint32_t, struct os_linux_timespec__struct_3100__3100 *);
zig_extern int printf(char const *, ...);
static uint8_t logger_max_level__4271;
zig_extern int printf(char const *, ...);
static uint8_t const builtin_mode__266;
static bool const debug_runtime_safety__159;
zig_extern int flock(int32_t, int);
static bool const fs_Dir_have_flock__3961;
zig_extern intptr_t write(int32_t, uint8_t const *, uintptr_t);
static uint8_t const c_native_abi__1865;
zig_extern int *zig_e___errno_location(void) zig_mangled(zig_e___errno_location, "__errno_location");
static bool const posix_unexpected_error_tracing__1854;
zig_extern intptr_t read(int32_t, uint8_t *, uintptr_t);
static uint16_t const fmt_max_format_args__7205;
static bool const debug_default_enable_segfault_handler__205;
static uint8_t const log_default_level__7777;
static struct std_Options__4590 const std_options__97;
zig_extern int faccessat(int32_t, uint8_t const *, unsigned int, unsigned int);
static bool const posix_lfs64_abi__1853;
zig_extern int openat64(int, uint8_t const *, uint32_t, ...);
zig_extern int close(int32_t);
static uint8_t const (*const fmt_ANY__7208)[4];
static uint8_t const unicode_native_endian__7485;
static uint32_t const unicode_replacement_character__7486;
zig_extern void callgrind_zero_stats(void);
zig_extern void callgrind_start_instrumentation(void);
zig_extern void callgrind_stop_instrumentation(void);
zig_extern void callgrind_dump_stats_at(uint8_t const *);
static bool const mem_backend_supports_vectors__817;
static bool const builtin_valgrind_support__270;
static nav__3694_38 const heap_page_size_min_default__3694;
static uintptr_t const heap_page_size_min__3673;
static uint8_t const c_MARKER_TYPE_SAMPLE_START__245;
static uint8_t const c_MARKER_TYPE_SAMPLE_END__246;
static uint8_t const c_MARKER_TYPE_BENCHMARK_START__247;
static uint8_t const c_MARKER_TYPE_BENCHMARK_END__248;
static struct mem_Allocator_VTable__592 const heap_CAllocator_vtable__3705;
zig_extern uintptr_t malloc_usable_size(void const *);
static bool const heap_CAllocator_supports_malloc_size__3706;
static bool const heap_CAllocator_supports_posix_memalign__3708;
zig_extern int posix_memalign(void **, uintptr_t, uintptr_t);
zig_extern void free(void *);
static struct Target_Cpu_Model__856 const Target_x86_cpu_znver4__3833;
enum {
 zig_error_ModeError = 1u,
 zig_error_OutOfMemory = 2u,
 zig_error_NotInstrumented = 3u,
 zig_error_PermissionDenied = 4u,
 zig_error_FileNotFound = 5u,
 zig_error_NameTooLong = 6u,
 zig_error_InputOutput = 7u,
 zig_error_SystemResources = 8u,
 zig_error_BadPathName = 9u,
 zig_error_FileBusy = 10u,
 zig_error_SymLinkLoop = 11u,
 zig_error_ReadOnlyFileSystem = 12u,
 zig_error_InvalidUtf8 = 13u,
 zig_error_InvalidWtf8 = 14u,
 zig_error_Unexpected = 15u,
 zig_error_SharingViolation = 16u,
 zig_error_PathAlreadyExists = 17u,
 zig_error_AccessDenied = 18u,
 zig_error_PipeBusy = 19u,
 zig_error_NoDevice = 20u,
 zig_error_NetworkNotFound = 21u,
 zig_error_AntivirusInterference = 22u,
 zig_error_ProcessFdQuotaExceeded = 23u,
 zig_error_SystemFdQuotaExceeded = 24u,
 zig_error_FileTooBig = 25u,
 zig_error_IsDir = 26u,
 zig_error_NoSpaceLeft = 27u,
 zig_error_NotDir = 28u,
 zig_error_DeviceBusy = 29u,
 zig_error_FileLocksNotSupported = 30u,
 zig_error_WouldBlock = 31u,
 zig_error_DiskQuota = 32u,
 zig_error_InvalidArgument = 33u,
 zig_error_BrokenPipe = 34u,
 zig_error_OperationAborted = 35u,
 zig_error_NotOpenForWriting = 36u,
 zig_error_LockViolation = 37u,
 zig_error_ConnectionResetByPeer = 38u,
 zig_error_ProcessNotFound = 39u,
 zig_error_AckTimeout = 40u,
 zig_error_ConnectionTimedOut = 41u,
 zig_error_NotOpenForReading = 42u,
 zig_error_SocketNotConnected = 43u,
 zig_error_Canceled = 44u,
 zig_error_NotReady = 45u,
 zig_error_UnexpectedEof = 46u,
 zig_error_EndOfStream = 47u,
 zig_error_UnexpectedError = 48u,
 zig_error_UnexpectedResponse = 49u,
 zig_error_UnsupportedClock = 50u,
 zig_error_Overflow = 51u,
 zig_error_Utf8ExpectedContinuation = 52u,
 zig_error_Utf8OverlongEncoding = 53u,
 zig_error_Utf8EncodesSurrogateHalf = 54u,
 zig_error_Utf8CodepointTooLarge = 55u,
 zig_error_Utf8InvalidStartByte = 56u,
 zig_error_TruncatedInput = 57u,
 zig_error_Utf8CannotEncodeSurrogateHalf = 58u,
 zig_error_CodepointTooLarge = 59u,
};
static uint8_t const zig_errorName_ModeError[10] = "ModeError";
static uint8_t const zig_errorName_OutOfMemory[12] = "OutOfMemory";
static uint8_t const zig_errorName_NotInstrumented[16] = "NotInstrumented";
static uint8_t const zig_errorName_PermissionDenied[17] = "PermissionDenied";
static uint8_t const zig_errorName_FileNotFound[13] = "FileNotFound";
static uint8_t const zig_errorName_NameTooLong[12] = "NameTooLong";
static uint8_t const zig_errorName_InputOutput[12] = "InputOutput";
static uint8_t const zig_errorName_SystemResources[16] = "SystemResources";
static uint8_t const zig_errorName_BadPathName[12] = "BadPathName";
static uint8_t const zig_errorName_FileBusy[9] = "FileBusy";
static uint8_t const zig_errorName_SymLinkLoop[12] = "SymLinkLoop";
static uint8_t const zig_errorName_ReadOnlyFileSystem[19] = "ReadOnlyFileSystem";
static uint8_t const zig_errorName_InvalidUtf8[12] = "InvalidUtf8";
static uint8_t const zig_errorName_InvalidWtf8[12] = "InvalidWtf8";
static uint8_t const zig_errorName_Unexpected[11] = "Unexpected";
static uint8_t const zig_errorName_SharingViolation[17] = "SharingViolation";
static uint8_t const zig_errorName_PathAlreadyExists[18] = "PathAlreadyExists";
static uint8_t const zig_errorName_AccessDenied[13] = "AccessDenied";
static uint8_t const zig_errorName_PipeBusy[9] = "PipeBusy";
static uint8_t const zig_errorName_NoDevice[9] = "NoDevice";
static uint8_t const zig_errorName_NetworkNotFound[16] = "NetworkNotFound";
static uint8_t const zig_errorName_AntivirusInterference[22] = "AntivirusInterference";
static uint8_t const zig_errorName_ProcessFdQuotaExceeded[23] = "ProcessFdQuotaExceeded";
static uint8_t const zig_errorName_SystemFdQuotaExceeded[22] = "SystemFdQuotaExceeded";
static uint8_t const zig_errorName_FileTooBig[11] = "FileTooBig";
static uint8_t const zig_errorName_IsDir[6] = "IsDir";
static uint8_t const zig_errorName_NoSpaceLeft[12] = "NoSpaceLeft";
static uint8_t const zig_errorName_NotDir[7] = "NotDir";
static uint8_t const zig_errorName_DeviceBusy[11] = "DeviceBusy";
static uint8_t const zig_errorName_FileLocksNotSupported[22] = "FileLocksNotSupported";
static uint8_t const zig_errorName_WouldBlock[11] = "WouldBlock";
static uint8_t const zig_errorName_DiskQuota[10] = "DiskQuota";
static uint8_t const zig_errorName_InvalidArgument[16] = "InvalidArgument";
static uint8_t const zig_errorName_BrokenPipe[11] = "BrokenPipe";
static uint8_t const zig_errorName_OperationAborted[17] = "OperationAborted";
static uint8_t const zig_errorName_NotOpenForWriting[18] = "NotOpenForWriting";
static uint8_t const zig_errorName_LockViolation[14] = "LockViolation";
static uint8_t const zig_errorName_ConnectionResetByPeer[22] = "ConnectionResetByPeer";
static uint8_t const zig_errorName_ProcessNotFound[16] = "ProcessNotFound";
static uint8_t const zig_errorName_AckTimeout[11] = "AckTimeout";
static uint8_t const zig_errorName_ConnectionTimedOut[19] = "ConnectionTimedOut";
static uint8_t const zig_errorName_NotOpenForReading[18] = "NotOpenForReading";
static uint8_t const zig_errorName_SocketNotConnected[19] = "SocketNotConnected";
static uint8_t const zig_errorName_Canceled[9] = "Canceled";
static uint8_t const zig_errorName_NotReady[9] = "NotReady";
static uint8_t const zig_errorName_UnexpectedEof[14] = "UnexpectedEof";
static uint8_t const zig_errorName_EndOfStream[12] = "EndOfStream";
static uint8_t const zig_errorName_UnexpectedError[16] = "UnexpectedError";
static uint8_t const zig_errorName_UnexpectedResponse[19] = "UnexpectedResponse";
static uint8_t const zig_errorName_UnsupportedClock[17] = "UnsupportedClock";
static uint8_t const zig_errorName_Overflow[9] = "Overflow";
static uint8_t const zig_errorName_Utf8ExpectedContinuation[25] = "Utf8ExpectedContinuation";
static uint8_t const zig_errorName_Utf8OverlongEncoding[21] = "Utf8OverlongEncoding";
static uint8_t const zig_errorName_Utf8EncodesSurrogateHalf[25] = "Utf8EncodesSurrogateHalf";
static uint8_t const zig_errorName_Utf8CodepointTooLarge[22] = "Utf8CodepointTooLarge";
static uint8_t const zig_errorName_Utf8InvalidStartByte[21] = "Utf8InvalidStartByte";
static uint8_t const zig_errorName_TruncatedInput[15] = "TruncatedInput";
static uint8_t const zig_errorName_Utf8CannotEncodeSurrogateHalf[30] = "Utf8CannotEncodeSurrogateHalf";
static uint8_t const zig_errorName_CodepointTooLarge[18] = "CodepointTooLarge";
static struct anon__lazy_57 const zig_errorName[60] = {{zig_errorName_ModeError, 9ul}, {zig_errorName_OutOfMemory, 11ul}, {zig_errorName_NotInstrumented, 15ul}, {zig_errorName_PermissionDenied, 16ul}, {zig_errorName_FileNotFound, 12ul}, {zig_errorName_NameTooLong, 11ul}, {zig_errorName_InputOutput, 11ul}, {zig_errorName_SystemResources, 15ul}, {zig_errorName_BadPathName, 11ul}, {zig_errorName_FileBusy, 8ul}, {zig_errorName_SymLinkLoop, 11ul}, {zig_errorName_ReadOnlyFileSystem, 18ul}, {zig_errorName_InvalidUtf8, 11ul}, {zig_errorName_InvalidWtf8, 11ul}, {zig_errorName_Unexpected, 10ul}, {zig_errorName_SharingViolation, 16ul}, {zig_errorName_PathAlreadyExists, 17ul}, {zig_errorName_AccessDenied, 12ul}, {zig_errorName_PipeBusy, 8ul}, {zig_errorName_NoDevice, 8ul}, {zig_errorName_NetworkNotFound, 15ul}, {zig_errorName_AntivirusInterference, 21ul}, {zig_errorName_ProcessFdQuotaExceeded, 22ul}, {zig_errorName_SystemFdQuotaExceeded, 21ul}, {zig_errorName_FileTooBig, 10ul}, {zig_errorName_IsDir, 5ul}, {zig_errorName_NoSpaceLeft, 11ul}, {zig_errorName_NotDir, 6ul}, {zig_errorName_DeviceBusy, 10ul}, {zig_errorName_FileLocksNotSupported, 21ul}, {zig_errorName_WouldBlock, 10ul}, {zig_errorName_DiskQuota, 9ul}, {zig_errorName_InvalidArgument, 15ul}, {zig_errorName_BrokenPipe, 10ul}, {zig_errorName_OperationAborted, 16ul}, {zig_errorName_NotOpenForWriting, 17ul}, {zig_errorName_LockViolation, 13ul}, {zig_errorName_ConnectionResetByPeer, 21ul}, {zig_errorName_ProcessNotFound, 15ul}, {zig_errorName_AckTimeout, 10ul}, {zig_errorName_ConnectionTimedOut, 18ul}, {zig_errorName_NotOpenForReading, 17ul}, {zig_errorName_SocketNotConnected, 18ul}, {zig_errorName_Canceled, 8ul}, {zig_errorName_NotReady, 8ul}, {zig_errorName_UnexpectedEof, 13ul}, {zig_errorName_EndOfStream, 11ul}, {zig_errorName_UnexpectedError, 15ul}, {zig_errorName_UnexpectedResponse, 18ul}, {zig_errorName_UnsupportedClock, 16ul}, {zig_errorName_Overflow, 8ul}, {zig_errorName_Utf8ExpectedContinuation, 24ul}, {zig_errorName_Utf8OverlongEncoding, 20ul}, {zig_errorName_Utf8EncodesSurrogateHalf, 24ul}, {zig_errorName_Utf8CodepointTooLarge, 21ul}, {zig_errorName_Utf8InvalidStartByte, 20ul}, {zig_errorName_TruncatedInput, 14ul}, {zig_errorName_Utf8CannotEncodeSurrogateHalf, 29ul}, {zig_errorName_CodepointTooLarge, 17ul}};
static struct anon__lazy_57 zig_tag_name_shared_IntegrationMode__563(uint8_t const tag) {
 switch (tag) {
  case UINT8_C(0): {
   static uint8_t const name[5] = "Perf";
   return (struct anon__lazy_57){name, (uintptr_t)4ul};
  }
  case UINT8_C(1): {
   static uint8_t const name[11] = "Simulation";
   return (struct anon__lazy_57){name, (uintptr_t)10ul};
  }
  case UINT8_C(2): {
   static uint8_t const name[9] = "Analysis";
   return (struct anon__lazy_57){name, (uintptr_t)8ul};
  }
 }
 while (true) zig_breakpoint();
}

static uint8_t const __anon_1941[21] = "/tmp/runner.ctl.fifo";

static uint8_t const __anon_2012[21] = "/tmp/runner.ack.fifo";

static char const __anon_3681[33] = {73,110,118,97,108,105,100,32,112,114,111,116,111,99,111,108,32,100,101,116,101,99,116,101,100,58,32,37,46,42,115,10,0};

static uint8_t const __anon_2558[62] = "The stream end was found before all required bytes were read.";

static uint8_t const __anon_3768[34] = "[ERROR] logger formatting failed\n";

static uint8_t const __anon_4279[49] = "[DEBUG] waitForAck received unexpected response: ";

static uint8_t const __anon_4659[1] = "\n";

static uint8_t const __anon_4761[68] = "[ERROR] instrument-hooks: CodSpeed runner rejected protocol version ";

static uint8_t const __anon_4872[82] = "[ERROR] instrument-hooks: please update the CodSpeed action to the latest version\n";

static uint8_t const __anon_4926[32] = "[ERROR] instrument-hooks: error ";

static uint8_t const __anon_5059[22] = " during version check\n";

static uint8_t const __anon_2281[15] = "StartBenchmark";

static uint8_t const __anon_2286[14] = "StopBenchmark";

static uint8_t const __anon_2291[4] = "Ack";

static uint8_t const __anon_2296[9] = "PingPerf";

static uint8_t const __anon_2304[4] = "Err";

static uint8_t const __anon_2319[19] = "GetIntegrationMode";

static uint8_t const __anon_5389[8] = "{ ... }";

static uint8_t const __anon_5566[3] = "\357\277\275";

static uint8_t const __anon_5638[25] = "ExecutedBenchmark { pid: ";

static uint8_t const __anon_5704[7] = ", uri: ";

static uint8_t const __anon_5729[2] = " }";

static uint8_t const __anon_5782[23] = "SetIntegration { name: ";

static uint8_t const __anon_5807[11] = ", version: ";

static uint8_t const __anon_5863[17] = "AddMarker { pid: ";

static uint8_t const __anon_5885[10] = ", marker: ";

static uint8_t const __anon_5955[31] = "SetVersion { protocol_version: ";

static uint8_t const __anon_6010[24] = "IntegrationModeResponse ";

static uint8_t const __anon_6125[23] = "shared.IntegrationMode";

static uint8_t const __anon_6130[2] = ".";

static uint8_t const __anon_6221[12] = "SampleStart(";

static uint8_t const __anon_6231[1] = ")";

static uint8_t const __anon_6250[10] = "SampleEnd(";

static uint8_t const __anon_6282[15] = "BenchmarkStart(";

static uint8_t const __anon_6310[13] = "BenchmarkEnd(";

static uint8_t const __anon_6848[10] = "Metadata: ";

static uint8_t const __anon_6863[1] = " ";

static uint8_t const __anon_6876[1] = "\000";

static uint8_t const __anon_5243[4] = "any";

static uint8_t const __anon_6988[7] = "znver4";

void c_instrument_hooks_set_feature__236(uint64_t const a0, bool const a1) {
 uint64_t t0;
 t0 = a0;
 features_set_feature__323(t0, a1);
 return;
}

static void features_set_feature__323(uint64_t const a0, bool const a1) {
 uint64_t t0;
 uintptr_t t1;
 if (a1) {
  t0 = a0;
  t1 = t0;
  bit_set_IntegerBitSet_2864_29_set__362(&features_features__322, t1);
  goto zig_block_0;
 }
 t0 = a0;
 t1 = t0;
 bit_set_IntegerBitSet_2864_29_unset__364(&features_features__322, t1);
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void bit_set_IntegerBitSet_2864_29_set__362(uint64_t *const a0, uintptr_t const a1) {
 uint64_t *const *t1;
 uint64_t t2;
 uint64_t t5;
 uint64_t *t4;
 uint64_t *t0;
 bool t3;
 t0 = a0;
 t1 = (uint64_t *const *)&t0;
 t2 = a1;
 t3 = t2 < UINT64_C(64);
 debug_assert__180(t3);
 t4 = (*t1);
 t4 = (uint64_t *)((uint8_t *)t4 + (uintptr_t)0ul);
 t2 = (*t4);
 t5 = bit_set_IntegerBitSet_2864_29_maskBit__383(a1);
 t5 = t2 | t5;
 (*t4) = t5;
 return;
}

static void bit_set_IntegerBitSet_2864_29_unset__364(uint64_t *const a0, uintptr_t const a1) {
 uint64_t *const *t1;
 uint64_t t2;
 uint64_t t5;
 uint64_t *t4;
 uint64_t *t0;
 bool t3;
 t0 = a0;
 t1 = (uint64_t *const *)&t0;
 t2 = a1;
 t3 = t2 < UINT64_C(64);
 debug_assert__180(t3);
 t4 = (*t1);
 t4 = (uint64_t *)((uint8_t *)t4 + (uintptr_t)0ul);
 t2 = (*t4);
 t5 = bit_set_IntegerBitSet_2864_29_maskBit__383(a1);
 t5 = zig_not_u64(t5, UINT8_C(64));
 t5 = t2 & t5;
 (*t4) = t5;
 return;
}

static void debug_assert__180(bool const a0) {
 bool t0;
 t0 = !a0;
 if (t0) {
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static uint64_t bit_set_IntegerBitSet_2864_29_maskBit__383(uintptr_t const a0) {
 uint64_t t1;
 uint8_t t0;
 t0 = (uint8_t)a0;
 t1 = zig_shlw_u64(UINT64_C(1), t0, UINT8_C(64));
 return t1;
}

static nav__739_39 instruments_valgrind_ValgrindInstrument_init__739(struct mem_Allocator__589 const a0) {
 struct instruments_valgrind_ValgrindInstrument__554 t2;
 nav__739_39 t3;
 uint8_t t0;
 bool t1;
 t0 = running_on_valgrind();
 t1 = t0 > UINT8_C(0);
 t1 = !t1;
 if (t1) {
  return (nav__739_39){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},zig_error_NotInstrumented};
 }
 goto zig_block_0;

 zig_block_0:;
 t2.allocator = a0;
 t3.payload = t2;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__1115_39 fifo_UnixPipe_openPipe__1115(nav__1115_41 const a0) {
 nav__1115_39 t1;
 struct fs_File__636 t2;
 int32_t t3;
 uint16_t t0;
 t0 = fs_accessAbsolute__1183(a0, (struct fs_File_OpenFlags__1954){UINT8_C(2),UINT8_C(0),false,false});
 if (t0) {
  t1.payload = (struct fs_File__636){-INT32_C(0x55555556)};
  t1.error = t0;
  return t1;
 }
 t1 = fs_openFileAbsolute__1180(a0, (struct fs_File_OpenFlags__1954){UINT8_C(2),UINT8_C(0),false,false});
 if (t1.error) {
  t0 = t1.error;
  t1.payload = (struct fs_File__636){-INT32_C(0x55555556)};
  t1.error = t0;
  return t1;
 }
 t2 = t1.payload;
 t3 = t2.handle;
 utils_setNonBlocking__3970(t3);
 t1.payload = t2;
 t1.error = UINT16_C(0);
 return t1;
}

static nav__1117_39 fifo_UnixPipe_openWrite__1117(struct mem_Allocator__589 const a0, nav__1117_42 const a1) {
 nav__1117_39 t2;
 struct fifo_UnixPipe_Writer__628 t4;
 nav__1117_59 t0;
 struct fs_File__636 t3;
 uint16_t t1;
 t0 = fifo_UnixPipe_openPipe__1115(a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct fifo_UnixPipe_Writer__628){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = fifo_UnixPipe_Writer_init__1118(t3, a0);
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__1116_39 fifo_UnixPipe_openRead__1116(struct mem_Allocator__589 const a0, nav__1116_42 const a1) {
 nav__1116_39 t2;
 struct fifo_UnixPipe_Reader__630 t4;
 nav__1116_59 t0;
 struct fs_File__636 t3;
 uint16_t t1;
 t0 = fifo_UnixPipe_openPipe__1115(a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct fifo_UnixPipe_Reader__630){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = fifo_UnixPipe_Reader_init__1123(t3, a0);
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__1082_39 runner_fifo_RunnerFifo_init__1082(struct mem_Allocator__589 const a0) {
 struct runner_fifo_RunnerFifo__621 *t1;
 struct mem_Allocator__589 *t2;
 struct fifo_UnixPipe_Writer__628 *t3;
 nav__1082_66 t4;
 nav__1082_39 t6;
 nav__1082_39 t0;
 struct fifo_UnixPipe_Writer__628 t7;
 struct fifo_UnixPipe_Reader__630 *t8;
 nav__1082_71 t9;
 struct fifo_UnixPipe_Reader__630 t10;
 uint16_t t5;
 t0.error = UINT16_C(0);
 t1 = &t0.payload;
 t2 = (struct mem_Allocator__589 *)&t1->allocator;
 (*t2) = a0;
 t3 = (struct fifo_UnixPipe_Writer__628 *)&t1->writer;
 t4 = fifo_UnixPipe_openWrite__1117(a0, (nav__1082_64){(uint8_t const *)&__anon_1941,(uintptr_t)20ul});
 if (t4.error) {
  t5 = t4.error;
  t6.payload = (struct runner_fifo_RunnerFifo__621){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}};
  t6.error = t5;
  return t6;
 }
 t7 = t4.payload;
 (*t3) = t7;
 t8 = (struct fifo_UnixPipe_Reader__630 *)&t1->reader;
 t9 = fifo_UnixPipe_openRead__1116(a0, (nav__1082_64){(uint8_t const *)&__anon_2012,(uintptr_t)20ul});
 if (t9.error) {
  t5 = t9.error;
  t6.payload = (struct runner_fifo_RunnerFifo__621){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}};
  t6.error = t5;
  return t6;
 }
 t10 = t9.payload;
 (*t8) = t10;
 return t0;
}

static uint16_t fifo_UnixPipe_Writer_sendCmd__1121(struct fifo_UnixPipe_Writer__628 *const a0, struct shared_Command__2038 const a1) {
 struct fifo_UnixPipe_Writer__628 *const *t1;
 struct fifo_UnixPipe_Writer__628 *t2;
 struct fifo_UnixPipe_Writer__628 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t4;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t6;
 nav__1121_65 t7;
 struct fs_File__636 *t8;
 uintptr_t t10;
 uint32_t const *t13;
 uint8_t const (*t14)[4];
 nav__1121_44 t15;
 struct fs_File__636 t9;
 uint32_t t11;
 uint32_t t12;
 uint16_t t5;
 t0 = a0;
 t1 = (struct fifo_UnixPipe_Writer__628 *const *)&t0;
 t2 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&t2->buffer;
 array_list_ArrayListAligned_28u8_2cnull_29_clearRetainingCapacity__1397(t3);
 t2 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&t2->buffer;
 t4 = array_list_ArrayListAligned_28u8_2cnull_29_writer__1387(t3);
 t5 = bincode_serialize__anon_2090__4101(t4, a1);
 if (t5) {
  return t5;
 }
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&a0->buffer;
 t6 = (*t3);
 t7 = t6.items;
 t2 = (*t1);
 t8 = (struct fs_File__636 *)&t2->file;
 t9 = (*t8);
 t10 = t7.len;
 t11 = (uint32_t)t10;
 t12 = t11;
 t13 = (uint32_t const *)&t12;
 t14 = mem_asBytes__anon_2128__4102(t13);
 t15.ptr = &(*t14)[(uintptr_t)0ul];
 t15.len = (uintptr_t)4ul;
 t5 = fs_File_writeAll__1274(t9, t15);
 if (t5) {
  return t5;
 }
 t2 = (*t1);
 t8 = (struct fs_File__636 *)&t2->file;
 t9 = (*t8);
 memcpy(&t15, &t7, sizeof(nav__1121_44));
 t5 = fs_File_writeAll__1274(t9, t15);
 if (t5) {
  return t5;
 }
 return 0;
}

static nav__4211_38 bincode_deserializeInt__anon_2394__4211(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t2;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *const *t5;
 void const **t7;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t8;
 void const *t9;
 nav__4211_50 (**t10)(void const *, nav__4211_52);
 struct io_Reader__2409 t11;
 struct io_Reader__2409 t6;
 struct io_Reader__2409 t12;
 struct io_Reader__2409 t15;
 struct io_Reader__2409 const *t13;
 uint8_t const (*t21)[4];
 nav__4211_38 t14;
 nav__4211_38 t18;
 uint32_t t22;
 nav__4211_64 t16;
 uint16_t t17;
 uint8_t t19[4];
 uint8_t t20[4];
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__4211_50 (**)(void const *, nav__4211_52))&t6.readFn;
 (*t10) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4207;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Reader__2409 const *)&t12;
 t11 = (*t13);
 t15 = t11;
 t13 = (struct io_Reader__2409 const *)&t15;
 t11 = (*t13);
 t16 = io_Reader_readBytesNoEof__anon_2440__4249(t11);
 if (t16.error) {
  t17 = t16.error;
  t18.payload = UINT32_C(0xaaaaaaaa);
  t18.error = t17;
  t14 = t18;
  goto zig_block_0;
 }
 memcpy(t19, t16.payload, sizeof(uint8_t[4]));
 memcpy((char *)&t20, t19, sizeof(uint8_t[4]));
 t21 = (uint8_t const (*)[4])&t20;
 memcpy(t19, (const char *)t21, sizeof(uint8_t[4]));
 memcpy(&t22, &t19, sizeof(uint32_t));
 t22 = zig_wrap_u32(t22, UINT8_C(32));
 t18.payload = t22;
 t18.error = UINT16_C(0);
 t14 = t18;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t18, &t14, sizeof(nav__4211_38));
 if (t18.error) {
  t17 = t18.error;
  t18.payload = UINT32_C(0xaaaaaaaa);
  t18.error = t17;
  return t18;
 }
 t22 = t18.payload;
 t18.payload = t22;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__4210_38 bincode_deserializeAlloc__anon_2383__4210(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4210_38 t0;
 uint32_t t2;
 uint16_t t1;
 (void)a1;
 t0 = bincode_deserializeInt__anon_2394__4211(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = UINT32_C(0xaaaaaaaa);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4253_40 bincode_deserializePointerAlloc__anon_2520__4253(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *t6;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *t7;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *t32;
 struct mem_Allocator__589 const *t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t5;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t31;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *const *t8;
 void const **t10;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t11;
 void const *t12;
 nav__4253_59 (**t13)(void const *, nav__4253_39);
 struct io_Reader__2409 t14;
 struct io_Reader__2409 t9;
 struct io_Reader__2409 t15;
 struct io_Reader__2409 t18;
 struct io_Reader__2409 t33;
 struct io_Reader__2409 t34;
 struct io_Reader__2409 const *t16;
 nav__4253_69 t17;
 nav__4253_69 t21;
 uint8_t const (*t24)[8];
 uint64_t t25;
 uint64_t t38;
 nav__4253_40 t26;
 uintptr_t t27;
 uintptr_t t37;
 struct mem_Allocator__589 t29;
 struct mem_Allocator__589 t2;
 nav__4253_39 t30;
 nav__4253_39 t28;
 nav__4253_59 t35;
 nav__4253_59 t36;
 nav__4253_72 t19;
 uint16_t t20;
 uint8_t t22[8];
 uint8_t t23[8];
 bool t39;
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__589 const *)&t2;
 t4 = (*t1);
 t5 = t4;
 t6 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *)&t5;
 t7 = t6;
 t8 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *const *)&t7;
 t10 = (void const **)&t9.context;
 t6 = (*t8);
 t11 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t6->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__4253_59 (**)(void const *, nav__4253_39))&t9.readFn;
 (*t13) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4207;
 t14 = t9;
 t15 = t14;
 t16 = (struct io_Reader__2409 const *)&t15;
 t14 = (*t16);
 t18 = t14;
 t16 = (struct io_Reader__2409 const *)&t18;
 t14 = (*t16);
 t19 = io_Reader_readBytesNoEof__anon_2533__4254(t14);
 if (t19.error) {
  t20 = t19.error;
  t21.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t21.error = t20;
  t17 = t21;
  goto zig_block_0;
 }
 memcpy(t22, t19.payload, sizeof(uint8_t[8]));
 memcpy((char *)&t23, t22, sizeof(uint8_t[8]));
 t24 = (uint8_t const (*)[8])&t23;
 memcpy(t22, (const char *)t24, sizeof(uint8_t[8]));
 memcpy(&t25, &t22, sizeof(uint64_t));
 t25 = zig_wrap_u64(t25, UINT8_C(64));
 t21.payload = t25;
 t21.error = UINT16_C(0);
 t17 = t21;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t21, &t17, sizeof(nav__4253_69));
 if (t21.error) {
  t20 = t21.error;
  t26.payload = (nav__4253_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t26.error = t20;
  return t26;
 }
 t25 = t21.payload;
 t27 = t25;
 t29 = (*t3);
 t26 = mem_Allocator_alloc__anon_2542__4255(t29, t27);
 if (t26.error) {
  t20 = t26.error;
  t26.payload = (nav__4253_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t26.error = t20;
  return t26;
 }
 t30 = t26.payload;
 t28 = t30;
 t4 = (*t1);
 t30 = t28;
 t31 = t4;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *)&t31;
 t32 = t1;
 t8 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *const *)&t32;
 t10 = (void const **)&t33.context;
 t1 = (*t8);
 t11 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t1->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__4253_59 (**)(void const *, nav__4253_39))&t33.readFn;
 (*t13) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4207;
 t14 = t33;
 t34 = t14;
 t16 = (struct io_Reader__2409 const *)&t34;
 t14 = (*t16);
 t35 = io_Reader_readAll__4214(t14, t30);
 memcpy(&t36, &t35, sizeof(nav__4253_59));
 t37 = t36.payload;
 t25 = t37;
 t38 = t27;
 t39 = t25 != t38;
 if (t39) {
  bincode_invalidProtocol__anon_2564__4256();
  zig_unreachable();
 }
 goto zig_block_1;

 zig_block_1:;
 t30 = t28;
 t26.payload = t30;
 t26.error = UINT16_C(0);
 return t26;
}

static nav__4252_40 bincode_deserializeAlloc__anon_2507__4252(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4252_57 t0;
 nav__4252_40 t2;
 nav__4252_55 t3;
 nav__4252_39 t4;
 uint16_t t1;
 t0 = bincode_deserializePointerAlloc__anon_2520__4253(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (nav__4252_39){(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 memcpy(&t4, &t3, sizeof(nav__4252_39));
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__4251_39 bincode_deserializeStructAlloc__anon_2482__4251(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 uint32_t *t1;
 nav__4251_39 t4;
 nav__4251_44 *t6;
 nav__4251_60 t7;
 nav__4251_44 t8;
 struct shared_Command__struct_2041__2041 t9;
 struct shared_Command__struct_2041__2041 t0;
 nav__4251_57 t2;
 uint32_t t5;
 uint16_t t3;
 t1 = (uint32_t *)&t0.pid;
 t2 = bincode_deserializeAlloc__anon_2383__4210(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_2041__2041){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t6 = (nav__4251_44 *)&t0.uri;
 t7 = bincode_deserializeAlloc__anon_2507__4252(a0, a1);
 if (t7.error) {
  t3 = t7.error;
  t4.payload = (struct shared_Command__struct_2041__2041){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)};
  t4.error = t3;
  return t4;
 }
 t8 = t7.payload;
 (*t6) = t8;
 t9 = t0;
 t4.payload = t9;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__4250_39 bincode_deserializeAlloc__anon_2453__4250(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4250_39 t0;
 struct shared_Command__struct_2041__2041 t2;
 uint16_t t1;
 t0 = bincode_deserializeStructAlloc__anon_2482__4251(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__struct_2041__2041){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static uint16_t bincode_deserializeAlloc__anon_2579__4257(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 (void)a0;
 (void)a1;
 return 0;
}

static nav__4259_39 bincode_deserializeStructAlloc__anon_2643__4259(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4259_44 *t1;
 nav__4259_57 t2;
 nav__4259_39 t4;
 nav__4259_44 t5;
 struct shared_Command__struct_2042__2042 t6;
 struct shared_Command__struct_2042__2042 t0;
 uint16_t t3;
 t1 = (nav__4259_44 *)&t0.name;
 t2 = bincode_deserializeAlloc__anon_2507__4252(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_2042__2042){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t1 = (nav__4259_44 *)&t0.version;
 t2 = bincode_deserializeAlloc__anon_2507__4252(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_2042__2042){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t6 = t0;
 t4.payload = t6;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__4258_39 bincode_deserializeAlloc__anon_2624__4258(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4258_39 t0;
 struct shared_Command__struct_2042__2042 t2;
 uint16_t t1;
 t0 = bincode_deserializeStructAlloc__anon_2643__4259(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__struct_2042__2042){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4265_38 bincode_deserializeInt__anon_2766__4265(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t2;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *const *t5;
 void const **t7;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t8;
 void const *t9;
 nav__4265_50 (**t10)(void const *, nav__4265_52);
 struct io_Reader__2409 t11;
 struct io_Reader__2409 t6;
 struct io_Reader__2409 t12;
 struct io_Reader__2409 t15;
 struct io_Reader__2409 const *t13;
 nav__4265_38 t14;
 nav__4265_38 t18;
 uint8_t const (*t21)[8];
 uint64_t t22;
 nav__4265_64 t16;
 uint16_t t17;
 uint8_t t19[8];
 uint8_t t20[8];
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__4265_50 (**)(void const *, nav__4265_52))&t6.readFn;
 (*t10) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4207;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Reader__2409 const *)&t12;
 t11 = (*t13);
 t15 = t11;
 t13 = (struct io_Reader__2409 const *)&t15;
 t11 = (*t13);
 t16 = io_Reader_readBytesNoEof__anon_2533__4254(t11);
 if (t16.error) {
  t17 = t16.error;
  t18.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t18.error = t17;
  t14 = t18;
  goto zig_block_0;
 }
 memcpy(t19, t16.payload, sizeof(uint8_t[8]));
 memcpy((char *)&t20, t19, sizeof(uint8_t[8]));
 t21 = (uint8_t const (*)[8])&t20;
 memcpy(t19, (const char *)t21, sizeof(uint8_t[8]));
 memcpy(&t22, &t19, sizeof(uint64_t));
 t22 = zig_wrap_u64(t22, UINT8_C(64));
 t18.payload = t22;
 t18.error = UINT16_C(0);
 t14 = t18;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t18, &t14, sizeof(nav__4265_38));
 if (t18.error) {
  t17 = t18.error;
  t18.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t18.error = t17;
  return t18;
 }
 t22 = t18.payload;
 t18.payload = t22;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__4264_38 bincode_deserializeAlloc__anon_2761__4264(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4264_38 t0;
 uint64_t t2;
 uint16_t t1;
 (void)a1;
 t0 = bincode_deserializeInt__anon_2766__4265(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4263_39 bincode_deserializeUnionAlloc__anon_2748__4263(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4263_39 t2;
 nav__4263_56 t6;
 uint64_t t7;
 struct shared_MarkerType__2045 t8;
 nav__4263_54 t0;
 uint32_t t3;
 uint16_t t1;
 uint8_t t4;
 bool t5;
 t0 = bincode_deserializeAlloc__anon_2383__4210(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct shared_MarkerType__2045){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = (uint8_t)t3;
 t5 = t4 == UINT8_C(0);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2761__4264(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__2045){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(0);
  t8.payload.SampleStart = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = t4 == UINT8_C(1);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2761__4264(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__2045){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(1);
  t8.payload.SampleEnd = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_1;

 zig_block_1:;
 t5 = t4 == UINT8_C(2);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2761__4264(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__2045){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(2);
  t8.payload.BenchmarkStart = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_2;

 zig_block_2:;
 t5 = t4 == UINT8_C(3);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2761__4264(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__2045){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(3);
  t8.payload.BenchmarkEnd = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_3;

 zig_block_3:;
 zig_unreachable();
}

static nav__4262_39 bincode_deserializeAlloc__anon_2714__4262(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4262_39 t0;
 struct shared_MarkerType__2045 t2;
 uint16_t t1;
 t0 = bincode_deserializeUnionAlloc__anon_2748__4263(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_MarkerType__2045){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4261_39 bincode_deserializeStructAlloc__anon_2696__4261(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 uint32_t *t1;
 nav__4261_39 t4;
 struct shared_MarkerType__2045 *t6;
 nav__4261_60 t7;
 struct shared_MarkerType__2045 t8;
 struct shared_Command__struct_2043__2043 t9;
 struct shared_Command__struct_2043__2043 t0;
 nav__4261_57 t2;
 uint32_t t5;
 uint16_t t3;
 t1 = (uint32_t *)&t0.pid;
 t2 = bincode_deserializeAlloc__anon_2383__4210(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_2043__2043){{{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)},UINT32_C(0xaaaaaaaa)};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t6 = (struct shared_MarkerType__2045 *)&t0.marker;
 t7 = bincode_deserializeAlloc__anon_2714__4262(a0, a1);
 if (t7.error) {
  t3 = t7.error;
  t4.payload = (struct shared_Command__struct_2043__2043){{{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)},UINT32_C(0xaaaaaaaa)};
  t4.error = t3;
  return t4;
 }
 t8 = t7.payload;
 (*t6) = t8;
 t9 = t0;
 t4.payload = t9;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__4260_39 bincode_deserializeAlloc__anon_2682__4260(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4260_39 t0;
 struct shared_Command__struct_2043__2043 t2;
 uint16_t t1;
 t0 = bincode_deserializeStructAlloc__anon_2696__4261(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__struct_2043__2043){{{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)},UINT32_C(0xaaaaaaaa)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4267_38 bincode_deserializeEnum__anon_2852__4267(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0) {
 nav__4267_46 t0;
 uint32_t t3;
 uint16_t t1;
 nav__4267_38 t2;
 uint8_t t4;
 t0 = bincode_deserializeInt__anon_2394__4211(a0);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = UINT8_C(0x2);
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = (uint8_t)t3;
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__4266_38 bincode_deserializeAlloc__anon_2823__4266(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4266_38 t0;
 uint16_t t1;
 uint8_t t2;
 (void)a1;
 t0 = bincode_deserializeEnum__anon_2852__4267(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = UINT8_C(0x2);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4209_39 bincode_deserializeUnionAlloc__anon_2378__4209(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4209_39 t2;
 nav__4209_68 t6;
 struct shared_Command__struct_2041__2041 t7;
 struct shared_Command__2038 t8;
 nav__4209_70 t9;
 struct shared_Command__struct_2042__2042 t10;
 nav__4209_72 t11;
 struct shared_Command__struct_2043__2043 t12;
 nav__4209_74 t13;
 uint64_t t14;
 nav__4209_66 t0;
 uint32_t t3;
 uint16_t t1;
 nav__4209_76 t15;
 uint8_t t4;
 bool t5;
 t0 = bincode_deserializeAlloc__anon_2383__4210(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = (uint8_t)t3;
 t5 = t4 == UINT8_C(0);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2453__4250(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(0);
  t8.payload.ExecutedBenchmark = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = t4 == UINT8_C(1);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2579__4257(a0, a1);
  return (nav__4209_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(1)},0};
 }
 goto zig_block_1;

 zig_block_1:;
 t5 = t4 == UINT8_C(2);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2579__4257(a0, a1);
  return (nav__4209_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(2)},0};
 }
 goto zig_block_2;

 zig_block_2:;
 t5 = t4 == UINT8_C(3);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2579__4257(a0, a1);
  return (nav__4209_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(3)},0};
 }
 goto zig_block_3;

 zig_block_3:;
 t5 = t4 == UINT8_C(4);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2579__4257(a0, a1);
  return (nav__4209_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(4)},0};
 }
 goto zig_block_4;

 zig_block_4:;
 t5 = t4 == UINT8_C(5);
 if (t5) {
  t9 = bincode_deserializeAlloc__anon_2624__4258(a0, a1);
  if (t9.error) {
   t1 = t9.error;
   t2.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t10 = t9.payload;
  t8.tag = UINT8_C(5);
  t8.payload.SetIntegration = t10;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_5;

 zig_block_5:;
 t5 = t4 == UINT8_C(6);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2579__4257(a0, a1);
  return (nav__4209_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(6)},0};
 }
 goto zig_block_6;

 zig_block_6:;
 t5 = t4 == UINT8_C(7);
 if (t5) {
  t11 = bincode_deserializeAlloc__anon_2682__4260(a0, a1);
  if (t11.error) {
   t1 = t11.error;
   t2.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t12 = t11.payload;
  t8.tag = UINT8_C(7);
  t8.payload.AddMarker = t12;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_7;

 zig_block_7:;
 t5 = t4 == UINT8_C(8);
 if (t5) {
  t13 = bincode_deserializeAlloc__anon_2761__4264(a0, a1);
  if (t13.error) {
   t1 = t13.error;
   t2.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t14 = t13.payload;
  t8.tag = UINT8_C(8);
  t8.payload.SetVersion = t14;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_8;

 zig_block_8:;
 t5 = t4 == UINT8_C(9);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2579__4257(a0, a1);
  return (nav__4209_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(9)},0};
 }
 goto zig_block_9;

 zig_block_9:;
 t5 = t4 == UINT8_C(10);
 if (t5) {
  t15 = bincode_deserializeAlloc__anon_2823__4266(a0, a1);
  if (t15.error) {
   t1 = t15.error;
   t2.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t4 = t15.payload;
  t8.tag = UINT8_C(10);
  t8.payload.IntegrationModeResponse = t4;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_10;

 zig_block_10:;
 zig_unreachable();
}

static nav__4208_39 bincode_deserializeAlloc__anon_2271__4208(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 const a0, struct mem_Allocator__589 const a1) {
 nav__4208_39 t0;
 struct shared_Command__2038 t2;
 uint16_t t1;
 t0 = bincode_deserializeUnionAlloc__anon_2378__4209(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__1126_39 fifo_UnixPipe_Reader_recvCmd__1126(struct fifo_UnixPipe_Reader__630 *const a0) {
 struct fifo_UnixPipe_Reader__630 *const *t1;
 uintptr_t t3;
 uintptr_t t10;
 struct fifo_UnixPipe_Reader__630 *t4;
 struct fifo_UnixPipe_Reader__630 *t0;
 struct fs_File__636 *t5;
 nav__1126_68 t7;
 nav__1126_76 t8;
 nav__1126_39 t12;
 nav__1126_39 t13;
 uint64_t t14;
 uint64_t t20;
 uint8_t const (*t15)[4];
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t18;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t19;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t22;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t21;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t23;
 struct mem_Allocator__589 *t24;
 struct mem_Allocator__589 t25;
 struct shared_Command__2038 t26;
 struct fs_File__636 t6;
 uint32_t t17;
 uint16_t t11;
 bool t9;
 uint8_t t16[4];
 uint8_t t2[4];
 t0 = a0;
 t1 = (struct fifo_UnixPipe_Reader__630 *const *)&t0;
 t4 = (*t1);
 t5 = (struct fs_File__636 *)&t4->file;
 t6 = (*t5);
 t7.ptr = &t2[(uintptr_t)0ul];
 t7.len = (uintptr_t)4ul;
 t8 = fs_File_readAll__1264(t6, t7);
 t9 = t8.error == UINT16_C(0);
 if (t9) {
  t10 = t8.payload;
  t3 = t10;
  goto zig_block_0;
 }
 t11 = t8.error;
 switch (t11) {
  case zig_error_WouldBlock:
  case zig_error_BrokenPipe: {
   t12 = (nav__1126_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)},zig_error_NotReady};
   goto zig_block_1;
  }
  default: {
   t13.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
   t13.error = t11;
   t12 = t13;
   goto zig_block_1;
  }
 }

 zig_block_1:;
 return t12;

 zig_block_0:;
 t14 = t3;
 t9 = t14 < UINT64_C(4);
 if (t9) {
  return (nav__1126_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)},zig_error_UnexpectedEof};
 }
 goto zig_block_2;

 zig_block_2:;
 t15 = (uint8_t const (*)[4])&t2;
 memcpy(t16, (const char *)t15, sizeof(uint8_t[4]));
 memcpy(&t17, &t16, sizeof(uint32_t));
 t17 = zig_wrap_u32(t17, UINT8_C(32));
 t4 = (*t1);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&t4->buffer;
 t3 = (uintptr_t)t17;
 t11 = array_list_ArrayListAligned_28u8_2cnull_29_resize__1394(t18, t3);
 if (t11) {
  t13.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
  t13.error = t11;
  return t13;
 }
 t4 = (*t1);
 t5 = (struct fs_File__636 *)&t4->file;
 t6 = (*t5);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&a0->buffer;
 t19 = (*t18);
 t7 = t19.items;
 t8 = fs_File_readAll__1264(t6, t7);
 t9 = t8.error == UINT16_C(0);
 if (t9) {
  t10 = t8.payload;
  t3 = t10;
  goto zig_block_3;
 }
 t11 = t8.error;
 switch (t11) {
  case zig_error_WouldBlock:
  case zig_error_BrokenPipe: {
   t13 = (nav__1126_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)},zig_error_NotReady};
   goto zig_block_4;
  }
  default: {
   t12.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
   t12.error = t11;
   t13 = t12;
   goto zig_block_4;
  }
 }

 zig_block_4:;
 return t13;

 zig_block_3:;
 t14 = t3;
 t20 = (uint64_t)t17;
 t9 = t14 < t20;
 if (t9) {
  return (nav__1126_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)},zig_error_UnexpectedEof};
 }
 goto zig_block_5;

 zig_block_5:;
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&a0->buffer;
 t19 = (*t18);
 t7 = t19.items;
 t22 = io_fixed_buffer_stream_fixedBufferStream__anon_2242__4174(t7);
 t21 = t22;
 t23 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_reader__4163(&t21);
 t24 = (struct mem_Allocator__589 *)&a0->allocator;
 t25 = (*t24);
 t13 = bincode_deserializeAlloc__anon_2271__4208(t23, t25);
 if (t13.error) {
  t11 = t13.error;
  t13.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
  t13.error = t11;
  return t13;
 }
 t26 = t13.payload;
 t13.payload = t26;
 t13.error = UINT16_C(0);
 return t13;
}

static nav__1127_39 fifo_UnixPipe_Reader_waitForResponse__1127(struct fifo_UnixPipe_Reader__630 *const a0, nav__1127_42 const a1) {
 zig_i128 t2;
 zig_i128 t6;
 struct fifo_UnixPipe_Reader__630 *const *t1;
 uint64_t t3;
 uint64_t t5;
 struct shared_Command__2038 t7;
 struct shared_Command__2038 t10;
 struct fifo_UnixPipe_Reader__630 *t8;
 struct fifo_UnixPipe_Reader__630 *t0;
 nav__1127_39 t9;
 uint16_t t11;
 bool t4;
 t0 = a0;
 t1 = (struct fifo_UnixPipe_Reader__630 *const *)&t0;
 t2 = time_nanoTimestamp__4115();
 t4 = a1.is_null != true;
 if (t4) {
  t5 = a1.payload;
  t3 = t5;
  goto zig_block_0;
 }
 t3 = UINT64_C(1000000000);
 goto zig_block_0;

 zig_block_0:;
 zig_loop_13:
 t6 = time_nanoTimestamp__4115();
 t6 = zig_sub_i128(t6, t2);
 t5 = zig_lo_i128(t6);
 t4 = t5 > t3;
 if (t4) {
  return (nav__1127_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)},zig_error_AckTimeout};
 }
 goto zig_block_2;

 zig_block_2:;
 t8 = (*t1);
 t9 = fifo_UnixPipe_Reader_recvCmd__1126(t8);
 t4 = t9.error == UINT16_C(0);
 if (t4) {
  t10 = t9.payload;
  t7 = t10;
  goto zig_block_3;
 }
 t11 = t9.error;
 switch (t11) {
  case zig_error_NotReady:
  case zig_error_UnexpectedEof: {
   utils_sleep__3969(UINT64_C(10000000));
   goto zig_block_1;
  }
  default: {
   t9.payload = (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(0xa)};
   t9.error = t11;
   return t9;
  }
 }

 zig_block_3:;
 t9.payload = t7;
 t9.error = UINT16_C(0);
 return t9;

 zig_block_1:;
 goto zig_loop_13;
}

static uint16_t fifo_UnixPipe_Reader_waitForAck__1128(struct fifo_UnixPipe_Reader__630 *const a0, nav__1128_40 const a1) {
 struct fifo_UnixPipe_Reader__630 *const *t1;
 struct fifo_UnixPipe_Reader__630 *t2;
 struct fifo_UnixPipe_Reader__630 *t0;
 nav__1128_60 t3;
 struct shared_Command__2038 t5;
 struct shared_Command__2038 t6;
 struct shared_Command__2038 const *t7;
 struct mem_Allocator__589 *t9;
 struct mem_Allocator__589 t10;
 nav__1128_77 t11;
 uint16_t t4;
 uint8_t t8;
 t0 = a0;
 t1 = (struct fifo_UnixPipe_Reader__630 *const *)&t0;
 t2 = (*t1);
 t3 = fifo_UnixPipe_Reader_waitForResponse__1127(t2, a1);
 if (t3.error) {
  t4 = t3.error;
  return t4;
 }
 t5 = t3.payload;
 t6 = t5;
 t7 = (struct shared_Command__2038 const *)&t6;
 t8 = t5.tag;
 switch (t8) {
  case UINT8_C(3): {
   t5 = (*t7);
   t9 = (struct mem_Allocator__589 *)&a0->allocator;
   t10 = (*t9);
   shared_Command_deinit__3974(t5, t10);
   return 0;
  }
  case UINT8_C(6): {
   t5 = (*t7);
   t9 = (struct mem_Allocator__589 *)&a0->allocator;
   t10 = (*t9);
   shared_Command_deinit__3974(t5, t10);
   return zig_error_UnexpectedError;
  }
  default: {
   t11.f0 = t5;
   logger_debug__anon_2894__4277(t11);
   t5 = (*t7);
   t9 = (struct mem_Allocator__589 *)&a0->allocator;
   t10 = (*t9);
   shared_Command_deinit__3974(t5, t10);
   return zig_error_UnexpectedResponse;
  }
 }
}

static uint16_t runner_fifo_RunnerFifo_send_version__1092(struct runner_fifo_RunnerFifo__621 *const a0, uint64_t const a1) {
 struct runner_fifo_RunnerFifo__621 *const *t1;
 struct runner_fifo_RunnerFifo__621 *t2;
 struct runner_fifo_RunnerFifo__621 *t0;
 struct fifo_UnixPipe_Writer__628 *t3;
 struct shared_Command__2038 t4;
 struct fifo_UnixPipe_Reader__630 *t6;
 uint16_t t5;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__621 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_UnixPipe_Writer__628 *)&t2->writer;
 t4.tag = UINT8_C(8);
 t4.payload.SetVersion = a1;
 t5 = fifo_UnixPipe_Writer_sendCmd__1121(t3, t4);
 if (t5) {
  return t5;
 }
 t2 = (*t1);
 t6 = (struct fifo_UnixPipe_Reader__630 *)&t2->reader;
 t5 = fifo_UnixPipe_Reader_waitForAck__1128(t6, (nav__1092_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t runner_fifo_RunnerFifo_validate_protocol_version__1083(struct runner_fifo_RunnerFifo__621 *const a0) {
 struct runner_fifo_RunnerFifo__621 *const *t1;
 struct runner_fifo_RunnerFifo__621 *t2;
 struct runner_fifo_RunnerFifo__621 *t0;
 nav__1083_43 t5;
 nav__1083_45 t6;
 uint16_t t3;
 bool t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__621 *const *)&t0;
 t2 = (*t1);
 t3 = runner_fifo_RunnerFifo_send_version__1092(t2, UINT64_C(2));
 t4 = t3 == UINT16_C(0);
 if (t4) {
  goto zig_block_0;
 }
 switch (t3) {
  case zig_error_AckTimeout: {
   return 0;
  }
  case zig_error_UnexpectedError: {
   logger_err__anon_2917__4278();
   logger_err__anon_2925__4279();
   posix_exit__1582(UINT8_C(1));
   zig_unreachable();
  }
  default: {
   t5 = zig_errorName[t3 - 1];
   t6.f0 = t5;
   logger_err__anon_2938__4280(t6);
   posix_exit__1582(UINT8_C(1));
   zig_unreachable();
  }
 }

 zig_block_0:;
 return 0;
}

static nav__1093_38 runner_fifo_RunnerFifo_get_integration_mode__1093(struct runner_fifo_RunnerFifo__621 *const a0) {
 struct runner_fifo_RunnerFifo__621 *const *t1;
 struct runner_fifo_RunnerFifo__621 *t2;
 struct runner_fifo_RunnerFifo__621 *t0;
 struct fifo_UnixPipe_Writer__628 *t3;
 struct fifo_UnixPipe_Reader__630 *t6;
 nav__1093_81 t7;
 struct shared_Command__2038 t8;
 struct shared_Command__2038 t9;
 struct shared_Command__2038 const *t10;
 struct mem_Allocator__589 *t13;
 struct mem_Allocator__589 t14;
 uint16_t t4;
 nav__1093_38 t5;
 uint8_t t11;
 bool t12;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__621 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_UnixPipe_Writer__628 *)&t2->writer;
 t4 = fifo_UnixPipe_Writer_sendCmd__1121(t3, (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(9)});
 if (t4) {
  t5.payload = UINT8_C(0x2);
  t5.error = t4;
  return t5;
 }
 t2 = (*t1);
 t6 = (struct fifo_UnixPipe_Reader__630 *)&t2->reader;
 t7 = fifo_UnixPipe_Reader_waitForResponse__1127(t6, (nav__1093_79){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t7.error) {
  t4 = t7.error;
  t5.payload = UINT8_C(0x2);
  t5.error = t4;
  return t5;
 }
 t8 = t7.payload;
 t9 = t8;
 t10 = (struct shared_Command__2038 const *)&t9;
 t11 = t8.tag;
 t12 = t11 == UINT8_C(10);
 if (t12) {
  t11 = t8.payload.IntegrationModeResponse;
  t8 = (*t10);
  t13 = (struct mem_Allocator__589 *)&a0->allocator;
  t14 = (*t13);
  shared_Command_deinit__3974(t8, t14);
  t5.payload = t11;
  t5.error = UINT16_C(0);
  return t5;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = (*t10);
 t13 = (struct mem_Allocator__589 *)&a0->allocator;
 t14 = (*t13);
 shared_Command_deinit__3974(t8, t14);
 return (nav__1093_38){zig_error_UnexpectedResponse,UINT8_C(0x2)};
}

static nav__773_39 instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_init__773(struct mem_Allocator__589 const a0) {
 nav__773_62 t1;
 nav__773_39 t3;
 struct runner_fifo_RunnerFifo__621 t4;
 struct runner_fifo_RunnerFifo__621 t0;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 t9;
 uint16_t t2;
 nav__773_65 t6;
 uint8_t t5;
 uint8_t t8;
 bool t7;
 t1 = runner_fifo_RunnerFifo_init__1082(a0);
 if (t1.error) {
  t2 = t1.error;
  t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
  t3.error = t2;
  return t3;
 }
 t4 = t1.payload;
 t0 = t4;
 (void)runner_fifo_RunnerFifo_validate_protocol_version__1083(&t0);
 t6 = runner_fifo_RunnerFifo_get_integration_mode__1093(&t0);
 t7 = t6.error == UINT16_C(0);
 if (t7) {
  t8 = t6.payload;
  t5 = t8;
  goto zig_block_0;
 }
 t2 = t6.error;
 runner_fifo_RunnerFifo_deinit__1084(&t0);
 t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
 t3.error = t2;
 return t3;

 zig_block_0:;
 t7 = t5 != UINT8_C(2);
 if (t7) {
  runner_fifo_RunnerFifo_deinit__1084(&t0);
  return (nav__773_39){{{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}},zig_error_ModeError};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t0;
 t9.fifo = t4;
 t3.payload = t9;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__760_39 instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_init__760(struct mem_Allocator__589 const a0) {
 nav__760_62 t1;
 nav__760_39 t3;
 struct runner_fifo_RunnerFifo__621 t4;
 struct runner_fifo_RunnerFifo__621 t0;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 t9;
 uint16_t t2;
 nav__760_65 t6;
 uint8_t t5;
 uint8_t t8;
 bool t7;
 t1 = runner_fifo_RunnerFifo_init__1082(a0);
 if (t1.error) {
  t2 = t1.error;
  t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
  t3.error = t2;
  return t3;
 }
 t4 = t1.payload;
 t0 = t4;
 (void)runner_fifo_RunnerFifo_validate_protocol_version__1083(&t0);
 t6 = runner_fifo_RunnerFifo_get_integration_mode__1093(&t0);
 t7 = t6.error == UINT16_C(0);
 if (t7) {
  t8 = t6.payload;
  t5 = t8;
  goto zig_block_0;
 }
 t2 = t6.error;
 runner_fifo_RunnerFifo_deinit__1084(&t0);
 t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
 t3.error = t2;
 return t3;

 zig_block_0:;
 t7 = t5 != UINT8_C(0);
 if (t7) {
  runner_fifo_RunnerFifo_deinit__1084(&t0);
  return (nav__760_39){{{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}},zig_error_ModeError};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t0;
 t9.fifo = t4;
 t3.payload = t9;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__726_39 instruments_root_InstrumentHooks_init__726(struct mem_Allocator__589 const a0) {
 nav__726_69 t0;
 struct instruments_valgrind_ValgrindInstrument__554 t2;
 struct instruments_root_InstrumentHooks__547 t3;
 nav__726_39 t4;
 nav__726_71 t5;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 t6;
 nav__726_73 t7;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 t8;
 bool t1;
 t0 = instruments_valgrind_ValgrindInstrument_init__739(a0);
 t1 = t0.error == UINT16_C(0);
 if (t1) {
  t2 = t0.payload;
  t3.tag = UINT8_C(0);
  t3.payload.valgrind = t2;
  t4.payload = t3;
  t4.error = UINT16_C(0);
  return t4;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_init__773(a0);
 t1 = t5.error == UINT16_C(0);
 if (t1) {
  t6 = t5.payload;
  t3.tag = UINT8_C(2);
  t3.payload.analysis = t6;
  t4.payload = t3;
  t4.error = UINT16_C(0);
  return t4;
 }
 goto zig_block_1;

 zig_block_1:;
 t7 = instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_init__760(a0);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t8 = t7.payload;
  t3.tag = UINT8_C(1);
  t3.payload.perf = t8;
  t4.payload = t3;
  t4.error = UINT16_C(0);
  return t4;
 }
 goto zig_block_2;

 zig_block_2:;
 return (nav__726_39){{{{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}}},UINT8_C(3)},0};
}

struct instruments_root_InstrumentHooks__547 *c_instrument_hooks_init__237(void) {
 struct instruments_root_InstrumentHooks__547 *t0;
 struct instruments_root_InstrumentHooks__547 *t3;
 nav__237_46 t1;
 nav__237_78 t4;
 struct instruments_root_InstrumentHooks__547 t5;
 bool t2;
 t1 = mem_Allocator_create__anon_901__3717((struct mem_Allocator__589){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)&heap_CAllocator_vtable__3705)});
 t2 = t1.error == UINT16_C(0);
 if (t2) {
  t3 = t1.payload;
  t0 = t3;
  goto zig_block_0;
 }
 return NULL;

 zig_block_0:;
 t4 = instruments_root_InstrumentHooks_init__726((struct mem_Allocator__589){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)&heap_CAllocator_vtable__3705)});
 t5 = t4.payload;
 (*t0) = t5;
 t3 = (struct instruments_root_InstrumentHooks__547 *)t0;
 return t3;
}

static nav__3717_40 mem_Allocator_create__anon_901__3717(struct mem_Allocator__589 const a0) {
 struct mem_Allocator__589 const *t1;
 struct mem_Allocator__589 t2;
 struct mem_Allocator__589 t0;
 uintptr_t t3;
 nav__3717_51 t4;
 nav__3717_40 t6;
 uint8_t *t7;
 struct instruments_root_InstrumentHooks__547 *t8;
 uint16_t t5;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t4 = mem_Allocator_allocBytesWithAlignment__anon_2980__4281(t2, (uintptr_t)152ul, t3);
 if (t4.error) {
  t5 = t4.error;
  t6.payload = ((struct instruments_root_InstrumentHooks__547 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul);
  t6.error = t5;
  return t6;
 }
 t7 = t4.payload;
 t8 = (struct instruments_root_InstrumentHooks__547 *)t7;
 t6.payload = t8;
 t6.error = UINT16_C(0);
 return t6;
}

static uint16_t fs_accessAbsolute__1183(nav__1183_39 const a0, struct fs_File_OpenFlags__1954 const a1) {
 struct fs_Dir__1956 const *t3;
 struct fs_Dir__1956 t1;
 struct fs_Dir__1956 t2;
 uint16_t t4;
 bool t0;
 t0 = fs_path_isAbsolute__4316(a0);
 debug_assert__180(t0);
 t1 = fs_cwd__1175();
 t2 = t1;
 t3 = (struct fs_Dir__1956 const *)&t2;
 t1 = (*t3);
 t4 = fs_Dir_access__3918(t1, a0, a1);
 if (t4) {
  return t4;
 }
 return 0;
}

static nav__1180_39 fs_openFileAbsolute__1180(nav__1180_41 const a0, struct fs_File_OpenFlags__1954 const a1) {
 struct fs_Dir__1956 const *t3;
 struct fs_Dir__1956 t1;
 struct fs_Dir__1956 t2;
 nav__1180_39 t4;
 bool t0;
 t0 = fs_path_isAbsolute__4316(a0);
 debug_assert__180(t0);
 t1 = fs_cwd__1175();
 t2 = t1;
 t3 = (struct fs_Dir__1956 const *)&t2;
 t1 = (*t3);
 t4 = fs_Dir_openFile__3854(t1, a0, a1);
 return t4;
}

static void utils_setNonBlocking__3970(int32_t const a0) {
 int t0;
 int t1;
 t0 = a0;
 t0 = fcntl(t0, 3, 0);
 t0 = t0 | 2048;
 t1 = a0;
 (void)fcntl(t1, 4, t0);
 return;
}

static struct fifo_UnixPipe_Writer__628 fifo_UnixPipe_Writer_init__1118(struct fs_File__636 const a0, struct mem_Allocator__589 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t0;
 struct fs_File__636 *t5;
 struct mem_Allocator__589 *t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t7;
 struct fifo_UnixPipe_Writer__628 t4;
 uint16_t t2;
 bool t3;
 t1 = array_list_ArrayListAligned_28u8_2cnull_29_init__1362(a1);
 t0 = t1;
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1399(&t0, (uintptr_t)1024ul);
 t3 = t2 == UINT16_C(0);
 if (t3) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (struct fs_File__636 *)&t4.file;
 (*t5) = a0;
 t6 = (struct mem_Allocator__589 *)&t4.allocator;
 (*t6) = a1;
 t7 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&t4.buffer;
 t1 = t0;
 (*t7) = t1;
 return t4;
}

static struct fifo_UnixPipe_Reader__630 fifo_UnixPipe_Reader_init__1123(struct fs_File__636 const a0, struct mem_Allocator__589 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t0;
 struct fs_File__636 *t5;
 struct mem_Allocator__589 *t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t7;
 struct fifo_UnixPipe_Reader__630 t4;
 uint16_t t2;
 bool t3;
 t1 = array_list_ArrayListAligned_28u8_2cnull_29_init__1362(a1);
 t0 = t1;
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1399(&t0, (uintptr_t)1024ul);
 t3 = t2 == UINT16_C(0);
 if (t3) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (struct fs_File__636 *)&t4.file;
 (*t5) = a0;
 t6 = (struct mem_Allocator__589 *)&t4.allocator;
 (*t6) = a1;
 t7 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&t4.buffer;
 t1 = t0;
 (*t7) = t1;
 return t4;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_clearRetainingCapacity__1397(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t0;
 nav__1397_43 *t3;
 uintptr_t *t4;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t0;
 t2 = (*t1);
 t3 = (nav__1397_43 *)&t2->items;
 t4 = &t3->len;
 (*t4) = (uintptr_t)0ul;
 return;
}

static struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 array_list_ArrayListAligned_28u8_2cnull_29_writer__1387(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 **t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static uint16_t bincode_serialize__anon_2090__4101(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_Command__2038 const a1) {
 uint16_t t0;
 t0 = bincode_serializeUnion__anon_3043__5311(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint8_t const (*mem_asBytes__anon_2128__4102(uint32_t const *const a0))[4] {
 uint8_t const (*t0)[4];
 t0 = (uint8_t const (*)[4])a0;
 return t0;
}

static uint16_t fs_File_writeAll__1274(struct fs_File__636 const a0, nav__1274_40 const a1) {
 struct fs_File__636 const *t1;
 nav__1274_40 const *t3;
 uintptr_t t5;
 uintptr_t t6;
 uintptr_t t13;
 uintptr_t t4;
 uint64_t t7;
 uint64_t t8;
 nav__1274_40 t11;
 nav__1274_40 t2;
 uint8_t const *t12;
 nav__1274_47 t14;
 struct fs_File__636 t10;
 struct fs_File__636 t0;
 uint16_t t15;
 bool t9;
 t0 = a0;
 t1 = (struct fs_File__636 const *)&t0;
 t2 = a1;
 t3 = (nav__1274_40 const *)&t2;
 t4 = (uintptr_t)0ul;
 zig_loop_11:
 t5 = t4;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 < t8;
 if (t9) {
  t6 = t4;
  t10 = (*t1);
  t5 = t4;
  t11 = (*t3);
  t12 = t11.ptr;
  t12 = (uint8_t const *)(((uintptr_t)t12) + (t5*sizeof(uint8_t)));
  t13 = t11.len;
  t5 = t13 - t5;
  t11.ptr = t12;
  t11.len = t5;
  t14 = fs_File_write__1273(t10, t11);
  if (t14.error) {
   t15 = t14.error;
   return t15;
  }
  t5 = t14.payload;
  t5 = t6 + t5;
  t4 = t5;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_11;

 zig_block_0:;
 return 0;
}

static zig_i128 time_nanoTimestamp__4115(void) {
 zig_i128 t6;
 zig_i128 t7;
 nav__4115_41 t0;
 struct os_linux_timespec__struct_3100__3100 t2;
 struct os_linux_timespec__struct_3100__3100 t3;
 intptr_t t5;
 uint16_t t4;
 bool t1;
 t0 = posix_clock_gettime__1773(UINT32_C(0));
 t1 = t0.error == UINT16_C(0);
 if (t1) {
  t3 = t0.payload;
  t2 = t3;
  goto zig_block_0;
 }
 t4 = t0.error;
 switch (t4) {
  case zig_error_UnsupportedClock:
  case zig_error_Unexpected: {
   return zig_make_i128(INT64_C(0), UINT64_C(0));
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t5 = t2.sec;
 t6 = zig_make_i128(0, t5);
 t6 = zig_mul_i128(t6, zig_make_i128(INT64_C(0), UINT64_C(1000000000)));
 t5 = t2.nsec;
 t7 = zig_make_i128(0, t5);
 t7 = zig_add_i128(t6, t7);
 return t7;
}

static nav__1264_38 fs_File_readAll__1264(struct fs_File__636 const a0, nav__1264_41 const a1) {
 struct fs_File__636 const *t1;
 nav__1264_41 const *t3;
 uintptr_t t5;
 uintptr_t t6;
 uintptr_t t4;
 uint64_t t7;
 uint64_t t8;
 nav__1264_41 t11;
 nav__1264_41 t2;
 uint8_t *t12;
 nav__1264_38 t13;
 struct fs_File__636 t10;
 struct fs_File__636 t0;
 uint16_t t14;
 bool t9;
 t0 = a0;
 t1 = (struct fs_File__636 const *)&t0;
 t2 = a1;
 t3 = (nav__1264_41 const *)&t2;
 t4 = (uintptr_t)0ul;
 zig_loop_11:
 t5 = t4;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 != t8;
 if (t9) {
  t10 = (*t1);
  t6 = t4;
  t11 = (*t3);
  t12 = t11.ptr;
  t12 = (uint8_t *)(((uintptr_t)t12) + (t6*sizeof(uint8_t)));
  t5 = t11.len;
  t6 = t5 - t6;
  t11.ptr = t12;
  t11.len = t6;
  t13 = fs_File_read__1263(t10, t11);
  if (t13.error) {
   t14 = t13.error;
   t13.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t13.error = t14;
   return t13;
  }
  t6 = t13.payload;
  t8 = t6;
  t9 = t8 == UINT64_C(0);
  if (t9) {
   goto zig_block_0;
  }
  goto zig_block_2;

  zig_block_2:;
  t5 = t4;
  t6 = t5 + t6;
  t4 = t6;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_11;

 zig_block_0:;
 t6 = t4;
 t13.payload = t6;
 t13.error = UINT16_C(0);
 return t13;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_resize__1394(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t0;
 nav__1394_43 *t4;
 uintptr_t *t5;
 uint16_t t3;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t0;
 t2 = (*t1);
 t3 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1399(t2, a1);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t4 = (nav__1394_43 *)&t2->items;
 t5 = &t4->len;
 (*t5) = a1;
 return 0;
}

static struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 io_fixed_buffer_stream_fixedBufferStream__anon_2242__4174(nav__4174_40 const a0) {
 nav__4174_40 *t1;
 uintptr_t *t2;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t0;
 t1 = (nav__4174_40 *)&t0.buffer;
 (*t1) = a0;
 t2 = (uintptr_t *)&t0.pos;
 (*t2) = (uintptr_t)0ul;
 return t0;
}

static struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_reader__4163(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 **t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2263 t0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static nav__4207_38 io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4207(void const *const a0, nav__4207_41 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t0;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *t1;
 nav__4207_38 t2;
 nav__4207_38 t3;
 t0 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)a0;
 t1 = (*t0);
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_read__4166(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__4207_38));
 return t3;
}

static nav__4249_39 io_Reader_readBytesNoEof__anon_2440__4249(struct io_Reader__2409 const a0) {
 struct io_Reader__2409 const *t1;
 struct io_Reader__2409 t3;
 struct io_Reader__2409 t0;
 nav__4249_46 t4;
 uint16_t t5;
 nav__4249_39 t6;
 uint8_t t7[4];
 uint8_t t2[4];
 t0 = a0;
 t1 = (struct io_Reader__2409 const *)&t0;
 t3 = (*t1);
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)4ul;
 t5 = io_Reader_readNoEof__4216(t3, t4);
 if (t5) {
  memcpy(t6.payload, "\252\252\252\252", sizeof(uint8_t[4]));
  t6.error = t5;
  return t6;
 }
 memcpy(t7, (const char *)&t2, sizeof(uint8_t[4]));
 memcpy(t6.payload, t7, sizeof(uint8_t[4]));
 t6.error = UINT16_C(0);
 return t6;
}

static nav__4254_39 io_Reader_readBytesNoEof__anon_2533__4254(struct io_Reader__2409 const a0) {
 struct io_Reader__2409 const *t1;
 struct io_Reader__2409 t3;
 struct io_Reader__2409 t0;
 nav__4254_46 t4;
 uint16_t t5;
 nav__4254_39 t6;
 uint8_t t7[8];
 uint8_t t2[8];
 t0 = a0;
 t1 = (struct io_Reader__2409 const *)&t0;
 t3 = (*t1);
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)8ul;
 t5 = io_Reader_readNoEof__4216(t3, t4);
 if (t5) {
  memcpy(t6.payload, "\252\252\252\252\252\252\252\252", sizeof(uint8_t[8]));
  t6.error = t5;
  return t6;
 }
 memcpy(t7, (const char *)&t2, sizeof(uint8_t[8]));
 memcpy(t6.payload, t7, sizeof(uint8_t[8]));
 t6.error = UINT16_C(0);
 return t6;
}

static nav__4255_40 mem_Allocator_alloc__anon_2542__4255(struct mem_Allocator__589 const a0, uintptr_t const a1) {
 struct mem_Allocator__589 const *t1;
 struct mem_Allocator__589 t2;
 struct mem_Allocator__589 t0;
 struct mem_Allocator__589 t5;
 uintptr_t t3;
 nav__4255_40 t4;
 nav__4255_40 t8;
 nav__4255_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__4255_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__589 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_3137__5312(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__4255_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static nav__4214_38 io_Reader_readAll__4214(struct io_Reader__2409 const a0, nav__4214_41 const a1) {
 uintptr_t t0;
 nav__4214_38 t1;
 t0 = a1.len;
 t1 = io_Reader_readAtLeast__4215(a0, a1, t0);
 return t1;
}

static zig_noreturn void bincode_invalidProtocol__anon_2564__4256(void) {
 utils_print__anon_3157__5313();
 posix_exit__1582(UINT8_C(1));
 zig_unreachable();
}

static void utils_sleep__3969(uint64_t const a0) {
 uint64_t t0;
 uint64_t t1;
 long *t3;
 long t4;
 long t7;
 nav__3969_42 t5;
 struct cimport_struct_timespec__3171 const *t9;
 struct cimport_struct_timespec__3171 *t10;
 struct cimport_struct_timespec__3171 t13;
 struct cimport_struct_timespec__3171 t2;
 struct cimport_struct_timespec__3171 t8;
 int t11;
 int32_t t12;
 bool t6;
 t0 = a0 / UINT64_C(1000000000);
 t1 = a0 % UINT64_C(1000000000);
 t3 = (long *)&t2.tv_sec;
 t5 = math_cast__anon_3184__6247(t0);
 t6 = t5.is_null != true;
 if (t6) {
  t7 = t5.payload;
  t4 = t7;
  goto zig_block_0;
 }
 t4 = LONG_MAX;
 goto zig_block_0;

 zig_block_0:;
 (*t3) = t4;
 t3 = (long *)&t2.tv_nsec;
 t5 = math_cast__anon_3184__6247(t1);
 t6 = t5.is_null != true;
 if (t6) {
  t7 = t5.payload;
  t4 = t7;
  goto zig_block_1;
 }
 t4 = LONG_MAX;
 goto zig_block_1;

 zig_block_1:;
 (*t3) = t4;
 zig_loop_28:
 t9 = (struct cimport_struct_timespec__3171 const *)&t2;
 t10 = (struct cimport_struct_timespec__3171 *)&t8;
 t11 = nanosleep(t9, t10);
 t12 = t11;
 t6 = t12 == INT32_C(4);
 if (t6) {
  t13 = t8;
  t2 = t13;
  goto zig_block_2;
 }
 return;

 zig_block_2:;
 goto zig_loop_28;
}

static void shared_Command_deinit__3974(struct shared_Command__2038 const a0, struct mem_Allocator__589 const a1) {
 struct mem_Allocator__589 const *t1;
 struct shared_Command__struct_2042__2042 t3;
 struct mem_Allocator__589 t4;
 struct mem_Allocator__589 t0;
 nav__3974_43 t5;
 struct shared_Command__struct_2041__2041 t6;
 uint8_t t2;
 t0 = a1;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(5): {
   t3 = a0.payload.SetIntegration;
   t4 = (*t1);
   t5 = t3.name;
   mem_Allocator_free__anon_3261__7028(t4, t5);
   t4 = (*t1);
   t5 = t3.version;
   mem_Allocator_free__anon_3261__7028(t4, t5);
   goto zig_block_0;
  }
  case UINT8_C(0): {
   t6 = a0.payload.ExecutedBenchmark;
   t4 = (*t1);
   t5 = t6.uri;
   mem_Allocator_free__anon_3261__7028(t4, t5);
   goto zig_block_0;
  }
  case UINT8_C(8): {
   goto zig_block_0;
  }
  case UINT8_C(9): {
   goto zig_block_0;
  }
  case UINT8_C(10): {
   goto zig_block_0;
  }
  default: {
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return;
}

static void logger_debug__anon_2894__4277(nav__4277_39 const a0) {
 logger_logWithPrefix__anon_3269__7029(a0);
 return;
}

static void logger_err__anon_2917__4278(void) {
 logger_logWithPrefix__anon_3271__7030();
 return;
}

static void logger_err__anon_2925__4279(void) {
 logger_logWithPrefix__anon_3272__7031();
 return;
}

static zig_noreturn void posix_exit__1582(uint8_t const a0) {
 int t0;
 t0 = (int)a0;
 exit(t0);
 zig_unreachable();
}

static void logger_err__anon_2938__4280(nav__4280_40 const a0) {
 logger_logWithPrefix__anon_3277__7033(a0);
 return;
}

static void runner_fifo_RunnerFifo_deinit__1084(struct runner_fifo_RunnerFifo__621 *const a0) {
 struct runner_fifo_RunnerFifo__621 *const *t1;
 struct runner_fifo_RunnerFifo__621 *t2;
 struct runner_fifo_RunnerFifo__621 *t0;
 struct fifo_UnixPipe_Writer__628 *t3;
 struct fifo_UnixPipe_Reader__630 *t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__621 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_UnixPipe_Writer__628 *)&t2->writer;
 fifo_UnixPipe_Writer_deinit__1122(t3);
 t2 = (*t1);
 t4 = (struct fifo_UnixPipe_Reader__630 *)&t2->reader;
 fifo_UnixPipe_Reader_deinit__1129(t4);
 return;
}

static nav__4281_39 mem_Allocator_allocBytesWithAlignment__anon_2980__4281(struct mem_Allocator__589 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__589 const *t1;
 uint64_t t2;
 uint8_t *t4;
 uint8_t *t13;
 uint8_t *t14;
 uint8_t *t15;
 struct mem_Allocator__589 t5;
 struct mem_Allocator__589 t0;
 struct mem_Allocator__589 t7;
 struct mem_Allocator_VTable__592 const *const *t8;
 struct mem_Allocator_VTable__592 const *t9;
 uint8_t *(*const *t10)(void *, uintptr_t, uint8_t, uintptr_t);
 uint8_t *(*t11)(void *, uintptr_t, uint8_t, uintptr_t);
 void *t12;
 uint8_t *const *t16;
 nav__4281_52 t17;
 nav__4281_39 t18;
 bool t3;
 uint8_t t6;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = a1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  return (nav__4281_39){(uint8_t *)(uintptr_t)0xfffffffffffffff8ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (*t1);
 t6 = mem_Alignment_fromByteUnits__1066((uintptr_t)8ul);
 t7 = t5;
 t1 = (struct mem_Allocator__589 const *)&t7;
 t8 = (struct mem_Allocator_VTable__592 const *const *)&t1->vtable;
 t9 = (*t8);
 t10 = (uint8_t *(*const *)(void *, uintptr_t, uint8_t, uintptr_t))&t9->alloc;
 t11 = (*t10);
 t12 = t5.ptr;
 t13 = t11(t12, a1, t6, a2);
 t3 = t13 != NULL;
 if (t3) {
  t14 = t13;
  t4 = t14;
  goto zig_block_1;
 }
 return (nav__4281_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_1:;
 t15 = t4;
 t16 = (uint8_t *const *)&t15;
 t14 = (*t16);
 t14 = (uint8_t *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t17.ptr = t14;
 t17.len = a1;
 t14 = (uint8_t *)t4;
 t18.payload = t14;
 t18.error = UINT16_C(0);
 return t18;
}

static bool fs_path_isAbsolute__4316(nav__4316_39 const a0) {
 bool t0;
 t0 = fs_path_isAbsolutePosix__4322(a0);
 return t0;
}

static struct fs_Dir__1956 fs_cwd__1175(void) {
 struct fs_Dir__1956 t0;
 t0 = (struct fs_Dir__1956){-INT32_C(100)};
 return t0;
}

static uint16_t fs_Dir_access__3918(struct fs_Dir__1956 const a0, nav__3918_40 const a1, struct fs_File_OpenFlags__1954 const a2) {
 struct fs_Dir__1956 const *t1;
 uint8_t const (*t6)[4096];
 uint8_t const *t8;
 struct fs_Dir__1956 t7;
 struct fs_Dir__1956 t0;
 nav__3918_49 t2;
 uint16_t t3;
 uint8_t t4[4096];
 uint8_t t5[4096];
 t0 = a0;
 t1 = (struct fs_Dir__1956 const *)&t0;
 t2 = posix_toPosixPath__1857(a1);
 if (t2.error) {
  t3 = t2.error;
  return t3;
 }
 memcpy(t4, t2.payload, sizeof(uint8_t[4096]));
 memcpy((char *)&t5, t4, sizeof(uint8_t[4096]));
 t6 = (uint8_t const (*)[4096])&t5;
 t7 = (*t1);
 t8 = (uint8_t const *)t6;
 t3 = fs_Dir_accessZ__3919(t7, t8, a2);
 return t3;
}

static nav__3854_39 fs_Dir_openFile__3854(struct fs_Dir__1956 const a0, nav__3854_42 const a1, struct fs_File_OpenFlags__1954 const a2) {
 struct fs_Dir__1956 const *t1;
 uint8_t const (*t8)[4096];
 uint8_t const *t10;
 nav__3854_39 t5;
 struct fs_Dir__1956 t9;
 struct fs_Dir__1956 t0;
 nav__3854_53 t3;
 uint16_t t4;
 uint8_t t6[4096];
 uint8_t t7[4096];
 struct fs_File_OpenFlags__1954 t2;
 t0 = a0;
 t1 = (struct fs_Dir__1956 const *)&t0;
 t2 = a2;
 t3 = posix_toPosixPath__1857(a1);
 if (t3.error) {
  t4 = t3.error;
  t5.payload = (struct fs_File__636){-INT32_C(0x55555556)};
  t5.error = t4;
  return t5;
 }
 memcpy(t6, t3.payload, sizeof(uint8_t[4096]));
 memcpy((char *)&t7, t6, sizeof(uint8_t[4096]));
 t8 = (uint8_t const (*)[4096])&t7;
 t9 = (*t1);
 t10 = (uint8_t const *)t8;
 t5 = fs_Dir_openFileZ__3855(t9, t10, a2);
 return t5;
}

static struct array_list_ArrayListAligned_28u8_2cnull_29__648 array_list_ArrayListAligned_28u8_2cnull_29_init__1362(struct mem_Allocator__589 const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t0;
 t0.items = (nav__1362_42){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 t0.capacity = (uintptr_t)0ul;
 t0.allocator = a0;
 return t0;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1399(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t0;
 uint16_t t8;
 bool t6;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a1;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = array_list_ArrayListAlignedUnmanaged_28u8_2cnull_29_growCapacity__7104(t3, a1);
 t7 = (*t1);
 t8 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacityPrecise__1400(t7, t3);
 return t8;
}

static uint16_t bincode_serializeUnion__anon_3043__5311(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_Command__2038 const a1) {
 struct shared_Command__struct_2041__2041 t4;
 struct shared_Command__struct_2042__2042 t5;
 struct shared_Command__struct_2043__2043 t6;
 uint64_t t7;
 uint32_t t1;
 uint16_t t2;
 uint8_t t0;
 bool t3;
 t0 = a1.tag;
 t1 = (uint32_t)t0;
 t2 = bincode_serialize__anon_3381__7105(a0, t1);
 if (t2) {
  return t2;
 }
 t0 = a1.tag;
 t3 = t0 == UINT8_C(0);
 if (t3) {
  t4 = a1.payload.ExecutedBenchmark;
  t2 = bincode_serialize__anon_3383__7106(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(1);
 if (t3) {
  t2 = bincode_serialize__anon_3385__7107(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(2);
 if (t3) {
  t2 = bincode_serialize__anon_3385__7107(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(3);
 if (t3) {
  t2 = bincode_serialize__anon_3385__7107(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(4);
 if (t3) {
  t2 = bincode_serialize__anon_3385__7107(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_4;
 }
 goto zig_block_4;

 zig_block_4:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(5);
 if (t3) {
  t5 = a1.payload.SetIntegration;
  t2 = bincode_serialize__anon_3387__7108(a0, t5);
  if (t2) {
   return t2;
  }
  goto zig_block_5;
 }
 goto zig_block_5;

 zig_block_5:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(6);
 if (t3) {
  t2 = bincode_serialize__anon_3385__7107(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_6;
 }
 goto zig_block_6;

 zig_block_6:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(7);
 if (t3) {
  t6 = a1.payload.AddMarker;
  t2 = bincode_serialize__anon_3389__7109(a0, t6);
  if (t2) {
   return t2;
  }
  goto zig_block_7;
 }
 goto zig_block_7;

 zig_block_7:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(8);
 if (t3) {
  t7 = a1.payload.SetVersion;
  t2 = bincode_serialize__anon_3391__7110(a0, t7);
  if (t2) {
   return t2;
  }
  goto zig_block_8;
 }
 goto zig_block_8;

 zig_block_8:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(9);
 if (t3) {
  t2 = bincode_serialize__anon_3385__7107(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_9;
 }
 goto zig_block_9;

 zig_block_9:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(10);
 if (t3) {
  t0 = a1.payload.IntegrationModeResponse;
  t2 = bincode_serialize__anon_3393__7111(a0, t0);
  if (t2) {
   return t2;
  }
  goto zig_block_10;
 }
 goto zig_block_10;

 zig_block_10:;
 return 0;
}

static nav__1273_38 fs_File_write__1273(struct fs_File__636 const a0, nav__1273_41 const a1) {
 nav__1273_38 t1;
 int32_t t0;
 t0 = a0.handle;
 t1 = posix_write__1592(t0, a1);
 return t1;
}

static nav__1773_39 posix_clock_gettime__1773(uint32_t const a0) {
 struct os_linux_timespec__struct_3100__3100 t3;
 struct os_linux_timespec__struct_3100__3100 t0;
 nav__1773_39 t4;
 int t1;
 uint16_t t2;
 t1 = clock_gettime(a0, &t0);
 t2 = posix_errno__anon_3559__7178(t1);
 switch (t2) {
  case UINT16_C(0): {
   t3 = t0;
   t4.payload = t3;
   t4.error = UINT16_C(0);
   return t4;
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(22): {
   return (nav__1773_39){{-(intptr_t)0x5555555555555556,-(intptr_t)0x5555555555555556},zig_error_UnsupportedClock};
  }
  default: {
   t2 = posix_unexpectedErrno__1856(t2);
   t4.payload = (struct os_linux_timespec__struct_3100__3100){-(intptr_t)0x5555555555555556,-(intptr_t)0x5555555555555556};
   t4.error = t2;
   return t4;
  }
 }
}

static nav__1263_38 fs_File_read__1263(struct fs_File__636 const a0, nav__1263_41 const a1) {
 nav__1263_38 t1;
 int32_t t0;
 t0 = a0.handle;
 t1 = posix_read__1584(t0, a1);
 return t1;
}

static nav__4166_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_read__4166(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const a0, nav__4166_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t1;
 nav__4166_42 const *t3;
 uintptr_t t4;
 uintptr_t t7;
 uintptr_t t9;
 nav__4166_42 *t5;
 nav__4166_42 t6;
 nav__4166_42 t13;
 nav__4166_42 t2;
 uintptr_t *t8;
 uint64_t t10;
 uint8_t *t11;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *t12;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *t0;
 nav__4166_38 t14;
 t0 = a0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t0;
 t2 = a1;
 t3 = (nav__4166_42 const *)&t2;
 t4 = a1.len;
 t5 = (nav__4166_42 *)&a0->buffer;
 t6 = (*t5);
 t7 = t6.len;
 t8 = (uintptr_t *)&a0->pos;
 t9 = (*t8);
 t9 = t7 - t9;
 t9 = (t4 < t9) ? t4 : t9;
 t10 = t9;
 t8 = (uintptr_t *)&a0->pos;
 t9 = (*t8);
 t4 = t10;
 t4 = t9 + t4;
 t6 = (*t3);
 t11 = t6.ptr;
 t11 = (uint8_t *)(((uintptr_t)t11) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t9 = t10;
 t6.ptr = t11;
 t6.len = t9;
 t12 = (*t1);
 t5 = (nav__4166_42 *)&t12->buffer;
 t8 = (uintptr_t *)&a0->pos;
 t9 = (*t8);
 t13 = (*t5);
 t11 = t13.ptr;
 t11 = (uint8_t *)(((uintptr_t)t11) + (t9*sizeof(uint8_t)));
 t9 = t4 - t9;
 t13.ptr = t11;
 t13.len = t9;
 t11 = t13.ptr;
 if (t6.len != 0) memcpy(t6.ptr, t11, t6.len * sizeof(uint8_t));
 t12 = (*t1);
 t8 = (uintptr_t *)&t12->pos;
 (*t8) = t4;
 t4 = t10;
 t14.payload = t4;
 t14.error = UINT16_C(0);
 return t14;
}

static uint16_t io_Reader_readNoEof__4216(struct io_Reader__2409 const a0, nav__4216_40 const a1) {
 struct io_Reader__2409 const *t1;
 struct io_Reader__2409 t2;
 struct io_Reader__2409 t0;
 nav__4216_43 t3;
 uintptr_t t5;
 uintptr_t t6;
 uint64_t t7;
 uint64_t t8;
 uint16_t t4;
 bool t9;
 t0 = a0;
 t1 = (struct io_Reader__2409 const *)&t0;
 t2 = (*t1);
 t3 = io_Reader_readAll__4214(t2, a1);
 if (t3.error) {
  t4 = t3.error;
  return t4;
 }
 t5 = t3.payload;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 < t8;
 if (t9) {
  return zig_error_EndOfStream;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static nav__5312_39 mem_Allocator_allocWithSizeAndAlignment__anon_3137__5312(struct mem_Allocator__589 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__589 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__5312_49 t3;
 struct mem_Allocator__589 t6;
 struct mem_Allocator__589 t0;
 nav__5312_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t3 = math_mul__anon_3583__7179((uintptr_t)1ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__5312_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_3585__7180(t6, t2, a2);
 return t7;
}

static nav__4215_38 io_Reader_readAtLeast__4215(struct io_Reader__2409 const a0, nav__4215_41 const a1, uintptr_t const a2) {
 struct io_Reader__2409 const *t1;
 nav__4215_41 const *t3;
 uintptr_t t4;
 uintptr_t t12;
 uintptr_t t8;
 uint64_t t5;
 uint64_t t6;
 struct io_Reader__2409 t9;
 struct io_Reader__2409 t0;
 nav__4215_41 t10;
 nav__4215_41 t2;
 uint8_t *t11;
 nav__4215_38 t13;
 uint16_t t14;
 bool t7;
 t0 = a0;
 t1 = (struct io_Reader__2409 const *)&t0;
 t2 = a1;
 t3 = (nav__4215_41 const *)&t2;
 t4 = a1.len;
 t5 = a2;
 t6 = t4;
 t7 = t5 <= t6;
 debug_assert__180(t7);
 t8 = (uintptr_t)0ul;
 zig_loop_17:
 t4 = t8;
 t6 = t4;
 t5 = a2;
 t7 = t6 < t5;
 if (t7) {
  t9 = (*t1);
  t4 = t8;
  t10 = (*t3);
  t11 = t10.ptr;
  t11 = (uint8_t *)(((uintptr_t)t11) + (t4*sizeof(uint8_t)));
  t12 = t10.len;
  t4 = t12 - t4;
  t10.ptr = t11;
  t10.len = t4;
  t13 = io_Reader_read__4213(t9, t10);
  if (t13.error) {
   t14 = t13.error;
   t13.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t13.error = t14;
   return t13;
  }
  t4 = t13.payload;
  t5 = t4;
  t7 = t5 == UINT64_C(0);
  if (t7) {
   goto zig_block_0;
  }
  goto zig_block_2;

  zig_block_2:;
  t12 = t8;
  t4 = t12 + t4;
  t8 = t4;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_17;

 zig_block_0:;
 t12 = t8;
 t13.payload = t12;
 t13.error = UINT16_C(0);
 return t13;
}

static void utils_print__anon_3157__5313(void) {
 (void)printf((char const *)&__anon_3681, 61, (uint8_t const *)&__anon_2558);
 return;
}

static nav__6247_38 math_cast__anon_3184__6247(uint64_t const a0) {
 long t1;
 nav__6247_38 t2;
 bool t0;
 t0 = a0 > UINT64_C(9223372036854775807);
 if (t0) {
  return (nav__6247_38){-0x5555555555555556l,true};
 }
 t1 = (long)a0;
 t2.is_null = false;
 t2.payload = t1;
 return t2;
}

static void mem_Allocator_free__anon_3261__7028(struct mem_Allocator__589 const a0, nav__7028_40 const a1) {
 struct mem_Allocator__589 const *t1;
 nav__7028_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t const *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 nav__7028_51 t10;
 struct mem_Allocator__589 t11;
 struct mem_Allocator__589 t0;
 struct mem_Allocator__589 t13;
 struct mem_Allocator_VTable__592 const *const *t14;
 struct mem_Allocator_VTable__592 const *t15;
 void (*const *t16)(void *, nav__7028_51, uint8_t, uintptr_t);
 void (*t17)(void *, nav__7028_51, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = mem_sliceAsBytes__anon_3715__7182(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 t11 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 t12 = mem_Alignment_fromByteUnits__1066((uintptr_t)1ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t11;
 t1 = (struct mem_Allocator__589 const *)&t13;
 t14 = (struct mem_Allocator_VTable__592 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__7028_51, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t11.ptr;
 t17(t18, t10, t12, t3);
 return;
}

static void logger_logWithPrefix__anon_3269__7029(nav__7029_39 const a0) {
 nav__7029_59 t3;
 nav__7029_59 t4;
 nav__7029_62 t5;
 uint8_t *t6;
 char const *t7;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4271);
 t1 = UINT8_C(0) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)512ul;
 t5 = fmt_bufPrint__anon_3764__7342(t4, a0);
 t1 = t5.error == UINT16_C(0);
 if (t1) {
  t4 = t5.payload;
  t3 = t4;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3768);
 return;

 zig_block_1:;
 t6 = t3.ptr;
 t7 = (char const *)t6;
 (void)printf(t7);
 return;
}

static void logger_logWithPrefix__anon_3271__7030(void) {
 nav__7030_41 t3;
 nav__7030_41 t4;
 nav__7030_44 t5;
 uint8_t *t6;
 char const *t7;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4271);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)512ul;
 t5 = fmt_bufPrint__anon_3785__7344(t4);
 t1 = t5.error == UINT16_C(0);
 if (t1) {
  t4 = t5.payload;
  t3 = t4;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3768);
 return;

 zig_block_1:;
 t6 = t3.ptr;
 t7 = (char const *)t6;
 (void)printf(t7);
 return;
}

static void logger_logWithPrefix__anon_3272__7031(void) {
 nav__7031_41 t3;
 nav__7031_41 t4;
 nav__7031_44 t5;
 uint8_t *t6;
 char const *t7;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4271);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)512ul;
 t5 = fmt_bufPrint__anon_3795__7345(t4);
 t1 = t5.error == UINT16_C(0);
 if (t1) {
  t4 = t5.payload;
  t3 = t4;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3768);
 return;

 zig_block_1:;
 t6 = t3.ptr;
 t7 = (char const *)t6;
 (void)printf(t7);
 return;
}

static void logger_logWithPrefix__anon_3277__7033(nav__7033_40 const a0) {
 nav__7033_47 t3;
 nav__7033_47 t4;
 nav__7033_50 t5;
 uint8_t *t6;
 char const *t7;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4271);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)512ul;
 t5 = fmt_bufPrint__anon_3807__7346(t4, a0);
 t1 = t5.error == UINT16_C(0);
 if (t1) {
  t4 = t5.payload;
  t3 = t4;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3768);
 return;

 zig_block_1:;
 t6 = t3.ptr;
 t7 = (char const *)t6;
 (void)printf(t7);
 return;
}

static void fifo_UnixPipe_Writer_deinit__1122(struct fifo_UnixPipe_Writer__628 *const a0) {
 struct fifo_UnixPipe_Writer__628 *const *t1;
 struct fifo_UnixPipe_Writer__628 *t2;
 struct fifo_UnixPipe_Writer__628 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t3;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t4;
 struct fs_File__636 *t5;
 struct fs_File__636 t6;
 t0 = a0;
 t1 = (struct fifo_UnixPipe_Writer__628 *const *)&t0;
 t2 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&t2->buffer;
 t4 = (*t3);
 array_list_ArrayListAligned_28u8_2cnull_29_deinit__1364(t4);
 t2 = (*t1);
 t5 = (struct fs_File__636 *)&t2->file;
 t6 = (*t5);
 fs_File_close__1220(t6);
 return;
}

static void fifo_UnixPipe_Reader_deinit__1129(struct fifo_UnixPipe_Reader__630 *const a0) {
 struct fifo_UnixPipe_Reader__630 *const *t1;
 uintptr_t t3;
 uintptr_t t10;
 struct fifo_UnixPipe_Reader__630 *t4;
 struct fifo_UnixPipe_Reader__630 *t0;
 struct fs_File__636 *t5;
 nav__1129_50 t7;
 nav__1129_58 t8;
 uint64_t t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t12;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t13;
 struct fs_File__636 t6;
 bool t9;
 uint8_t t2[4096];
 t0 = a0;
 t1 = (struct fifo_UnixPipe_Reader__630 *const *)&t0;
 zig_loop_7:
 t4 = (*t1);
 t5 = (struct fs_File__636 *)&t4->file;
 t6 = (*t5);
 t7.ptr = &t2[(uintptr_t)0ul];
 t7.len = (uintptr_t)4096ul;
 t8 = fs_File_read__1263(t6, t7);
 t9 = t8.error == UINT16_C(0);
 if (t9) {
  t10 = t8.payload;
  t3 = t10;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t11 = t3;
 t9 = t11 == UINT64_C(0);
 if (t9) {
  goto zig_block_0;
 }
 goto zig_block_2;

 zig_block_2:;
 goto zig_loop_7;

 zig_block_0:;
 t4 = (*t1);
 t12 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *)&t4->buffer;
 t13 = (*t12);
 array_list_ArrayListAligned_28u8_2cnull_29_deinit__1364(t13);
 t4 = (*t1);
 t5 = (struct fs_File__636 *)&t4->file;
 t6 = (*t5);
 fs_File_close__1220(t6);
 return;
}

static uint8_t mem_Alignment_fromByteUnits__1066(uintptr_t const a0) {
 bool t0;
 uint8_t t1;
 t0 = math_isPowerOfTwo__anon_3820__7347(a0);
 debug_assert__180(t0);
 t1 = zig_ctz_u64(a0, UINT8_C(64));
 return t1;
}

static bool fs_path_isAbsolutePosix__4322(nav__4322_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 bool t2;
 bool t3;
 uint8_t t4;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 > UINT64_C(0);
 if (t2) {
  t4 = a0.ptr[(uintptr_t)0ul];
  t2 = t4 == UINT8_C(47);
  t3 = t2;
  goto zig_block_0;
 }
 t3 = false;
 goto zig_block_0;

 zig_block_0:;
 return t3;
}

static nav__1857_39 posix_toPosixPath__1857(nav__1857_41 const a0) {
 uintptr_t t1;
 uint64_t t2;
 uint8_t *t4;
 nav__1857_47 t5;
 uint8_t const *t6;
 nav__1857_39 t8;
 bool t3;
 uint8_t t7[4096];
 uint8_t t0[4096];
 t1 = a0.len;
 t2 = t1;
 t3 = t2 >= UINT64_C(4096);
 if (t3) {
  return (nav__1857_39){zig_error_NameTooLong,{'\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa'}};
 }
 goto zig_block_0;

 zig_block_0:;
 t1 = a0.len;
 t4 = (uint8_t *)&t0;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5.ptr = t4;
 t5.len = t1;
 t6 = a0.ptr;
 if (t5.len != 0) memcpy(t5.ptr, t6, t5.len * sizeof(uint8_t));
 t1 = a0.len;
 t4 = (uint8_t *)&t0[t1];
 (*t4) = UINT8_C(0);
 memcpy(t7, (const char *)&t0, sizeof(uint8_t[4096]));
 memcpy(t8.payload, t7, sizeof(uint8_t[4096]));
 t8.error = UINT16_C(0);
 return t8;
}

static uint16_t fs_Dir_accessZ__3919(struct fs_Dir__1956 const a0, uint8_t const *const a1, struct fs_File_OpenFlags__1954 const a2) {
 uint32_t t2;
 int32_t t3;
 struct fs_Dir__1956 t0;
 uint16_t t4;
 uint8_t t1;
 t0 = a0;
 t1 = a2.mode;
 switch (t1) {
  case UINT8_C(0): {
   t2 = UINT32_C(0);
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t2 = UINT32_C(2);
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t2 = UINT32_C(6);
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t3 = a0.fd;
 t4 = posix_faccessatZ__1748(t3, a1, t2, UINT32_C(0));
 return t4;
}

static nav__3855_39 fs_Dir_openFileZ__3855(struct fs_Dir__1956 const a0, uint8_t const *const a1, struct fs_File_OpenFlags__1954 const a2) {
 uint32_t *t2;
 struct fs_File__636 *t14;
 int32_t *t15;
 int32_t t6;
 int32_t t11;
 int32_t t12;
 uint32_t t7;
 uint32_t t1;
 nav__3855_50 t8;
 nav__3855_39 t10;
 nav__3855_39 t13;
 struct fs_Dir__1956 t0;
 uint16_t t9;
 uint8_t t3;
 uint8_t t4;
 bool t5;
 t0 = a0;
 t2 = (uint32_t *)&t1;
 t3 = a2.mode;
 switch (t3) {
  case UINT8_C(0): {
   t4 = UINT8_C(0);
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t4 = UINT8_C(1);
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t4 = UINT8_C(2);
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffc)), zig_shl_u32((uint32_t)t4, UINT8_C(0)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffc3)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(2)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffbf)), zig_shl_u32((uint32_t)false, UINT8_C(6)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffff7f)), zig_shl_u32((uint32_t)false, UINT8_C(7)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffeff)), zig_shl_u32((uint32_t)false, UINT8_C(8)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffdff)), zig_shl_u32((uint32_t)false, UINT8_C(9)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffbff)), zig_shl_u32((uint32_t)false, UINT8_C(10)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffff7ff)), zig_shl_u32((uint32_t)false, UINT8_C(11)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffefff)), zig_shl_u32((uint32_t)false, UINT8_C(12)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffdfff)), zig_shl_u32((uint32_t)false, UINT8_C(13)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffbfff)), zig_shl_u32((uint32_t)false, UINT8_C(14)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffff7fff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(15)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffeffff)), zig_shl_u32((uint32_t)false, UINT8_C(16)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffdffff)), zig_shl_u32((uint32_t)false, UINT8_C(17)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffbffff)), zig_shl_u32((uint32_t)false, UINT8_C(18)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfff7ffff)), zig_shl_u32((uint32_t)false, UINT8_C(19)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffefffff)), zig_shl_u32((uint32_t)false, UINT8_C(20)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffdfffff)), zig_shl_u32((uint32_t)false, UINT8_C(21)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffbfffff)), zig_shl_u32((uint32_t)false, UINT8_C(22)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0x7fffff)), zig_shl_u32((uint32_t)UINT16_C(0), UINT8_C(23)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfff7ffff)), zig_shl_u32((uint32_t)true, UINT8_C(19)));
 t2 = (uint32_t *)&t1;
 t5 = a2.allow_ctty;
 t5 = !t5;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffeff)), zig_shl_u32((uint32_t)t5, UINT8_C(8)));
 t6 = a0.fd;
 t7 = t1;
 t8 = posix_openatZ__1604(t6, a1, t7, (uintptr_t)0ul);
 if (t8.error) {
  t9 = t8.error;
  t10.payload = (struct fs_File__636){-INT32_C(0x55555556)};
  t10.error = t9;
  return t10;
 }
 t6 = t8.payload;
 t4 = a2.lock;
 t5 = t4 != UINT8_C(0);
 if (t5) {
  t5 = a2.lock_nonblocking;
  if (t5) {
   t11 = INT32_C(4);
   goto zig_block_2;
  }
  t11 = INT32_C(0);
  goto zig_block_2;

  zig_block_2:;
  t4 = a2.lock;
  switch (t4) {
   case UINT8_C(0): {
    zig_unreachable();
   }
   case UINT8_C(1): {
    t11 = INT32_C(1) | t11;
    t12 = t11;
    goto zig_block_3;
   }
   case UINT8_C(2): {
    t11 = INT32_C(2) | t11;
    t12 = t11;
    goto zig_block_3;
   }
   default: zig_unreachable();
  }

  zig_block_3:;
  t9 = posix_flock__1765(t6, t12);
  if (t9) {
   posix_close__1562(t6);
   t10.payload = (struct fs_File__636){-INT32_C(0x55555556)};
   t10.error = t9;
   return t10;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t13.error = UINT16_C(0);
 t14 = &t13.payload;
 t15 = (int32_t *)&t14->handle;
 (*t15) = t6;
 return t13;
}

static uintptr_t array_list_ArrayListAlignedUnmanaged_28u8_2cnull_29_growCapacity__7104(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)128ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacityPrecise__1400(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t8;
 nav__1400_43 t9;
 nav__1400_43 t12;
 nav__1400_43 t20;
 nav__1400_43 t21;
 nav__1400_43 t18;
 struct mem_Allocator__589 *t10;
 struct mem_Allocator__589 t11;
 nav__1400_43 *t13;
 uint8_t **t14;
 uint8_t *t15;
 nav__1400_55 t16;
 nav__1400_43 const *t19;
 uint16_t t17;
 bool t6;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a1;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t8 = (*t7);
 t9 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1410(t8);
 t7 = (*t1);
 t10 = (struct mem_Allocator__589 *)&t7->allocator;
 t11 = (*t10);
 t12 = mem_Allocator_remap__anon_3973__7434(t11, t9, a1);
 t6 = t12.ptr != NULL;
 if (t6) {
  t9 = t12;
  t7 = (*t1);
  t13 = (nav__1400_43 *)&t7->items;
  t14 = &t13->ptr;
  t15 = t9.ptr;
  (*t14) = t15;
  t7 = (*t1);
  t2 = (uintptr_t *)&t7->capacity;
  t3 = t9.len;
  (*t2) = t3;
  goto zig_block_1;
 }
 t7 = (*t1);
 t10 = (struct mem_Allocator__589 *)&t7->allocator;
 t11 = (*t10);
 t16 = mem_Allocator_alignedAlloc__anon_3978__7435(t11, a1);
 if (t16.error) {
  t17 = t16.error;
  return t17;
 }
 t12 = t16.payload;
 t18 = t12;
 t19 = (nav__1400_43 const *)&t18;
 t13 = (nav__1400_43 *)&a0->items;
 t20 = (*t13);
 t3 = t20.len;
 t20 = (*t19);
 t15 = t20.ptr;
 t15 = (uint8_t *)(((uintptr_t)t15) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t20.ptr = t15;
 t20.len = t3;
 t13 = (nav__1400_43 *)&a0->items;
 t21 = (*t13);
 t15 = t21.ptr;
 if (t20.len != 0) memcpy(t20.ptr, t15, t20.len * sizeof(uint8_t));
 t7 = (*t1);
 t10 = (struct mem_Allocator__589 *)&t7->allocator;
 t11 = (*t10);
 mem_Allocator_free__anon_3980__7436(t11, t9);
 t7 = (*t1);
 t13 = (nav__1400_43 *)&t7->items;
 t14 = &t13->ptr;
 t15 = t12.ptr;
 (*t14) = t15;
 t7 = (*t1);
 t2 = (uintptr_t *)&t7->capacity;
 t3 = t12.len;
 (*t2) = t3;
 goto zig_block_1;

 zig_block_1:;
 return 0;
}

static uint16_t bincode_serialize__anon_3381__7105(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, uint32_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeInt__anon_3985__7437(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3383__7106(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_Command__struct_2041__2041 const a1) {
 uint16_t t0;
 t0 = bincode_serializeStruct__anon_3990__7438(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3385__7107(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0) {
 (void)a0;
 return 0;
}

static uint16_t bincode_serialize__anon_3387__7108(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_Command__struct_2042__2042 const a1) {
 uint16_t t0;
 t0 = bincode_serializeStruct__anon_3991__7439(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3389__7109(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_Command__struct_2043__2043 const a1) {
 uint16_t t0;
 t0 = bincode_serializeStruct__anon_3992__7440(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3391__7110(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, uint64_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeInt__anon_3993__7441(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3393__7111(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, uint8_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeEnum__anon_3996__7442(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static nav__1592_38 posix_write__1592(int32_t const a0, nav__1592_40 const a1) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t const *t3;
 intptr_t t5;
 nav__1592_38 t7;
 uint32_t t4;
 uint16_t t6;
 bool t2;
 t0 = a1.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__1592_38){(uintptr_t)0ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 zig_loop_16:
 t3 = a1.ptr;
 t0 = a1.len;
 t0 = ((uintptr_t)2147479552ul < t0) ? (uintptr_t)2147479552ul : t0;
 t4 = (uint32_t)t0;
 t0 = (uintptr_t)t4;
 t5 = write(a0, t3, t0);
 t6 = posix_errno__anon_4007__7444(t5);
 switch (t6) {
  case UINT16_C(0): {
   t0 = (uintptr_t)t5;
   t7.payload = t0;
   t7.error = UINT16_C(0);
   return t7;
  }
  case UINT16_C(4): {
   goto zig_block_1;
  }
  case UINT16_C(22): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_InvalidArgument};
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(2): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ProcessNotFound};
  }
  case UINT16_C(11): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_WouldBlock};
  }
  case UINT16_C(9): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NotOpenForWriting};
  }
  case UINT16_C(89): {
   zig_unreachable();
  }
  case UINT16_C(122): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_DiskQuota};
  }
  case UINT16_C(27): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_FileTooBig};
  }
  case UINT16_C(5): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_InputOutput};
  }
  case UINT16_C(28): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoSpaceLeft};
  }
  case UINT16_C(13): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_AccessDenied};
  }
  case UINT16_C(1): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_AccessDenied};
  }
  case UINT16_C(32): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_BrokenPipe};
  }
  case UINT16_C(104): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ConnectionResetByPeer};
  }
  case UINT16_C(16): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_DeviceBusy};
  }
  case UINT16_C(6): {
   return (nav__1592_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoDevice};
  }
  default: {
   t6 = posix_unexpectedErrno__1856(t6);
   t7.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t7.error = t6;
   return t7;
  }
 }

 zig_block_1:;
 goto zig_loop_16;
}

static uint16_t posix_errno__anon_3559__7178(int const a0) {
 int *t3;
 int32_t t1;
 int t4;
 uint16_t t0;
 uint16_t t5;
 bool t2;
 t1 = a0;
 t2 = t1 == -INT32_C(1);
 if (t2) {
  t3 = zig_e___errno_location();
  t4 = (*t3);
  t5 = (uint16_t)t4;
  t0 = t5;
  goto zig_block_0;
 }
 t0 = UINT16_C(0);
 goto zig_block_0;

 zig_block_0:;
 return t0;
}

static uint16_t posix_unexpectedErrno__1856(uint16_t const a0) {
 (void)a0;
 return zig_error_Unexpected;
}

static nav__1584_38 posix_read__1584(int32_t const a0, nav__1584_40 const a1) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3;
 intptr_t t5;
 nav__1584_38 t7;
 uint32_t t4;
 uint16_t t6;
 bool t2;
 t0 = a1.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__1584_38){(uintptr_t)0ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 zig_loop_16:
 t3 = a1.ptr;
 t0 = a1.len;
 t0 = ((uintptr_t)2147479552ul < t0) ? (uintptr_t)2147479552ul : t0;
 t4 = (uint32_t)t0;
 t0 = (uintptr_t)t4;
 t5 = read(a0, t3, t0);
 t6 = posix_errno__anon_4007__7444(t5);
 switch (t6) {
  case UINT16_C(0): {
   t0 = (uintptr_t)t5;
   t7.payload = t0;
   t7.error = UINT16_C(0);
   return t7;
  }
  case UINT16_C(4): {
   goto zig_block_1;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(2): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ProcessNotFound};
  }
  case UINT16_C(11): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_WouldBlock};
  }
  case UINT16_C(125): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_Canceled};
  }
  case UINT16_C(9): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NotOpenForReading};
  }
  case UINT16_C(5): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_InputOutput};
  }
  case UINT16_C(21): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_IsDir};
  }
  case UINT16_C(105): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};
  }
  case UINT16_C(12): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};
  }
  case UINT16_C(107): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SocketNotConnected};
  }
  case UINT16_C(104): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ConnectionResetByPeer};
  }
  case UINT16_C(110): {
   return (nav__1584_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ConnectionTimedOut};
  }
  default: {
   t6 = posix_unexpectedErrno__1856(t6);
   t7.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t7.error = t6;
   return t7;
  }
 }

 zig_block_1:;
 goto zig_loop_16;
}

static nav__7179_38 math_mul__anon_3583__7179(uintptr_t const a0, uintptr_t const a1) {
 nav__7179_42 t0;
 uintptr_t t3;
 nav__7179_38 t4;
 uint8_t t1;
 bool t2;
 t0.f1 = zig_mulo_u64(&t0.f0, a0, a1, UINT8_C(64));
 t1 = t0.f1;
 t2 = t1 != UINT8_C(0);
 if (t2) {
  return (nav__7179_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_Overflow};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t0.f0;
 t4.payload = t3;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__7180_39 mem_Allocator_allocBytesWithAlignment__anon_3585__7180(struct mem_Allocator__589 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__589 const *t1;
 uint64_t t2;
 uint8_t *t4;
 uint8_t *t13;
 uint8_t *t14;
 uint8_t *t15;
 struct mem_Allocator__589 t5;
 struct mem_Allocator__589 t0;
 struct mem_Allocator__589 t7;
 struct mem_Allocator_VTable__592 const *const *t8;
 struct mem_Allocator_VTable__592 const *t9;
 uint8_t *(*const *t10)(void *, uintptr_t, uint8_t, uintptr_t);
 uint8_t *(*t11)(void *, uintptr_t, uint8_t, uintptr_t);
 void *t12;
 uint8_t *const *t16;
 nav__7180_52 t17;
 nav__7180_39 t18;
 bool t3;
 uint8_t t6;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = a1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  return (nav__7180_39){(uint8_t *)UINTPTR_MAX,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (*t1);
 t6 = mem_Alignment_fromByteUnits__1066((uintptr_t)1ul);
 t7 = t5;
 t1 = (struct mem_Allocator__589 const *)&t7;
 t8 = (struct mem_Allocator_VTable__592 const *const *)&t1->vtable;
 t9 = (*t8);
 t10 = (uint8_t *(*const *)(void *, uintptr_t, uint8_t, uintptr_t))&t9->alloc;
 t11 = (*t10);
 t12 = t5.ptr;
 t13 = t11(t12, a1, t6, a2);
 t3 = t13 != NULL;
 if (t3) {
  t14 = t13;
  t4 = t14;
  goto zig_block_1;
 }
 return (nav__7180_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_1:;
 t15 = t4;
 t16 = (uint8_t *const *)&t15;
 t14 = (*t16);
 t14 = (uint8_t *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t17.ptr = t14;
 t17.len = a1;
 t18.payload = t4;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__4213_38 io_Reader_read__4213(struct io_Reader__2409 const a0, nav__4213_41 const a1) {
 struct io_Reader__2409 const *t1;
 nav__4213_38 (*const *t2)(void const *, nav__4213_41);
 nav__4213_38 (*t3)(void const *, nav__4213_41);
 void const *t4;
 nav__4213_38 t5;
 struct io_Reader__2409 t0;
 t0 = a0;
 t1 = (struct io_Reader__2409 const *)&t0;
 t2 = (nav__4213_38 (*const *)(void const *, nav__4213_41))&t1->readFn;
 t3 = (*t2);
 t4 = a0.context;
 t5 = t3(t4, a1);
 return t5;
}

static nav__7182_39 mem_sliceAsBytes__anon_3715__7182(nav__7182_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t const *t4;
 uint8_t const *t5;
 uint8_t const *const *t6;
 nav__7182_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__7182_39){(uint8_t const *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t const *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t const *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uint16_t fmt_format__anon_4185__7476(struct io_Writer__4171 const a0, nav__7476_40 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 struct shared_Command__2038 t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7476_45){(uint8_t const *)&__anon_4279,(uintptr_t)49ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_4653__7843(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7476_45){(uint8_t const *)&__anon_4659,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7342_40 fmt_bufPrint__anon_3764__7342(nav__7342_39 const a0, nav__7342_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t9;
 void const *t10;
 nav__7342_71 (**t11)(void const *, nav__7342_48);
 struct io_Writer__4171 t12;
 struct io_Writer__4171 t7;
 nav__7342_39 t15;
 nav__7342_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2242__4174(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4164(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7342_71 (**)(void const *, nav__7342_48))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__7459;
 t12 = t7;
 t13 = fmt_format__anon_4185__7476(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7342_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4172(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_4674__7844(struct io_Writer__4171 const a0) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7844_43){(uint8_t const *)&__anon_4761,(uintptr_t)68ul});
 if (t3) {
  return t3;
 }
 t3 = fmt_formatType__anon_4769__7845(UINT64_C(2), (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7844_43){(uint8_t const *)&__anon_4659,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7344_40 fmt_bufPrint__anon_3785__7344(nav__7344_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t9;
 void const *t10;
 nav__7344_54 (**t11)(void const *, nav__7344_56);
 struct io_Writer__4171 t12;
 struct io_Writer__4171 t7;
 nav__7344_39 t15;
 nav__7344_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2242__4174(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4164(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7344_54 (**)(void const *, nav__7344_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__7459;
 t12 = t7;
 t13 = fmt_format__anon_4674__7844(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7344_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4172(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_4777__7846(struct io_Writer__4171 const a0) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7846_43){(uint8_t const *)&__anon_4872,(uintptr_t)82ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7345_40 fmt_bufPrint__anon_3795__7345(nav__7345_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t9;
 void const *t10;
 nav__7345_54 (**t11)(void const *, nav__7345_56);
 struct io_Writer__4171 t12;
 struct io_Writer__4171 t7;
 nav__7345_39 t15;
 nav__7345_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2242__4174(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4164(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7345_54 (**)(void const *, nav__7345_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__7459;
 t12 = t7;
 t13 = fmt_format__anon_4777__7846(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7345_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4172(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_4878__7847(struct io_Writer__4171 const a0, nav__7847_41 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 nav__7847_40 t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7847_40){(uint8_t const *)&__anon_4926,(uintptr_t)32ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5030__7848(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7847_40){(uint8_t const *)&__anon_5059,(uintptr_t)22ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7346_40 fmt_bufPrint__anon_3807__7346(nav__7346_39 const a0, nav__7346_43 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t9;
 void const *t10;
 nav__7346_59 (**t11)(void const *, nav__7346_42);
 struct io_Writer__4171 t12;
 struct io_Writer__4171 t7;
 nav__7346_39 t15;
 nav__7346_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2242__4174(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4164(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7346_59 (**)(void const *, nav__7346_42))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__7459;
 t12 = t7;
 t13 = fmt_format__anon_4878__7847(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7346_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4172(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_deinit__1364(struct array_list_ArrayListAligned_28u8_2cnull_29__648 const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 const *t1;
 struct mem_Allocator__589 const *t2;
 struct mem_Allocator__589 t3;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t4;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t0;
 nav__1364_41 t5;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 const *)&t0;
 t2 = (struct mem_Allocator__589 const *)&t1->allocator;
 t3 = (*t2);
 t4 = (*t1);
 t5 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1410(t4);
 mem_Allocator_free__anon_3980__7436(t3, t5);
 return;
}

static void fs_File_close__1220(struct fs_File__636 const a0) {
 int32_t t0;
 t0 = a0.handle;
 posix_close__1562(t0);
 return;
}

static bool math_isPowerOfTwo__anon_3820__7347(uintptr_t const a0) {
 uint64_t t0;
 uintptr_t t2;
 bool t1;
 t0 = a0;
 t1 = t0 > UINT64_C(0);
 debug_assert__180(t1);
 t2 = a0 - (uintptr_t)1ul;
 t2 = a0 & t2;
 t0 = t2;
 t1 = t0 == UINT64_C(0);
 return t1;
}

static uint16_t posix_faccessatZ__1748(int32_t const a0, uint8_t const *const a1, uint32_t const a2, uint32_t const a3) {
 unsigned int t0;
 unsigned int t1;
 int t2;
 uint16_t t3;
 t0 = a2;
 t1 = a3;
 t2 = faccessat(a0, a1, t0, t1);
 t3 = posix_errno__anon_3559__7178(t2);
 switch (t3) {
  case UINT16_C(0): {
   return 0;
  }
  case UINT16_C(13): {
   return zig_error_PermissionDenied;
  }
  case UINT16_C(1): {
   return zig_error_PermissionDenied;
  }
  case UINT16_C(30): {
   return zig_error_ReadOnlyFileSystem;
  }
  case UINT16_C(40): {
   return zig_error_SymLinkLoop;
  }
  case UINT16_C(26): {
   return zig_error_FileBusy;
  }
  case UINT16_C(20): {
   return zig_error_FileNotFound;
  }
  case UINT16_C(2): {
   return zig_error_FileNotFound;
  }
  case UINT16_C(36): {
   return zig_error_NameTooLong;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(5): {
   return zig_error_InputOutput;
  }
  case UINT16_C(12): {
   return zig_error_SystemResources;
  }
  case UINT16_C(84): {
   t3 = posix_unexpectedErrno__1856(t3);
   return t3;
  }
  default: {
   t3 = posix_unexpectedErrno__1856(t3);
   return t3;
  }
 }
}

static nav__1604_38 posix_openatZ__1604(int32_t const a0, uint8_t const *const a1, uint32_t const a2, uintptr_t const a3) {
 unsigned long t1;
 int t0;
 int32_t t3;
 nav__1604_38 t4;
 uint16_t t2;
 zig_loop_11:
 t0 = a0;
 t1 = a3;
 t0 = openat64(t0, a1, a2, t1);
 t2 = posix_errno__anon_3559__7178(t0);
 switch (t2) {
  case UINT16_C(0): {
   t3 = t0;
   t4.payload = t3;
   t4.error = UINT16_C(0);
   return t4;
  }
  case UINT16_C(4): {
   goto zig_block_0;
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(22): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_BadPathName};
  }
  case UINT16_C(9): {
   zig_unreachable();
  }
  case UINT16_C(13): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_AccessDenied};
  }
  case UINT16_C(27): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_FileTooBig};
  }
  case UINT16_C(75): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_FileTooBig};
  }
  case UINT16_C(21): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_IsDir};
  }
  case UINT16_C(40): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_SymLinkLoop};
  }
  case UINT16_C(24): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_ProcessFdQuotaExceeded};
  }
  case UINT16_C(36): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_NameTooLong};
  }
  case UINT16_C(23): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_SystemFdQuotaExceeded};
  }
  case UINT16_C(19): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_NoDevice};
  }
  case UINT16_C(2): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_FileNotFound};
  }
  case UINT16_C(12): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_SystemResources};
  }
  case UINT16_C(28): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_NoSpaceLeft};
  }
  case UINT16_C(20): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_NotDir};
  }
  case UINT16_C(1): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_AccessDenied};
  }
  case UINT16_C(17): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_PathAlreadyExists};
  }
  case UINT16_C(16): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_DeviceBusy};
  }
  case UINT16_C(95): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_FileLocksNotSupported};
  }
  case UINT16_C(11): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_WouldBlock};
  }
  case UINT16_C(26): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_FileBusy};
  }
  case UINT16_C(6): {
   return (nav__1604_38){-INT32_C(0x55555556),zig_error_NoDevice};
  }
  case UINT16_C(84): {
   t2 = posix_unexpectedErrno__1856(t2);
   t4.payload = -INT32_C(0x55555556);
   t4.error = t2;
   return t4;
  }
  default: {
   t2 = posix_unexpectedErrno__1856(t2);
   t4.payload = -INT32_C(0x55555556);
   t4.error = t2;
   return t4;
  }
 }

 zig_block_0:;
 goto zig_loop_11;
}

static uint16_t posix_flock__1765(int32_t const a0, int32_t const a1) {
 int t0;
 uint16_t t1;
 zig_loop_3:
 t0 = a1;
 t0 = flock(a0, t0);
 t1 = posix_errno__anon_3559__7178(t0);
 switch (t1) {
  case UINT16_C(0): {
   return 0;
  }
  case UINT16_C(9): {
   zig_unreachable();
  }
  case UINT16_C(4): {
   goto zig_block_0;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(37): {
   return zig_error_SystemResources;
  }
  case UINT16_C(11): {
   return zig_error_WouldBlock;
  }
  case UINT16_C(95): {
   return zig_error_FileLocksNotSupported;
  }
  default: {
   t1 = posix_unexpectedErrno__1856(t1);
   return t1;
  }
 }

 zig_block_0:;
 goto zig_loop_3;
}

static void posix_close__1562(int32_t const a0) {
 int t0;
 uint16_t t1;
 t0 = close(a0);
 t1 = posix_errno__anon_3559__7178(t0);
 switch (t1) {
  case UINT16_C(9): {
   zig_unreachable();
  }
  case UINT16_C(4): {
   return;
  }
  default: {
   return;
  }
 }
}

static nav__1410_39 array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1410(struct array_list_ArrayListAligned_28u8_2cnull_29__648 const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 const *t1;
 nav__1410_39 const *t2;
 uint8_t *const *t3;
 uintptr_t t4;
 uint8_t *t5;
 nav__1410_39 t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 t0;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 const *)&t0;
 t2 = (nav__1410_39 const *)&t1->items;
 t3 = &t2->ptr;
 t4 = a0.capacity;
 t5 = (*t3);
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t6.ptr = t5;
 t6.len = t4;
 return t6;
}

static nav__7434_39 mem_Allocator_remap__anon_3973__7434(struct mem_Allocator__589 const a0, nav__7434_39 const a1, uintptr_t const a2) {
 struct mem_Allocator__589 const *t1;
 nav__7434_39 const *t3;
 uint64_t t4;
 struct mem_Allocator__589 t6;
 struct mem_Allocator__589 t0;
 struct mem_Allocator__589 t15;
 nav__7434_39 t7;
 nav__7434_39 t10;
 nav__7434_39 t2;
 uint8_t *t8;
 uint8_t *t20;
 uint8_t *t21;
 uint8_t *t22;
 void *t9;
 uintptr_t t11;
 uintptr_t t13;
 nav__7434_50 t12;
 struct mem_Allocator_VTable__592 const *const *t16;
 struct mem_Allocator_VTable__592 const *t17;
 uint8_t *(*const *t18)(void *, nav__7434_39, uint8_t, uintptr_t, uintptr_t);
 uint8_t *(*t19)(void *, nav__7434_39, uint8_t, uintptr_t, uintptr_t);
 uint8_t *const *t23;
 bool t5;
 uint8_t t14;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = a1;
 t3 = (nav__7434_39 const *)&t2;
 t4 = a2;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  t6 = (*t1);
  mem_Allocator_free__anon_3980__7436(t6, a1);
  t7 = (*t3);
  t8 = t7.ptr;
  t8 = (uint8_t *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t9 = (void *)t8;
  t7.ptr = t9;
  t7.len = (uintptr_t)0ul;
  t10 = t7;
  return t10;
 }
 goto zig_block_0;

 zig_block_0:;
 t11 = a1.len;
 t4 = t11;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return (nav__7434_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};
 }
 goto zig_block_1;

 zig_block_1:;
 t10 = mem_sliceAsBytes__anon_5184__7852(a1);
 t12 = math_mul__anon_3583__7179((uintptr_t)1ul, a2);
 t5 = t12.error == UINT16_C(0);
 if (t5) {
  t13 = t12.payload;
  t11 = t13;
  goto zig_block_2;
 }
 return (nav__7434_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_2:;
 t6 = (*t1);
 t14 = mem_Alignment_fromByteUnits__1066((uintptr_t)1ul);
 t13 = (uintptr_t)zig_return_address();
 t15 = t6;
 t1 = (struct mem_Allocator__589 const *)&t15;
 t16 = (struct mem_Allocator_VTable__592 const *const *)&t1->vtable;
 t17 = (*t16);
 t18 = (uint8_t *(*const *)(void *, nav__7434_39, uint8_t, uintptr_t, uintptr_t))&t17->remap;
 t19 = (*t18);
 t9 = t6.ptr;
 t20 = t19(t9, t10, t14, t11, t13);
 t5 = t20 != NULL;
 if (t5) {
  t21 = t20;
  t8 = t21;
  goto zig_block_3;
 }
 return (nav__7434_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_3:;
 t22 = t8;
 t23 = (uint8_t *const *)&t22;
 t8 = (*t23);
 t8 = (uint8_t *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t8;
 t10.len = t11;
 memcpy(&t7, &t10, sizeof(nav__7434_39));
 t7 = mem_bytesAsSlice__anon_5197__7853(t7);
 t10 = t7;
 return t10;
}

static nav__7435_40 mem_Allocator_alignedAlloc__anon_3978__7435(struct mem_Allocator__589 const a0, uintptr_t const a1) {
 struct mem_Allocator__589 const *t1;
 struct mem_Allocator__589 t2;
 struct mem_Allocator__589 t0;
 struct mem_Allocator__589 t5;
 uintptr_t t3;
 nav__7435_40 t4;
 nav__7435_40 t8;
 nav__7435_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__7435_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__589 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_3137__5312(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__7435_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static void mem_Allocator_free__anon_3980__7436(struct mem_Allocator__589 const a0, nav__7436_40 const a1) {
 struct mem_Allocator__589 const *t1;
 nav__7436_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__589 t10;
 struct mem_Allocator__589 t0;
 struct mem_Allocator__589 t12;
 struct mem_Allocator_VTable__592 const *const *t13;
 struct mem_Allocator_VTable__592 const *t14;
 void (*const *t15)(void *, nav__7436_40, uint8_t, uintptr_t);
 void (*t16)(void *, nav__7436_40, uint8_t, uintptr_t);
 void *t17;
 bool t5;
 uint8_t t11;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = mem_sliceAsBytes__anon_5184__7852(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t11 = mem_Alignment_fromByteUnits__1066((uintptr_t)1ul);
 t3 = (uintptr_t)zig_return_address();
 t12 = t10;
 t1 = (struct mem_Allocator__589 const *)&t12;
 t13 = (struct mem_Allocator_VTable__592 const *const *)&t1->vtable;
 t14 = (*t13);
 t15 = (void (*const *)(void *, nav__7436_40, uint8_t, uintptr_t))&t14->free;
 t16 = (*t15);
 t17 = t10.ptr;
 t16(t17, t2, t11, t3);
 return;
}

static uint16_t bincode_serializeInt__anon_3985__7437(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, uint32_t const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t8;
 void const *t9;
 nav__7437_48 (**t10)(void const *, nav__7437_50);
 struct io_Writer__4171 t11;
 struct io_Writer__4171 t6;
 struct io_Writer__4171 t12;
 struct io_Writer__4171 t14;
 struct io_Writer__4171 const *t13;
 nav__7437_50 t17;
 uint16_t t18;
 uint16_t t19;
 uint8_t t16[4];
 uint8_t t15[4];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__7437_48 (**)(void const *, nav__7437_50))&t6.writeFn;
 (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4100;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Writer__4171 const *)&t12;
 t11 = (*t13);
 t14 = t11;
 t13 = (struct io_Writer__4171 const *)&t14;
 memcpy(&t16, &a1, sizeof(uint8_t[4]));
 memcpy((char *)&t15, t16, sizeof(uint8_t[4]));
 t11 = (*t13);
 t17.ptr = &t15[(uintptr_t)0ul];
 t17.len = (uintptr_t)4ul;
 t18 = io_Writer_writeAll__7467(t11, t17);
 memcpy(&t19, &t18, sizeof(uint16_t));
 if (t19) {
  return t19;
 }
 return 0;
}

static uint16_t bincode_serializeStruct__anon_3990__7438(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_Command__struct_2041__2041 const a1) {
 nav__7438_45 t2;
 uint32_t t0;
 uint16_t t1;
 t0 = a1.pid;
 t1 = bincode_serialize__anon_3381__7105(a0, t0);
 if (t1) {
  return t1;
 }
 t2 = a1.uri;
 t1 = bincode_serialize__anon_5228__7854(a0, t2);
 if (t1) {
  return t1;
 }
 return 0;
}

static uint16_t bincode_serializeStruct__anon_3991__7439(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_Command__struct_2042__2042 const a1) {
 nav__7439_45 t0;
 uint16_t t1;
 t0 = a1.name;
 t1 = bincode_serialize__anon_5228__7854(a0, t0);
 if (t1) {
  return t1;
 }
 t0 = a1.version;
 t1 = bincode_serialize__anon_5228__7854(a0, t0);
 if (t1) {
  return t1;
 }
 return 0;
}

static uint16_t bincode_serializeStruct__anon_3992__7440(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_Command__struct_2043__2043 const a1) {
 struct shared_MarkerType__2045 t2;
 uint32_t t0;
 uint16_t t1;
 t0 = a1.pid;
 t1 = bincode_serialize__anon_3381__7105(a0, t0);
 if (t1) {
  return t1;
 }
 t2 = a1.marker;
 t1 = bincode_serialize__anon_5230__7855(a0, t2);
 if (t1) {
  return t1;
 }
 return 0;
}

static uint16_t bincode_serializeInt__anon_3993__7441(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, uint64_t const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t8;
 void const *t9;
 nav__7441_48 (**t10)(void const *, nav__7441_50);
 struct io_Writer__4171 t11;
 struct io_Writer__4171 t6;
 struct io_Writer__4171 t12;
 struct io_Writer__4171 t14;
 struct io_Writer__4171 const *t13;
 nav__7441_50 t17;
 uint16_t t18;
 uint16_t t19;
 uint8_t t16[8];
 uint8_t t15[8];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__7441_48 (**)(void const *, nav__7441_50))&t6.writeFn;
 (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4100;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Writer__4171 const *)&t12;
 t11 = (*t13);
 t14 = t11;
 t13 = (struct io_Writer__4171 const *)&t14;
 memcpy(&t16, &a1, sizeof(uint8_t[8]));
 memcpy((char *)&t15, t16, sizeof(uint8_t[8]));
 t11 = (*t13);
 t17.ptr = &t15[(uintptr_t)0ul];
 t17.len = (uintptr_t)8ul;
 t18 = io_Writer_writeAll__7467(t11, t17);
 memcpy(&t19, &t18, sizeof(uint16_t));
 if (t19) {
  return t19;
 }
 return 0;
}

static uint16_t bincode_serializeEnum__anon_3996__7442(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, uint8_t const a1) {
 uint32_t t1;
 uint16_t t2;
 uint8_t t0;
 t0 = a1;
 t1 = (uint32_t)t0;
 t2 = bincode_serialize__anon_3381__7105(a0, t1);
 if (t2) {
  return t2;
 }
 return 0;
}

static uint16_t posix_errno__anon_4007__7444(intptr_t const a0) {
 int64_t t1;
 int *t3;
 int t4;
 uint16_t t0;
 uint16_t t5;
 bool t2;
 t1 = a0;
 t2 = t1 == -INT64_C(1);
 if (t2) {
  t3 = zig_e___errno_location();
  t4 = (*t3);
  t5 = (uint16_t)t4;
  t0 = t5;
  goto zig_block_0;
 }
 t0 = UINT16_C(0);
 goto zig_block_0;

 zig_block_0:;
 return t0;
}

static struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4164(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 **t1;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static nav__7459_38 io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__7459(void const *const a0, nav__7459_41 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t0;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *t1;
 nav__7459_38 t2;
 nav__7459_38 t3;
 t0 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)a0;
 t1 = (*t0);
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_write__4167(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__7459_38));
 return t3;
}

static uint16_t io_Writer_writeAll__7467(struct io_Writer__4171 const a0, nav__7467_40 const a1) {
 struct io_Writer__4171 const *t1;
 nav__7467_40 const *t3;
 uintptr_t t5;
 uintptr_t t6;
 uintptr_t t13;
 uintptr_t t4;
 uint64_t t7;
 uint64_t t8;
 struct io_Writer__4171 t10;
 struct io_Writer__4171 t0;
 nav__7467_40 t11;
 nav__7467_40 t2;
 uint8_t const *t12;
 nav__7467_43 t14;
 uint16_t t15;
 bool t9;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = a1;
 t3 = (nav__7467_40 const *)&t2;
 t4 = (uintptr_t)0ul;
 zig_loop_11:
 t5 = t4;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 != t8;
 if (t9) {
  t6 = t4;
  t10 = (*t1);
  t5 = t4;
  t11 = (*t3);
  t12 = t11.ptr;
  t12 = (uint8_t const *)(((uintptr_t)t12) + (t5*sizeof(uint8_t)));
  t13 = t11.len;
  t5 = t13 - t5;
  t11.ptr = t12;
  t11.len = t5;
  t14 = io_Writer_write__7466(t10, t11);
  if (t14.error) {
   t15 = t14.error;
   return t15;
  }
  t5 = t14.payload;
  t5 = t6 + t5;
  t4 = t5;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_11;

 zig_block_0:;
 return 0;
}

static uint16_t shared_Command_format__anon_5290__7856(struct shared_Command__2038 const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2) {
 struct io_Writer__4171 const *t1;
 struct shared_Command__struct_2041__2041 t3;
 struct io_Writer__4171 t4;
 struct io_Writer__4171 t0;
 nav__7856_44 t6;
 nav__7856_44 t10;
 nav__7856_66 t7;
 struct shared_Command__struct_2042__2042 t9;
 nav__7856_76 t11;
 struct shared_Command__struct_2043__2043 t12;
 struct shared_MarkerType__2045 t13;
 nav__7856_78 t14;
 uint64_t t15;
 nav__7856_80 t16;
 uint32_t t5;
 uint16_t t8;
 uint8_t t2;
 nav__7856_84 t17;
 (void)a1;
 t0 = a2;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a0.payload.ExecutedBenchmark;
   t4 = (*t1);
   t5 = t3.pid;
   t6 = t3.uri;
   t7.f0 = t5;
   t7.f1 = t6;
   t8 = io_Writer_print__anon_5310__7857(t4, t7);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__7467(t4, (nav__7856_44){(uint8_t const *)&__anon_2281,(uintptr_t)14ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__7467(t4, (nav__7856_44){(uint8_t const *)&__anon_2286,(uintptr_t)13ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__7467(t4, (nav__7856_44){(uint8_t const *)&__anon_2291,(uintptr_t)3ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(4): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__7467(t4, (nav__7856_44){(uint8_t const *)&__anon_2296,(uintptr_t)8ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(5): {
   t9 = a0.payload.SetIntegration;
   t4 = (*t1);
   t6 = t9.name;
   t10 = t9.version;
   t11.f0 = t6;
   t11.f1 = t10;
   t8 = io_Writer_print__anon_5331__7858(t4, t11);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(6): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__7467(t4, (nav__7856_44){(uint8_t const *)&__anon_2304,(uintptr_t)3ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(7): {
   t12 = a0.payload.AddMarker;
   t4 = (*t1);
   t5 = t12.pid;
   t13 = t12.marker;
   t14.f0 = t5;
   t14.f1 = t13;
   t8 = io_Writer_print__anon_5343__7859(t4, t14);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(8): {
   t15 = a0.payload.SetVersion;
   t4 = (*t1);
   t16.f0 = t15;
   t8 = io_Writer_print__anon_5352__7860(t4, t16);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(9): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__7467(t4, (nav__7856_44){(uint8_t const *)&__anon_2319,(uintptr_t)18ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(10): {
   t2 = a0.payload.IntegrationModeResponse;
   t4 = (*t1);
   t17.f0 = t2;
   t8 = io_Writer_print__anon_5365__7861(t4, t17);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatType__anon_4653__7843(struct shared_Command__2038 const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2, uintptr_t const a3) {
 struct shared_Command__2038 const *t1;
 struct shared_Command__2038 t3;
 struct shared_Command__2038 t0;
 struct io_Writer__4171 t2;
 uint16_t t4;
 (void)a3;
 t0 = a0;
 t1 = (struct shared_Command__2038 const *)&t0;
 t2 = a2;
 t3 = (*t1);
 t4 = shared_Command_format__anon_5290__7856(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static nav__4172_39 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4172(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 const *t1;
 nav__4172_39 const *t2;
 uintptr_t t3;
 nav__4172_39 t4;
 uint8_t *t5;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t0;
 t0 = a0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 const *)&t0;
 t2 = (nav__4172_39 const *)&t1->buffer;
 t3 = a0.pos;
 t4 = (*t2);
 t5 = t4.ptr;
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4.ptr = t5;
 t4.len = t3;
 return t4;
}

static uint16_t fmt_formatType__anon_4769__7845(uint64_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2, uintptr_t const a3) {
 uint64_t t0;
 struct io_Writer__4171 t1;
 uint16_t t2;
 uint16_t t3;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_5374__7862(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatType__anon_5030__7848(nav__7848_39 const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2, uintptr_t const a3) {
 struct io_Writer__4171 const *t2;
 uint64_t t3;
 struct io_Writer__4171 t5;
 struct io_Writer__4171 t1;
 nav__7848_39 t7;
 nav__7848_39 t0;
 uint16_t t6;
 uint16_t t8;
 bool t4;
 t0 = a0;
 t1 = a2;
 t2 = (struct io_Writer__4171 const *)&t1;
 t3 = a3;
 t4 = t3 == UINT64_C(0);
 if (t4) {
  t5 = (*t2);
  t6 = io_Writer_writeAll__7467(t5, (nav__7848_39){(uint8_t const *)&__anon_5389,(uintptr_t)7ul});
  return t6;
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t7, &a0, sizeof(nav__7848_39));
 t6 = fmt_formatBuf__anon_5399__7863(t7, a1, a2);
 memcpy(&t8, &t6, sizeof(uint16_t));
 return t8;
}

static nav__7852_39 mem_sliceAsBytes__anon_5184__7852(nav__7852_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__7852_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__7852_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static nav__7853_39 mem_bytesAsSlice__anon_5197__7853(nav__7853_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__7853_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__7853_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t0 = t0 / (uintptr_t)1ul;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static nav__4100_38 io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4100(void const *const a0, nav__4100_41 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t1;
 nav__4100_38 t2;
 nav__4100_38 t3;
 t0 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)a0;
 t1 = (*t0);
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_appendWrite__1388(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__4100_38));
 return t3;
}

static uint16_t bincode_serialize__anon_5228__7854(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, nav__7854_40 const a1) {
 uint16_t t0;
 t0 = bincode_serializePointer__anon_5411__7864(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_5230__7855(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_MarkerType__2045 const a1) {
 uint16_t t0;
 t0 = bincode_serializeUnion__anon_5412__7865(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static nav__4167_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_write__4167(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const a0, nav__4167_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t1;
 nav__4167_42 const *t3;
 uintptr_t t4;
 uintptr_t t10;
 uint64_t t5;
 uint64_t t11;
 uintptr_t *t7;
 nav__4167_50 *t8;
 nav__4167_50 t9;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *t12;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *t0;
 uint8_t *t13;
 nav__4167_42 t14;
 nav__4167_42 t2;
 uint8_t const *t15;
 nav__4167_38 t16;
 bool t6;
 t0 = a0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t0;
 t2 = a1;
 t3 = (nav__4167_42 const *)&t2;
 t4 = a1.len;
 t5 = t4;
 t6 = t5 == UINT64_C(0);
 if (t6) {
  return (nav__4167_38){(uintptr_t)0ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (uintptr_t *)&a0->pos;
 t4 = (*t7);
 t8 = (nav__4167_50 *)&a0->buffer;
 t9 = (*t8);
 t10 = t9.len;
 t5 = t4;
 t11 = t10;
 t6 = t5 >= t11;
 if (t6) {
  return (nav__4167_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoSpaceLeft};
 }
 goto zig_block_1;

 zig_block_1:;
 t8 = (nav__4167_50 *)&a0->buffer;
 t9 = (*t8);
 t10 = t9.len;
 t7 = (uintptr_t *)&a0->pos;
 t4 = (*t7);
 t4 = t10 - t4;
 t10 = a1.len;
 t10 = (t4 < t10) ? t4 : t10;
 t11 = t10;
 t12 = (*t1);
 t8 = (nav__4167_50 *)&t12->buffer;
 t7 = (uintptr_t *)&a0->pos;
 t10 = (*t7);
 t9 = (*t8);
 t13 = t9.ptr;
 t13 = (uint8_t *)(((uintptr_t)t13) + (t10*sizeof(uint8_t)));
 t10 = t11;
 t9.ptr = t13;
 t9.len = t10;
 t14 = (*t3);
 t15 = t14.ptr;
 t15 = (uint8_t const *)(((uintptr_t)t15) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10 = t11;
 t14.ptr = t15;
 t14.len = t10;
 t15 = t14.ptr;
 if (t9.len != 0) memcpy(t9.ptr, t15, t9.len * sizeof(uint8_t));
 t12 = (*t1);
 t7 = (uintptr_t *)&t12->pos;
 t10 = (*t7);
 t4 = t11;
 t4 = t10 + t4;
 (*t7) = t4;
 t6 = t11 == UINT64_C(0);
 if (t6) {
  return (nav__4167_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoSpaceLeft};
 }
 goto zig_block_2;

 zig_block_2:;
 t4 = t11;
 t16.payload = t4;
 t16.error = UINT16_C(0);
 return t16;
}

static nav__7466_38 io_Writer_write__7466(struct io_Writer__4171 const a0, nav__7466_41 const a1) {
 struct io_Writer__4171 const *t1;
 nav__7466_38 (*const *t2)(void const *, nav__7466_41);
 nav__7466_38 (*t3)(void const *, nav__7466_41);
 void const *t4;
 nav__7466_38 t5;
 struct io_Writer__4171 t0;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (nav__7466_38 (*const *)(void const *, nav__7466_41))&t1->writeFn;
 t3 = (*t2);
 t4 = a0.context;
 t5 = t3(t4, a1);
 return t5;
}

static uint16_t io_Writer_print__anon_5310__7857(struct io_Writer__4171 const a0, nav__7857_41 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5416__7866(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5331__7858(struct io_Writer__4171 const a0, nav__7858_41 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5420__7867(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5343__7859(struct io_Writer__4171 const a0, nav__7859_40 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5424__7868(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5352__7860(struct io_Writer__4171 const a0, nav__7860_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5428__7869(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5365__7861(struct io_Writer__4171 const a0, nav__7861_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5432__7870(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static nav__7488_38 unicode_utf8ByteSequenceLength__7488(uint8_t const a0) {
 nav__7488_38 t0;
 switch (a0) {
  default: if ((a0 >= UINT8_C(0) && a0 <= UINT8_C(127))) {
   t0 = (nav__7488_38){0,UINT8_C(1)};
   goto zig_block_0;
  }if ((a0 >= UINT8_C(192) && a0 <= UINT8_C(223))) {
   t0 = (nav__7488_38){0,UINT8_C(2)};
   goto zig_block_0;
  }if ((a0 >= UINT8_C(224) && a0 <= UINT8_C(239))) {
   t0 = (nav__7488_38){0,UINT8_C(3)};
   goto zig_block_0;
  }if ((a0 >= UINT8_C(240) && a0 <= UINT8_C(247))) {
   t0 = (nav__7488_38){0,UINT8_C(4)};
   goto zig_block_0;
  }{
   t0 = (nav__7488_38){zig_error_Utf8InvalidStartByte,UINT8_C(0x2)};
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return t0;
}

static nav__7504_38 unicode_utf8CountCodepoints__7504(nav__7504_40 const a0) {
 nav__7504_40 const *t1;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t2;
 uintptr_t t3;
 uint64_t t6;
 uint64_t t7;
 nav__7504_40 t9;
 nav__7504_40 t0;
 uint8_t const *t10;
 uint8_t const (*t11)[8];
 nav__7504_38 t16;
 nav__7504_50 t17;
 nav__7504_48 t14;
 uint16_t t15;
 bool t8;
 uint8_t t12[8];
 uint8_t t13;
 t0 = a0;
 t1 = (nav__7504_40 const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = (uintptr_t)0ul;
 zig_loop_9:
 t4 = t3;
 t5 = a0.len;
 t6 = t4;
 t7 = t5;
 t8 = t6 < t7;
 if (t8) {
  zig_loop_18:
  t5 = t3;
  t5 = t5 + (uintptr_t)8ul;
  t4 = a0.len;
  t7 = t5;
  t6 = t4;
  t8 = t7 <= t6;
  if (t8) {
   t4 = t3;
   t9 = (*t1);
   t10 = t9.ptr;
   t10 = (uint8_t const *)(((uintptr_t)t10) + (t4*sizeof(uint8_t)));
   t11 = (uint8_t const (*)[8])t10;
   memcpy(t12, (const char *)t11, sizeof(uint8_t[8]));
   memcpy(&t4, &t12, sizeof(uintptr_t));
   t4 = zig_wrap_u64(t4, UINT8_C(64));
   t4 = t4 & (uintptr_t)9259542123273814144ul;
   t6 = t4;
   t8 = t6 != UINT64_C(0);
   if (t8) {
    goto zig_block_2;
   }
   goto zig_block_4;

   zig_block_4:;
   t4 = t2;
   t4 = t4 + (uintptr_t)8ul;
   t2 = t4;
   t4 = t3;
   t4 = t4 + (uintptr_t)8ul;
   t3 = t4;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_18;

  zig_block_2:;
  t5 = t3;
  t4 = a0.len;
  t7 = t5;
  t6 = t4;
  t8 = t7 < t6;
  if (t8) {
   t4 = t3;
   t13 = a0.ptr[t4];
   t14 = unicode_utf8ByteSequenceLength__7488(t13);
   if (t14.error) {
    t15 = t14.error;
    t16.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
    t16.error = t15;
    return t16;
   }
   t13 = t14.payload;
   t4 = t3;
   t5 = (uintptr_t)t13;
   t5 = t4 + t5;
   t4 = a0.len;
   t6 = t5;
   t7 = t4;
   t8 = t6 > t7;
   if (t8) {
    return (nav__7504_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_TruncatedInput};
   }
   goto zig_block_6;

   zig_block_6:;
   switch (t13) {
    case UINT8_C(1): {
     goto zig_block_7;
    }
    default: {
     t4 = t3;
     t9 = (*t1);
     t10 = t9.ptr;
     t10 = (uint8_t const *)(((uintptr_t)t10) + (t4*sizeof(uint8_t)));
     t4 = (uintptr_t)t13;
     t9.ptr = t10;
     t9.len = t4;
     t17 = unicode_utf8Decode__7494(t9);
     if (t17.error) {
      t15 = t17.error;
      t16.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
      t16.error = t15;
      return t16;
     }
     goto zig_block_7;
    }
   }

   zig_block_7:;
   t4 = t3;
   t5 = (uintptr_t)t13;
   t5 = t4 + t5;
   t3 = t5;
   t5 = t2;
   t5 = t5 + (uintptr_t)1ul;
   t2 = t5;
   goto zig_block_5;
  }
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_9;

 zig_block_0:;
 t5 = t2;
 t16.payload = t5;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_formatBuf__anon_5399__7863(nav__7863_39 const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2) {
 struct io_Writer__4171 const *t1;
 nav__7863_44 t2;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t7;
 nav__7863_48 t6;
 uint64_t t8;
 uint64_t t9;
 struct io_Writer__4171 t10;
 struct io_Writer__4171 t0;
 nav__7863_57 t15;
 nav__7863_39 t17;
 nav__7863_39 t20;
 uint8_t *t19;
 uint32_t t14;
 uint16_t t11;
 uint16_t t12;
 nav__7863_60 t16;
 bool t3;
 uint8_t t18;
 uint8_t t13[4];
 t0 = a2;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = a1.width;
 t3 = t2.is_null != true;
 if (t3) {
  t4 = t2.payload;
  t6 = unicode_utf8CountCodepoints__7504(a0);
  t3 = t6.error == UINT16_C(0);
  if (t3) {
   t7 = t6.payload;
   t5 = t7;
   goto zig_block_1;
  }
  t7 = a0.len;
  t5 = t7;
  goto zig_block_1;

  zig_block_1:;
  t8 = t5;
  t9 = t4;
  t3 = t8 < t9;
  if (t3) {
   t5 = t4 - t5;
   t7 = t5;
   goto zig_block_2;
  }
  t7 = (uintptr_t)0ul;
  goto zig_block_2;

  zig_block_2:;
  t9 = t7;
  t3 = t9 == UINT64_C(0);
  if (t3) {
   t10 = (*t1);
   t11 = io_Writer_writeAll__7467(t10, a0);
   memcpy(&t12, &t11, sizeof(uint16_t));
   return t12;
  }
  goto zig_block_3;

  zig_block_3:;
  t14 = a1.fill;
  t15.ptr = &t13[(uintptr_t)0ul];
  t15.len = (uintptr_t)4ul;
  t16 = unicode_utf8Encode__7489(t14, t15);
  t3 = t16.error == UINT16_C(0);
  if (t3) {
   t18 = t16.payload;
   t19 = (uint8_t *)&t13;
   t19 = (uint8_t *)(((uintptr_t)t19) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t5 = (uintptr_t)t18;
   t15.ptr = t19;
   t15.len = t5;
   memcpy(&t20, &t15, sizeof(nav__7863_39));
   t17 = t20;
   goto zig_block_4;

  }
  t12 = t16.error;
  switch (t12) {
   case zig_error_Utf8CannotEncodeSurrogateHalf:
   case zig_error_CodepointTooLarge: {
    t17 = (nav__7863_39){(uint8_t const *)&__anon_5566,(uintptr_t)3ul};
    goto zig_block_4;
   }
   default: zig_unreachable();
  }

  zig_block_4:;
  t18 = a1.alignment;
  switch (t18) {
   case UINT8_C(0): {
    t10 = (*t1);
    t12 = io_Writer_writeAll__7467(t10, a0);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__7471(t10, t17, t7);
    if (t12) {
     return t12;
    }
    goto zig_block_6;
   }
   case UINT8_C(1): {
    t5 = t7 / (uintptr_t)2ul;
    t7 = t7 + (uintptr_t)1ul;
    t7 = t7 / (uintptr_t)2ul;
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__7471(t10, t17, t5);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeAll__7467(t10, a0);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__7471(t10, t17, t7);
    if (t12) {
     return t12;
    }
    goto zig_block_6;
   }
   case UINT8_C(2): {
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__7471(t10, t17, t7);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeAll__7467(t10, a0);
    if (t12) {
     return t12;
    }
    goto zig_block_6;
   }
   default: zig_unreachable();
  }

  zig_block_6:;
  goto zig_block_0;
 }
 t10 = (*t1);
 t12 = io_Writer_writeAll__7467(t10, a0);
 if (t12) {
  return t12;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatInt__anon_5459__7872(uint64_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4577 const a3, struct io_Writer__4171 const a4) {
 uint64_t t4;
 uint64_t t11;
 uint64_t t2;
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__7872_58 t12;
 nav__7872_47 t13;
 uint16_t t14;
 uint16_t t15;
 bool t0;
 uint8_t t8;
 nav__7872_56 t9;
 uint8_t t10[2];
 uint8_t t1[65];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)65ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_19:
  t4 = t2;
  t0 = t4 >= UINT64_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT64_C(100);
   t8 = (uint8_t)t4;
   t9 = fmt_digits2__7243(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT64_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_19;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT64_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t8 = (uint8_t)t4;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = (uint8_t)t4;
  t9 = fmt_digits2__7243(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_74:
 t4 = t2;
 t11 = (uint64_t)a1;
 t11 = t4 % t11;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__7259(t8, a2);
 (*t6) = t8;
 t11 = t2;
 t4 = (uint64_t)a1;
 t4 = t11 / t4;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT64_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_74;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)65ul - t5;
 t12.ptr = t6;
 t12.len = t5;
 memcpy(&t13, &t12, sizeof(nav__7872_47));
 t14 = fmt_formatBuf__anon_5399__7863(t13, a3, a4);
 memcpy(&t15, &t14, sizeof(uint16_t));
 return t15;
}

static uint16_t fmt_formatIntValue__anon_5442__7871(uint64_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_5459__7872(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_5374__7862(uint64_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_5442__7871(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static nav__1388_38 array_list_ArrayListAligned_28u8_2cnull_29_appendWrite__1388(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const a0, nav__1388_42 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t0;
 nav__1388_38 t4;
 uintptr_t t5;
 uint16_t t3;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t0;
 t2 = (*t1);
 t3 = array_list_ArrayListAligned_28u8_2cnull_29_appendSlice__1382(t2, a1);
 if (t3) {
  t4.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
  t4.error = t3;
  return t4;
 }
 t5 = a1.len;
 t4.payload = t5;
 t4.error = UINT16_C(0);
 return t4;
}

static uint16_t bincode_serializePointer__anon_5411__7864(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, nav__7864_40 const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *t24;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 t23;
 uintptr_t t3;
 uint64_t t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *const *t8;
 void const **t10;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t11;
 void const *t12;
 nav__7864_51 (**t13)(void const *, nav__7864_40);
 struct io_Writer__4171 t14;
 struct io_Writer__4171 t9;
 struct io_Writer__4171 t15;
 struct io_Writer__4171 t17;
 struct io_Writer__4171 t25;
 struct io_Writer__4171 t26;
 struct io_Writer__4171 const *t16;
 nav__7864_40 t20;
 uint16_t t21;
 uint16_t t22;
 uint8_t t19[8];
 uint8_t t18[8];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *)&t0;
 t2 = (*t1);
 t3 = a1.len;
 t4 = t3;
 t5 = t2;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *)&t5;
 t7 = t6;
 t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *const *)&t7;
 t10 = (void const **)&t9.context;
 t6 = (*t8);
 t11 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t6->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__7864_51 (**)(void const *, nav__7864_40))&t9.writeFn;
 (*t13) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4100;
 t14 = t9;
 t15 = t14;
 t16 = (struct io_Writer__4171 const *)&t15;
 t14 = (*t16);
 t17 = t14;
 t16 = (struct io_Writer__4171 const *)&t17;
 memcpy(&t19, &t4, sizeof(uint8_t[8]));
 memcpy((char *)&t18, t19, sizeof(uint8_t[8]));
 t14 = (*t16);
 t20.ptr = &t18[(uintptr_t)0ul];
 t20.len = (uintptr_t)8ul;
 t21 = io_Writer_writeAll__7467(t14, t20);
 memcpy(&t22, &t21, sizeof(uint16_t));
 if (t22) {
  return t22;
 }
 t2 = (*t1);
 t23 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *)&t23;
 t24 = t1;
 t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const *const *)&t24;
 t10 = (void const **)&t25.context;
 t1 = (*t8);
 t11 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t1->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__7864_51 (**)(void const *, nav__7864_40))&t25.writeFn;
 (*t13) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4100;
 t14 = t25;
 t26 = t14;
 t16 = (struct io_Writer__4171 const *)&t26;
 t14 = (*t16);
 t22 = io_Writer_writeAll__7467(t14, a1);
 memcpy(&t21, &t22, sizeof(uint16_t));
 if (t21) {
  return t21;
 }
 return 0;
}

static uint16_t bincode_serializeUnion__anon_5412__7865(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__2080 const a0, struct shared_MarkerType__2045 const a1) {
 uint64_t t4;
 uint32_t t1;
 uint16_t t2;
 uint8_t t0;
 bool t3;
 t0 = a1.tag;
 t1 = (uint32_t)t0;
 t2 = bincode_serialize__anon_3381__7105(a0, t1);
 if (t2) {
  return t2;
 }
 t0 = a1.tag;
 t3 = t0 == UINT8_C(0);
 if (t3) {
  t4 = a1.payload.SampleStart;
  t2 = bincode_serialize__anon_3391__7110(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(1);
 if (t3) {
  t4 = a1.payload.SampleEnd;
  t2 = bincode_serialize__anon_3391__7110(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(2);
 if (t3) {
  t4 = a1.payload.BenchmarkStart;
  t2 = bincode_serialize__anon_3391__7110(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(3);
 if (t3) {
  t4 = a1.payload.BenchmarkEnd;
  t2 = bincode_serialize__anon_3391__7110(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 return 0;
}

static uint16_t fmt_format__anon_5416__7866(struct io_Writer__4171 const a0, nav__7866_41 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 nav__7866_40 t5;
 uint32_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7866_40){(uint8_t const *)&__anon_5638,(uintptr_t)25ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5690__7873(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7866_40){(uint8_t const *)&__anon_5704,(uintptr_t)7ul});
 if (t3) {
  return t3;
 }
 t5 = a1.f1;
 t3 = fmt_formatType__anon_5720__7874(t5, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7866_40){(uint8_t const *)&__anon_5729,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5420__7867(struct io_Writer__4171 const a0, nav__7867_41 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 nav__7867_40 t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7867_40){(uint8_t const *)&__anon_5782,(uintptr_t)23ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5720__7874(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7867_40){(uint8_t const *)&__anon_5807,(uintptr_t)11ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f1;
 t3 = fmt_formatType__anon_5720__7874(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7867_40){(uint8_t const *)&__anon_5729,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5424__7868(struct io_Writer__4171 const a0, nav__7868_40 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 struct shared_MarkerType__2045 t5;
 uint32_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7868_45){(uint8_t const *)&__anon_5863,(uintptr_t)17ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5690__7873(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7868_45){(uint8_t const *)&__anon_5885,(uintptr_t)10ul});
 if (t3) {
  return t3;
 }
 t5 = a1.f1;
 t3 = fmt_formatType__anon_5895__7875(t5, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7868_45){(uint8_t const *)&__anon_5729,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5428__7869(struct io_Writer__4171 const a0, nav__7869_39 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7869_44){(uint8_t const *)&__anon_5955,(uintptr_t)31ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5964__7876(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7869_44){(uint8_t const *)&__anon_5729,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5432__7870(struct io_Writer__4171 const a0, nav__7870_39 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 uint16_t t3;
 uint8_t t4;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7870_44){(uint8_t const *)&__anon_6010,(uintptr_t)24ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_6018__7877(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7243_39 fmt_digits2__7243(uint8_t const a0) {
 uint8_t *t1;
 uint8_t t2;
 nav__7243_39 t3;
 uint8_t t0[2];
 t1 = (uint8_t *)&t0[(uintptr_t)0ul];
 t2 = a0 / UINT8_C(10);
 t2 = UINT8_C(48) + t2;
 (*t1) = t2;
 t1 = (uint8_t *)&t0[(uintptr_t)1ul];
 t2 = a0 % UINT8_C(10);
 t2 = UINT8_C(48) + t2;
 (*t1) = t2;
 memcpy(t3.array, t0, sizeof(uint8_t[2]));
 return t3;
}

static uint8_t fmt_digitToChar__7259(uint8_t const a0, uint8_t const a1) {
 uint8_t t0;
 uint8_t t1;
 bool t2;
 switch (a0) {
  default: if ((a0 >= UINT8_C(0) && a0 <= UINT8_C(9))) {
   t1 = a0 + UINT8_C(48);
   t0 = t1;
   goto zig_block_0;
  }if ((a0 >= UINT8_C(10) && a0 <= UINT8_C(35))) {
   t2 = a1 == UINT8_C(1);
   if (t2) {
    t1 = UINT8_C(65);
    goto zig_block_1;
   }
   t1 = UINT8_C(97);
   goto zig_block_1;

   zig_block_1:;
   t1 = t1 - UINT8_C(10);
   t1 = a0 + t1;
   t0 = t1;
   goto zig_block_0;
  }{
   zig_unreachable();
  }
 }

 zig_block_0:;
 return t0;
}

static nav__7494_38 unicode_utf8Decode__7494(nav__7494_40 const a0) {
 nav__7494_40 const *t1;
 uintptr_t t2;
 nav__7494_40 t7;
 nav__7494_40 t0;
 uint8_t const *t8;
 uint8_t const (*t9)[2];
 uint8_t const (*t13)[3];
 uint8_t const (*t16)[4];
 nav__7494_38 t3;
 nav__7494_38 t6;
 nav__7494_38 t12;
 uint32_t t5;
 uint8_t t4;
 uint8_t t10[2];
 nav__7494_48 t11;
 uint8_t t14[3];
 nav__7494_52 t15;
 uint8_t t17[4];
 nav__7494_56 t18;
 t0 = a0;
 t1 = (nav__7494_40 const *)&t0;
 t2 = a0.len;
 switch (t2) {
  case (uintptr_t)1ul: {
   t4 = a0.ptr[(uintptr_t)0ul];
   t5 = (uint32_t)t4;
   t6.payload = t5;
   t6.error = UINT16_C(0);
   t3 = t6;
   goto zig_block_0;
  }
  case (uintptr_t)2ul: {
   t7 = (*t1);
   t8 = t7.ptr;
   t8 = (uint8_t const *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t9 = (uint8_t const (*)[2])t8;
   memcpy(t10, (const char *)t9, sizeof(uint8_t[2]));
   memcpy(t11.array, t10, sizeof(nav__7494_48));
   t6 = unicode_utf8Decode2__7496(t11);
   memcpy(&t12, &t6, sizeof(nav__7494_38));
   t3 = t12;
   goto zig_block_0;
  }
  case (uintptr_t)3ul: {
   t7 = (*t1);
   t8 = t7.ptr;
   t8 = (uint8_t const *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t13 = (uint8_t const (*)[3])t8;
   memcpy(t14, (const char *)t13, sizeof(uint8_t[3]));
   memcpy(t15.array, t14, sizeof(nav__7494_52));
   t12 = unicode_utf8Decode3__7498(t15);
   memcpy(&t6, &t12, sizeof(nav__7494_38));
   t3 = t6;
   goto zig_block_0;
  }
  case (uintptr_t)4ul: {
   t7 = (*t1);
   t8 = t7.ptr;
   t8 = (uint8_t const *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t16 = (uint8_t const (*)[4])t8;
   memcpy(t17, (const char *)t16, sizeof(uint8_t[4]));
   memcpy(t18.array, t17, sizeof(nav__7494_56));
   t12 = unicode_utf8Decode4__7502(t18);
   memcpy(&t6, &t12, sizeof(nav__7494_38));
   t3 = t6;
   goto zig_block_0;
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 return t3;
}

static nav__7487_38 unicode_utf8CodepointSequenceLength__7487(uint32_t const a0) {
 bool t0;
 t0 = a0 < UINT32_C(128);
 if (t0) {
  return (nav__7487_38){0,UINT8_C(1)};
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a0 < UINT32_C(2048);
 if (t0) {
  return (nav__7487_38){0,UINT8_C(2)};
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a0 < UINT32_C(65536);
 if (t0) {
  return (nav__7487_38){0,UINT8_C(3)};
 }
 goto zig_block_2;

 zig_block_2:;
 t0 = a0 < UINT32_C(1114112);
 if (t0) {
  return (nav__7487_38){0,UINT8_C(4)};
 }
 goto zig_block_3;

 zig_block_3:;
 return (nav__7487_38){zig_error_CodepointTooLarge,UINT8_C(0x2)};
}

static nav__7878_38 unicode_utf8EncodeImpl__anon_6041__7878(uint32_t const a0, nav__7878_40 const a1) {
 nav__7878_40 const *t1;
 uintptr_t t5;
 uint64_t t6;
 uint64_t t7;
 nav__7878_40 t9;
 nav__7878_40 t0;
 uint8_t *t10;
 uint32_t t12;
 nav__7878_38 t2;
 uint16_t t3;
 uint8_t t4;
 uint8_t t11;
 bool t8;
 t0 = a1;
 t1 = (nav__7878_40 const *)&t0;
 t2 = unicode_utf8CodepointSequenceLength__7487(a0);
 if (t2.error) {
  t3 = t2.error;
  t2.payload = UINT8_C(0x2);
  t2.error = t3;
  return t2;
 }
 t4 = t2.payload;
 t5 = a1.len;
 t6 = t5;
 t7 = (uint64_t)t4;
 t8 = t6 >= t7;
 debug_assert__180(t8);
 switch (t4) {
  case UINT8_C(1): {
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t11 = (uint8_t)a0;
   (*t10) = t11;
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t12 = zig_shr_u32(a0, UINT8_C(6));
   t12 = UINT32_C(192) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)1ul];
   t12 = a0 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t8 = unicode_isSurrogateCodepoint__7567(a0);
   if (t8) {
    return (nav__7878_38){zig_error_Utf8CannotEncodeSurrogateHalf,UINT8_C(0x2)};
   }
   goto zig_block_1;

   zig_block_1:;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t12 = zig_shr_u32(a0, UINT8_C(12));
   t12 = UINT32_C(224) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)1ul];
   t12 = zig_shr_u32(a0, UINT8_C(6));
   t12 = t12 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)2ul];
   t12 = a0 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   goto zig_block_0;
  }
  case UINT8_C(4): {
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t12 = zig_shr_u32(a0, UINT8_C(18));
   t12 = UINT32_C(240) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)1ul];
   t12 = zig_shr_u32(a0, UINT8_C(12));
   t12 = t12 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)2ul];
   t12 = zig_shr_u32(a0, UINT8_C(6));
   t12 = t12 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)3ul];
   t12 = a0 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   goto zig_block_0;
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__7489_38 unicode_utf8Encode__7489(uint32_t const a0, nav__7489_40 const a1) {
 nav__7489_38 t0;
 nav__7489_38 t1;
 t0 = unicode_utf8EncodeImpl__anon_6041__7878(a0, a1);
 memcpy(&t1, &t0, sizeof(nav__7489_38));
 return t1;
}

static uint16_t io_Writer_writeBytesNTimes__7471(struct io_Writer__4171 const a0, nav__7471_40 const a1, uintptr_t const a2) {
 struct io_Writer__4171 const *t1;
 uintptr_t t3;
 uintptr_t t2;
 uint64_t t4;
 uint64_t t5;
 struct io_Writer__4171 t7;
 struct io_Writer__4171 t0;
 uint16_t t8;
 bool t6;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (uintptr_t)0ul;
 zig_loop_9:
 t3 = t2;
 t4 = t3;
 t5 = a2;
 t6 = t4 < t5;
 if (t6) {
  t7 = (*t1);
  t8 = io_Writer_writeAll__7467(t7, a1);
  if (t8) {
   return t8;
  }
  t3 = t2;
  t3 = t3 + (uintptr_t)1ul;
  t2 = t3;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_9;

 zig_block_0:;
 return 0;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_appendSlice__1382(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const a0, nav__1382_41 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t0;
 uintptr_t t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t0;
 t2 = (*t1);
 t3 = a1.len;
 t4 = array_list_ArrayListAligned_28u8_2cnull_29_ensureUnusedCapacity__1401(t2, t3);
 if (t4) {
  return t4;
 }
 t2 = (*t1);
 array_list_ArrayListAligned_28u8_2cnull_29_appendSliceAssumeCapacity__1383(t2, a1);
 return 0;
}

static uint16_t fmt_formatType__anon_5690__7873(uint32_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2, uintptr_t const a3) {
 struct io_Writer__4171 t1;
 uint32_t t0;
 uint16_t t2;
 uint16_t t3;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_6065__7879(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatType__anon_5720__7874(nav__7874_39 const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2, uintptr_t const a3) {
 struct io_Writer__4171 const *t2;
 uint64_t t3;
 struct io_Writer__4171 t5;
 struct io_Writer__4171 t1;
 nav__7874_39 t0;
 uint16_t t6;
 uint16_t t7;
 bool t4;
 t0 = a0;
 t1 = a2;
 t2 = (struct io_Writer__4171 const *)&t1;
 t3 = a3;
 t4 = t3 == UINT64_C(0);
 if (t4) {
  t5 = (*t2);
  t6 = io_Writer_writeAll__7467(t5, (nav__7874_39){(uint8_t const *)&__anon_5389,(uintptr_t)7ul});
  return t6;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = fmt_formatBuf__anon_5399__7863(a0, a1, a2);
 memcpy(&t7, &t6, sizeof(uint16_t));
 return t7;
}

static uint16_t shared_MarkerType_format__anon_6084__7880(struct shared_MarkerType__2045 const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2) {
 struct io_Writer__4171 const *t1;
 uint64_t t3;
 struct io_Writer__4171 t4;
 struct io_Writer__4171 t0;
 nav__7880_56 t5;
 uint16_t t6;
 uint8_t t2;
 (void)a1;
 t0 = a2;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a0.payload.SampleStart;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_6098__7881(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t3 = a0.payload.SampleEnd;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_6103__7882(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t3 = a0.payload.BenchmarkStart;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_6110__7883(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t3 = a0.payload.BenchmarkEnd;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_6116__7884(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatType__anon_5895__7875(struct shared_MarkerType__2045 const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2, uintptr_t const a3) {
 struct shared_MarkerType__2045 const *t1;
 struct shared_MarkerType__2045 t3;
 struct shared_MarkerType__2045 t0;
 struct io_Writer__4171 t2;
 uint16_t t4;
 (void)a3;
 t0 = a0;
 t1 = (struct shared_MarkerType__2045 const *)&t0;
 t2 = a2;
 t3 = (*t1);
 t4 = shared_MarkerType_format__anon_6084__7880(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t fmt_formatType__anon_5964__7876(uint64_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2, uintptr_t const a3) {
 uint64_t t0;
 struct io_Writer__4171 t1;
 uint16_t t2;
 uint16_t t3;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_6118__7885(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatType__anon_6018__7877(uint8_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2, uintptr_t const a3) {
 struct io_Writer__4171 const *t2;
 struct io_Writer__4171 t3;
 struct io_Writer__4171 t1;
 nav__7877_47 t5;
 nav__7877_47 t6;
 uint16_t t4;
 uint8_t t0;
 (void)a1;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = (struct io_Writer__4171 const *)&t1;
 t3 = (*t2);
 t4 = io_Writer_writeAll__7467(t3, (nav__7877_47){(uint8_t const *)&__anon_6125,(uintptr_t)22ul});
 if (t4) {
  return t4;
 }
 t3 = (*t2);
 t4 = io_Writer_writeAll__7467(t3, (nav__7877_47){(uint8_t const *)&__anon_6130,(uintptr_t)1ul});
 if (t4) {
  return t4;
 }
 t3 = (*t2);
 t5 = zig_tag_name_shared_IntegrationMode__563(a0);
 memcpy(&t6, &t5, sizeof(nav__7877_47));
 t4 = io_Writer_writeAll__7467(t3, t6);
 if (t4) {
  return t4;
 }
 return 0;
}

static nav__7496_38 unicode_utf8Decode2__7496(nav__7496_40 const a0) {
 uint32_t t3;
 uint32_t t4;
 uint32_t t2;
 nav__7496_38 t5;
 uint8_t t0;
 bool t1;
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(224);
 t1 = t0 == UINT8_C(192);
 debug_assert__180(t1);
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(31);
 t3 = (uint32_t)t0;
 t2 = t3;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__7496_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t4 = t2;
 t1 = t4 < UINT32_C(128);
 if (t1) {
  return (nav__7496_38){UINT32_C(0xaaaaa),zig_error_Utf8OverlongEncoding};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t2;
 t5.payload = t4;
 t5.error = UINT16_C(0);
 return t5;
}

static nav__7498_38 unicode_utf8Decode3__7498(nav__7498_40 const a0) {
 nav__7498_38 t1;
 uint32_t t3;
 uint16_t t2;
 nav__7498_40 t0;
 bool t4;
 bool t5;
 memcpy(t0.array, a0.array, sizeof(nav__7498_40));
 t1 = unicode_utf8Decode3AllowSurrogateHalf__7500(t0);
 if (t1.error) {
  t2 = t1.error;
  t1.payload = UINT32_C(0xaaaaa);
  t1.error = t2;
  return t1;
 }
 t3 = t1.payload;
 t4 = UINT32_C(55296) <= t3;
 if (t4) {
  t4 = t3 <= UINT32_C(57343);
  t5 = t4;
  goto zig_block_1;
 }
 t5 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t5) {
  return (nav__7498_38){UINT32_C(0xaaaaa),zig_error_Utf8EncodesSurrogateHalf};
 }
 goto zig_block_0;

 zig_block_0:;
 t1.payload = t3;
 t1.error = UINT16_C(0);
 return t1;
}

static nav__7502_38 unicode_utf8Decode4__7502(nav__7502_40 const a0) {
 uint32_t t3;
 uint32_t t4;
 uint32_t t2;
 nav__7502_38 t5;
 uint8_t t0;
 bool t1;
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(248);
 t1 = t0 == UINT8_C(240);
 debug_assert__180(t1);
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(7);
 t3 = (uint32_t)t0;
 t2 = t3;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__7502_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__7502_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t2;
 t4 = zig_shlw_u32(t4, UINT8_C(6), UINT8_C(21));
 t2 = t4;
 t4 = t2;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(63);
 t3 = (uint32_t)t0;
 t3 = t4 | t3;
 t2 = t3;
 t0 = a0.array[(uintptr_t)3ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__7502_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_2;

 zig_block_2:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)3ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t4 = t2;
 t1 = t4 < UINT32_C(65536);
 if (t1) {
  return (nav__7502_38){UINT32_C(0xaaaaa),zig_error_Utf8OverlongEncoding};
 }
 goto zig_block_3;

 zig_block_3:;
 t4 = t2;
 t1 = t4 > UINT32_C(1114111);
 if (t1) {
  return (nav__7502_38){UINT32_C(0xaaaaa),zig_error_Utf8CodepointTooLarge};
 }
 goto zig_block_4;

 zig_block_4:;
 t4 = t2;
 t5.payload = t4;
 t5.error = UINT16_C(0);
 return t5;
}

static bool unicode_isSurrogateCodepoint__7567(uint32_t const a0) {
 bool t0;
 switch (a0) {
  default: if ((a0 >= UINT32_C(55296) && a0 <= UINT32_C(57343))) {
   t0 = true;
   goto zig_block_0;
  }{
   t0 = false;
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return t0;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureUnusedCapacity__1401(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t0;
 nav__1401_43 *t3;
 nav__1401_43 t4;
 uintptr_t t5;
 nav__1401_52 t6;
 uint16_t t7;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t0;
 t2 = (*t1);
 t3 = (nav__1401_43 *)&a0->items;
 t4 = (*t3);
 t5 = t4.len;
 t6 = array_list_addOrOom__1326(t5, a1);
 if (t6.error) {
  t7 = t6.error;
  return t7;
 }
 t5 = t6.payload;
 t7 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1399(t2, t5);
 return t7;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_appendSliceAssumeCapacity__1383(struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const a0, nav__1383_41 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *t1;
 nav__1383_47 *t2;
 nav__1383_47 t3;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t7;
 uintptr_t *t6;
 uint64_t t8;
 uint64_t t9;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__648 *t0;
 uint8_t *t12;
 uint8_t const *t13;
 bool t10;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__648 *const *)&t0;
 t2 = (nav__1383_47 *)&a0->items;
 t3 = (*t2);
 t4 = t3.len;
 t5 = a1.len;
 t5 = t4 + t5;
 t6 = (uintptr_t *)&a0->capacity;
 t7 = (*t6);
 t8 = t5;
 t9 = t7;
 t10 = t8 <= t9;
 debug_assert__180(t10);
 t11 = (*t1);
 t2 = (nav__1383_47 *)&t11->items;
 t6 = &t2->len;
 (*t6) = t5;
 t11 = (*t1);
 t2 = (nav__1383_47 *)&t11->items;
 t5 = a1.len;
 t3 = (*t2);
 t12 = t3.ptr;
 t12 = (uint8_t *)(((uintptr_t)t12) + (t4*sizeof(uint8_t)));
 t3.ptr = t12;
 t3.len = t5;
 t13 = a1.ptr;
 if (t3.len != 0) memcpy(t3.ptr, t13, t3.len * sizeof(uint8_t));
 return;
}

static uint16_t fmt_formatInt__anon_6172__7887(uint32_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4577 const a3, struct io_Writer__4171 const a4) {
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__7887_58 t12;
 nav__7887_47 t13;
 uint32_t t4;
 uint32_t t11;
 uint32_t t2;
 uint16_t t14;
 uint16_t t15;
 bool t0;
 uint8_t t8;
 nav__7887_56 t9;
 uint8_t t10[2];
 uint8_t t1[33];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)33ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_19:
  t4 = t2;
  t0 = t4 >= UINT32_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT32_C(100);
   t8 = (uint8_t)t4;
   t9 = fmt_digits2__7243(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT32_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_19;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT32_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t8 = (uint8_t)t4;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = (uint8_t)t4;
  t9 = fmt_digits2__7243(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_74:
 t4 = t2;
 t11 = (uint32_t)a1;
 t11 = t4 % t11;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__7259(t8, a2);
 (*t6) = t8;
 t11 = t2;
 t4 = (uint32_t)a1;
 t4 = t11 / t4;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT32_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_74;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)33ul - t5;
 t12.ptr = t6;
 t12.len = t5;
 memcpy(&t13, &t12, sizeof(nav__7887_47));
 t14 = fmt_formatBuf__anon_5399__7863(t13, a3, a4);
 memcpy(&t15, &t14, sizeof(uint16_t));
 return t15;
}

static uint16_t fmt_formatIntValue__anon_6161__7886(uint32_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_6172__7887(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_6065__7879(uint32_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_6161__7886(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_6098__7881(struct io_Writer__4171 const a0, nav__7881_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_6182__7888(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_6103__7882(struct io_Writer__4171 const a0, nav__7882_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_6186__7889(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_6110__7883(struct io_Writer__4171 const a0, nav__7883_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_6190__7890(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_6116__7884(struct io_Writer__4171 const a0, nav__7884_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_6194__7891(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatIntValue__anon_6198__7892(uint64_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_5459__7872(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_6118__7885(uint64_t const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_6198__7892(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static nav__7500_38 unicode_utf8Decode3AllowSurrogateHalf__7500(nav__7500_40 const a0) {
 uint32_t t3;
 uint32_t t4;
 uint32_t t2;
 nav__7500_38 t5;
 uint8_t t0;
 bool t1;
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(240);
 t1 = t0 == UINT8_C(224);
 debug_assert__180(t1);
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(15);
 t3 = (uint32_t)t0;
 t2 = t3;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__7500_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__7500_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t2;
 t4 = zig_shlw_u32(t4, UINT8_C(6), UINT8_C(21));
 t2 = t4;
 t4 = t2;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(63);
 t3 = (uint32_t)t0;
 t3 = t4 | t3;
 t2 = t3;
 t3 = t2;
 t1 = t3 < UINT32_C(2048);
 if (t1) {
  return (nav__7500_38){UINT32_C(0xaaaaa),zig_error_Utf8OverlongEncoding};
 }
 goto zig_block_2;

 zig_block_2:;
 t3 = t2;
 t5.payload = t3;
 t5.error = UINT16_C(0);
 return t5;
}

static nav__1326_38 array_list_addOrOom__1326(uintptr_t const a0, uintptr_t const a1) {
 nav__1326_42 t2;
 uintptr_t t3;
 uintptr_t t0;
 uintptr_t const *t5;
 uint8_t const *t6;
 nav__1326_38 t8;
 uint8_t t4;
 uint8_t t1;
 bool t7;
 t2.f1 = zig_addo_u64(&t2.f0, a0, a1, UINT8_C(64));
 t3 = t2.f0;
 t0 = t3;
 t4 = t2.f1;
 t1 = t4;
 t5 = (uintptr_t const *)&t0;
 t6 = (uint8_t const *)&t1;
 t4 = (*t6);
 t7 = t4 != UINT8_C(0);
 if (t7) {
  return (nav__1326_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_OutOfMemory};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = (*t5);
 t8.payload = t3;
 t8.error = UINT16_C(0);
 return t8;
}

static uint16_t fmt_format__anon_6182__7888(struct io_Writer__4171 const a0, nav__7888_39 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7888_44){(uint8_t const *)&__anon_6221,(uintptr_t)12ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5964__7876(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7888_44){(uint8_t const *)&__anon_6231,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_6186__7889(struct io_Writer__4171 const a0, nav__7889_39 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7889_44){(uint8_t const *)&__anon_6250,(uintptr_t)10ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5964__7876(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7889_44){(uint8_t const *)&__anon_6231,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_6190__7890(struct io_Writer__4171 const a0, nav__7890_39 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7890_44){(uint8_t const *)&__anon_6282,(uintptr_t)15ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5964__7876(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7890_44){(uint8_t const *)&__anon_6231,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_6194__7891(struct io_Writer__4171 const a0, nav__7891_39 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7891_44){(uint8_t const *)&__anon_6310,(uintptr_t)13ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5964__7876(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7891_44){(uint8_t const *)&__anon_6231,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

void c_instrument_hooks_deinit__238(struct instruments_root_InstrumentHooks__547 *const a0) {
 struct instruments_root_InstrumentHooks__547 *t1;
 struct instruments_root_InstrumentHooks__547 *t4;
 struct instruments_root_InstrumentHooks__547 *t2;
 struct instruments_root_InstrumentHooks__547 *t5;
 struct instruments_root_InstrumentHooks__547 *const *t3;
 struct instruments_root_InstrumentHooks__547 t6;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t9;
 bool t0;
 uint8_t t7;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t2;
  t4 = (*t3);
  t5 = t4;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t5;
  t6 = (*t4);
  t7 = t6.tag;
  switch (t7) {
   case UINT8_C(0): {
    goto zig_block_1;
   }
   case UINT8_C(1): {
    t4 = (*t3);
    t8 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *)&t4->payload.perf;
    instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_deinit__761(t8);
    goto zig_block_1;
   }
   case UINT8_C(2): {
    t4 = (*t3);
    t9 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *)&t4->payload.analysis;
    instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_deinit__774(t9);
    goto zig_block_1;
   }
   case UINT8_C(3): {
    goto zig_block_1;
   }
   default: zig_unreachable();
  }

  zig_block_1:;
  mem_Allocator_destroy__anon_6345__7893((struct mem_Allocator__589){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)&heap_CAllocator_vtable__3705)}, t1);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_deinit__761(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 runner_fifo_RunnerFifo_deinit__1084(t3);
 return;
}

static void instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_deinit__774(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 runner_fifo_RunnerFifo_deinit__1084(t3);
 return;
}

static void mem_Allocator_destroy__anon_6345__7893(struct mem_Allocator__589 const a0, struct instruments_root_InstrumentHooks__547 *const a1) {
 struct mem_Allocator__589 const *t1;
 uint8_t *t2;
 uint8_t *t3;
 uint8_t *const *t4;
 struct mem_Allocator__589 t5;
 struct mem_Allocator__589 t0;
 struct mem_Allocator__589 t10;
 uint8_t (*t6)[152];
 nav__7893_52 t7;
 uintptr_t t9;
 struct mem_Allocator_VTable__592 const *const *t11;
 struct mem_Allocator_VTable__592 const *t12;
 void (*const *t13)(void *, nav__7893_52, uint8_t, uintptr_t);
 void (*t14)(void *, nav__7893_52, uint8_t, uintptr_t);
 void *t15;
 uint8_t t8;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = (uint8_t *)a1;
 t3 = t2;
 t4 = (uint8_t *const *)&t3;
 t5 = (*t1);
 t2 = (*t4);
 t2 = (uint8_t *)(((uintptr_t)t2) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t6 = (uint8_t (*)[152])t2;
 t7.ptr = &(*t6)[(uintptr_t)0ul];
 t7.len = (uintptr_t)152ul;
 t8 = mem_Alignment_fromByteUnits__1066((uintptr_t)8ul);
 t9 = (uintptr_t)zig_return_address();
 t10 = t5;
 t1 = (struct mem_Allocator__589 const *)&t10;
 t11 = (struct mem_Allocator_VTable__592 const *const *)&t1->vtable;
 t12 = (*t11);
 t13 = (void (*const *)(void *, nav__7893_52, uint8_t, uintptr_t))&t12->free;
 t14 = (*t13);
 t15 = t5.ptr;
 t14(t15, t7, t8, t9);
 return;
}

bool c_instrument_hooks_is_instrumented__239(struct instruments_root_InstrumentHooks__547 *const a0) {
 struct instruments_root_InstrumentHooks__547 *t1;
 struct instruments_root_InstrumentHooks__547 *t2;
 struct instruments_root_InstrumentHooks__547 *const *t3;
 struct instruments_root_InstrumentHooks__547 t4;
 bool t0;
 bool t6;
 uint8_t t5;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t2;
  t1 = (*t3);
  t4 = (*t1);
  t5 = t4.tag;
  switch (t5) {
   case UINT8_C(0): {
    t5 = running_on_valgrind();
    t6 = t5 > UINT8_C(0);
    t0 = t6;
    goto zig_block_1;
   }
   case UINT8_C(1): {
    t0 = true;
    goto zig_block_1;
   }
   case UINT8_C(2): {
    t0 = true;
    goto zig_block_1;
   }
   case UINT8_C(3): {
    t0 = false;
    goto zig_block_1;
   }
   default: zig_unreachable();
  }

  zig_block_1:;
  return t0;
 }
 goto zig_block_0;

 zig_block_0:;
 return false;
}

static zig_cold uint16_t runner_fifo_RunnerFifo_start_benchmark__1087(struct runner_fifo_RunnerFifo__621 *const a0) {
 struct runner_fifo_RunnerFifo__621 *const *t1;
 struct runner_fifo_RunnerFifo__621 *t2;
 struct runner_fifo_RunnerFifo__621 *t0;
 struct fifo_UnixPipe_Writer__628 *t3;
 struct fifo_UnixPipe_Reader__630 *t5;
 uint16_t t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__621 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_UnixPipe_Writer__628 *)&t2->writer;
 t4 = fifo_UnixPipe_Writer_sendCmd__1121(t3, (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(1)});
 if (t4) {
  return t4;
 }
 t2 = (*t1);
 t5 = (struct fifo_UnixPipe_Reader__630 *)&t2->reader;
 t4 = fifo_UnixPipe_Reader_waitForAck__1128(t5, (nav__1087_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_start_benchmark__763(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_start_benchmark__1087(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_start_benchmark__776(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_start_benchmark__1087(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_start_benchmark__240(struct instruments_root_InstrumentHooks__547 *const a0) {
 struct instruments_root_InstrumentHooks__547 *t1;
 struct instruments_root_InstrumentHooks__547 *t2;
 struct instruments_root_InstrumentHooks__547 *t5;
 struct instruments_root_InstrumentHooks__547 *const *t3;
 struct instruments_root_InstrumentHooks__547 t6;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t11;
 uint16_t t4;
 uint16_t t9;
 uint16_t t10;
 bool t0;
 uint8_t t7;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t2;
  t1 = (*t3);
  t5 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t5;
  t6 = (*t1);
  t7 = t6.tag;
  t0 = t7 == UINT8_C(1);
  if (t0) {
   t1 = (*t3);
   t8 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *)&t1->payload.perf;
   t9 = instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_start_benchmark__763(t8);
   memcpy(&t10, &t9, sizeof(uint16_t));
   t4 = t10;
   goto zig_block_2;
  }
  t6 = (*t1);
  t7 = t6.tag;
  t0 = t7 == UINT8_C(0);
  if (t0) {
   t0 = features_is_feature_enabled__324(UINT64_C(0));
   t0 = !t0;
   if (t0) {
    callgrind_zero_stats();
    callgrind_start_instrumentation();
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t4 = 0;
   goto zig_block_2;
  }
  t6 = (*t1);
  t7 = t6.tag;
  t0 = t7 == UINT8_C(2);
  if (t0) {
   t1 = (*t3);
   t11 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *)&t1->payload.analysis;
   t10 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_start_benchmark__776(t11);
   memcpy(&t9, &t10, sizeof(uint16_t));
   t4 = t9;
   goto zig_block_2;
  }
  goto zig_block_6;

  zig_block_6:;
  goto zig_block_4;

  zig_block_4:;
  goto zig_block_3;

  zig_block_3:;
  t4 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t10, &t4, sizeof(uint16_t));
  t0 = t10 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static bool features_is_feature_enabled__324(uint64_t const a0) {
 uint64_t t0;
 uint64_t t1;
 uintptr_t t2;
 bool t3;
 t0 = (*&features_features__322);
 t1 = a0;
 t2 = t1;
 t3 = bit_set_IntegerBitSet_2864_29_isSet__359(t0, t2);
 return t3;
}

static bool bit_set_IntegerBitSet_2864_29_isSet__359(uint64_t const a0, uintptr_t const a1) {
 uint64_t t0;
 uint64_t t2;
 bool t1;
 t0 = a1;
 t1 = t0 < UINT64_C(64);
 debug_assert__180(t1);
 t0 = zig_wrap_u64((uint64_t)a0, UINT8_C(64));
 t2 = bit_set_IntegerBitSet_2864_29_maskBit__383(a1);
 t2 = t0 & t2;
 t1 = t2 != UINT64_C(0);
 return t1;
}

static zig_cold uint16_t runner_fifo_RunnerFifo_stop_benchmark__1088(struct runner_fifo_RunnerFifo__621 *const a0) {
 struct runner_fifo_RunnerFifo__621 *const *t1;
 struct runner_fifo_RunnerFifo__621 *t2;
 struct runner_fifo_RunnerFifo__621 *t0;
 struct fifo_UnixPipe_Writer__628 *t3;
 struct fifo_UnixPipe_Reader__630 *t5;
 uint16_t t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__621 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_UnixPipe_Writer__628 *)&t2->writer;
 t4 = fifo_UnixPipe_Writer_sendCmd__1121(t3, (struct shared_Command__2038){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},UINT32_C(0xaaaaaaaa)}},UINT8_C(2)});
 if (t4) {
  return t4;
 }
 t2 = (*t1);
 t5 = (struct fifo_UnixPipe_Reader__630 *)&t2->reader;
 t4 = fifo_UnixPipe_Reader_waitForAck__1128(t5, (nav__1088_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_stop_benchmark__764(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_stop_benchmark__1088(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_stop_benchmark__777(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_stop_benchmark__1088(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_stop_benchmark__241(struct instruments_root_InstrumentHooks__547 *const a0) {
 struct instruments_root_InstrumentHooks__547 *t1;
 struct instruments_root_InstrumentHooks__547 *t2;
 struct instruments_root_InstrumentHooks__547 *t5;
 struct instruments_root_InstrumentHooks__547 *const *t3;
 struct instruments_root_InstrumentHooks__547 t6;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t11;
 uint16_t t4;
 uint16_t t9;
 uint16_t t10;
 bool t0;
 uint8_t t7;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t2;
  t1 = (*t3);
  t5 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t5;
  t6 = (*t1);
  t7 = t6.tag;
  t0 = t7 == UINT8_C(0);
  if (t0) {
   t0 = features_is_feature_enabled__324(UINT64_C(0));
   t0 = !t0;
   if (t0) {
    callgrind_stop_instrumentation();
    goto zig_block_4;
   }
   goto zig_block_4;

   zig_block_4:;
   t4 = 0;
   goto zig_block_2;
  }
  t6 = (*t1);
  t7 = t6.tag;
  t0 = t7 == UINT8_C(1);
  if (t0) {
   t1 = (*t3);
   t8 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *)&t1->payload.perf;
   t9 = instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_stop_benchmark__764(t8);
   memcpy(&t10, &t9, sizeof(uint16_t));
   t4 = t10;
   goto zig_block_2;
  }
  t6 = (*t1);
  t7 = t6.tag;
  t0 = t7 == UINT8_C(2);
  if (t0) {
   t1 = (*t3);
   t11 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *)&t1->payload.analysis;
   t10 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_stop_benchmark__777(t11);
   memcpy(&t9, &t10, sizeof(uint16_t));
   t4 = t9;
   goto zig_block_2;
  }
  goto zig_block_6;

  zig_block_6:;
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_3;

  zig_block_3:;
  t4 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t10, &t4, sizeof(uint16_t));
  t0 = t10 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static uint16_t runner_fifo_RunnerFifo_set_executed_benchmark__1089(struct runner_fifo_RunnerFifo__621 *const a0, uint32_t const a1, uint8_t const *const a2) {
 struct runner_fifo_RunnerFifo__621 *const *t1;
 struct runner_fifo_RunnerFifo__621 *t2;
 struct runner_fifo_RunnerFifo__621 *t0;
 struct fifo_UnixPipe_Writer__628 *t3;
 nav__1089_61 t4;
 nav__1089_61 t5;
 struct shared_Command__struct_2041__2041 t6;
 struct shared_Command__2038 t7;
 struct fifo_UnixPipe_Reader__630 *t9;
 uint16_t t8;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__621 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_UnixPipe_Writer__628 *)&t2->writer;
 t4 = mem_span__anon_6487__7898(a2);
 memcpy(&t5, &t4, sizeof(nav__1089_61));
 t6.uri = t5;
 t6.pid = a1;
 t7.tag = UINT8_C(0);
 t7.payload.ExecutedBenchmark = t6;
 t8 = fifo_UnixPipe_Writer_sendCmd__1121(t3, t7);
 if (t8) {
  return t8;
 }
 t2 = (*t1);
 t9 = (struct fifo_UnixPipe_Reader__630 *)&t2->reader;
 t8 = fifo_UnixPipe_Reader_waitForAck__1128(t9, (nav__1089_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t8) {
  return t8;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_set_executed_benchmark__765(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const a0, uint32_t const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_executed_benchmark__1089(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_executed_benchmark__778(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const a0, uint32_t const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_executed_benchmark__1089(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_executed_benchmark__242(struct instruments_root_InstrumentHooks__547 *const a0, uint32_t const a1, uint8_t const *const a2) {
 struct instruments_root_InstrumentHooks__547 *t1;
 struct instruments_root_InstrumentHooks__547 *t2;
 struct instruments_root_InstrumentHooks__547 *t5;
 struct instruments_root_InstrumentHooks__547 *const *t3;
 struct instruments_root_InstrumentHooks__547 t6;
 uint8_t const *t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t9;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t11;
 uint16_t t4;
 uint16_t t10;
 bool t0;
 uint8_t t7;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t2;
  t1 = (*t3);
  t5 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t5;
  t6 = (*t1);
  t7 = t6.tag;
  switch (t7) {
   case UINT8_C(0): {
    t8 = (uint8_t const *)a2;
    callgrind_dump_stats_at(t8);
    goto zig_block_3;
   }
   case UINT8_C(1): {
    t1 = (*t3);
    t9 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *)&t1->payload.perf;
    t10 = instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_set_executed_benchmark__765(t9, a1, a2);
    if (t10) {
     t4 = t10;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(2): {
    t1 = (*t3);
    t11 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *)&t1->payload.analysis;
    t10 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_executed_benchmark__778(t11, a1, a2);
    if (t10) {
     t4 = t10;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(3): {
    goto zig_block_3;
   }
   default: zig_unreachable();
  }

  zig_block_3:;
  t4 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t10, &t4, sizeof(uint16_t));
  t0 = t10 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static nav__7898_39 mem_span__anon_6487__7898(uint8_t const *const a0) {
 uint8_t const *const *t1;
 uintptr_t t2;
 uint8_t const *t3;
 uint8_t const *t0;
 nav__7898_39 t4;
 t0 = a0;
 t1 = (uint8_t const *const *)&t0;
 t2 = mem_len__anon_6504__7899(a0);
 t3 = (*t1);
 t3 = (uint8_t const *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4.ptr = t3;
 t4.len = t2;
 return t4;
}

static uintptr_t mem_len__anon_6504__7899(uint8_t const *const a0) {
 uint8_t const *t1;
 uintptr_t t2;
 bool t0;
 t0 = a0 != NULL;
 debug_assert__180(t0);
 t1 = (uint8_t const *)a0;
 t2 = mem_indexOfSentinel__anon_6511__7900(t1);
 return t2;
}

static uintptr_t mem_indexOfSentinel__anon_6511__7900(uint8_t const *const a0) {
 static uint8_t const t11[64] = "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000";
 static uint8_t const t18[64] = "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000";
 uint8_t const *const *t1;
 uintptr_t t4;
 uintptr_t t6;
 uintptr_t t16;
 uintptr_t t2;
 uint8_t const *t5;
 uint8_t const *t0;
 uint64_t t7;
 uint8_t const (*t8)[64];
 uint8_t const (*t17)[64];
 bool t3;
 uint8_t t9[64];
 uint8_t t10[64];
 bool t12[64];
 nav__7900_45 t13;
 nav__7900_47 t14;
 uint8_t t15;
 t0 = a0;
 t1 = (uint8_t const *const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = math_isPowerOfTwo__anon_6522__7901();
 if (t3) {
  debug_assert__180(true);
  t4 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)&t5[t4];
  t4 = (uintptr_t)t5;
  t6 = t4 & (uintptr_t)4095ul;
  t7 = t6;
  t3 = t7 <= UINT64_C(4032);
  if (t3) {
   t6 = t2;
   t5 = (*t1);
   t5 = (uint8_t const *)(((uintptr_t)t5) + (t6*sizeof(uint8_t)));
   t8 = (uint8_t const (*)[64])t5;
   memcpy(t9, (const char *)t8, sizeof(uint8_t[64]));
   memcpy(&t10, &t9, sizeof(uint8_t[64]));
   for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)64ul; t6 += (uintptr_t)1ul) {
    t12[t6] = t10[t6] == t11[t6];
   }
   t3 = false;
   for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)64ul; t6 += (uintptr_t)1ul) {
    t3 |= t12[t6];
   }
   if (t3) {
    t4 = t2;
    memcpy(t13.array, t12, sizeof(nav__7900_45));
    t14 = simd_firstTrue__anon_6579__7908(t13);
    t15 = t14.payload;
    t6 = (uintptr_t)t15;
    t6 = t4 + t6;
    return t6;
   }
   goto zig_block_2;

   zig_block_2:;
   t6 = t2;
   t16 = mem_alignForward__anon_6582__7909(t4, (uintptr_t)64ul);
   t4 = t16 - t4;
   t4 = t4 / (uintptr_t)1ul;
   t4 = t6 + t4;
   t2 = t4;
   goto zig_block_1;
  }
  zig_loop_63:
  t4 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)&t5[t4];
  t4 = (uintptr_t)t5;
  t4 = t4 & (uintptr_t)63ul;
  t7 = t4;
  t3 = t7 != UINT64_C(0);
  if (t3) {
   t4 = t2;
   t15 = a0[t4];
   t3 = t15 == UINT8_C(0);
   if (t3) {
    t4 = t2;
    return t4;
   }
   goto zig_block_5;

   zig_block_5:;
   t4 = t2;
   t4 = t4 + (uintptr_t)1ul;
   t2 = t4;
   goto zig_block_4;
  }
  goto zig_block_3;

  zig_block_4:;
  goto zig_loop_63;

  zig_block_3:;
  goto zig_block_1;

  zig_block_1:;
  t6 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)&t5[t6];
  t6 = (uintptr_t)t5;
  t3 = mem_isAligned__935(t6, (uintptr_t)64ul);
  debug_assert__180(t3);
  zig_loop_98:
  t6 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)(((uintptr_t)t5) + (t6*sizeof(uint8_t)));
  t8 = (uint8_t const (*)[64])t5;
  t17 = (uint8_t const (*)[64])t8;
  memcpy(t10, (const char *)t17, sizeof(uint8_t[64]));
  for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)64ul; t6 += (uintptr_t)1ul) {
   t12[t6] = t10[t6] == t18[t6];
  }
  t3 = false;
  for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)64ul; t6 += (uintptr_t)1ul) {
   t3 |= t12[t6];
  }
  if (t3) {
   t6 = t2;
   memcpy(t13.array, t12, sizeof(nav__7900_45));
   t14 = simd_firstTrue__anon_6579__7908(t13);
   t15 = t14.payload;
   t4 = (uintptr_t)t15;
   t4 = t6 + t4;
   return t4;
  }
  goto zig_block_6;

  zig_block_6:;
  t6 = t2;
  t6 = t6 + (uintptr_t)64ul;
  t2 = t6;
  goto zig_loop_98;
 }
 goto zig_block_0;

 zig_block_0:;
 zig_loop_129:
 t16 = t2;
 t15 = a0[t16];
 t3 = t15 != UINT8_C(0);
 if (t3) {
  t16 = t2;
  t16 = t16 + (uintptr_t)1ul;
  t2 = t16;
  goto zig_block_8;
 }
 goto zig_block_7;

 zig_block_8:;
 goto zig_loop_129;

 zig_block_7:;
 t16 = t2;
 return t16;
}

static bool math_isPowerOfTwo__anon_6522__7901(void) {
 debug_assert__180(true);
 return true;
}

static nav__7908_38 simd_firstTrue__anon_6579__7908(nav__7908_40 const a0) {
 static uint8_t const t2[64] = {UINT8_C(0),UINT8_C(1),UINT8_C(2),UINT8_C(3),UINT8_C(4),UINT8_C(5),UINT8_C(6),UINT8_C(7),UINT8_C(8),UINT8_C(9),UINT8_C(10),UINT8_C(11),UINT8_C(12),UINT8_C(13),UINT8_C(14),UINT8_C(15),UINT8_C(16),UINT8_C(17),UINT8_C(18),UINT8_C(19),UINT8_C(20),UINT8_C(21),UINT8_C(22),UINT8_C(23),UINT8_C(24),UINT8_C(25),UINT8_C(26),UINT8_C(27),UINT8_C(28),UINT8_C(29),UINT8_C(30),UINT8_C(31),UINT8_C(32),UINT8_C(33),UINT8_C(34),UINT8_C(35),UINT8_C(36),UINT8_C(37),UINT8_C(38),UINT8_C(39),UINT8_C(40),UINT8_C(41),UINT8_C(42),UINT8_C(43),UINT8_C(44),UINT8_C(45),UINT8_C(46),UINT8_C(47),UINT8_C(48),UINT8_C(49),UINT8_C(50),UINT8_C(51),UINT8_C(52),UINT8_C(53),UINT8_C(54),UINT8_C(55),UINT8_C(56),UINT8_C(57),UINT8_C(58),UINT8_C(59),UINT8_C(60),UINT8_C(61),UINT8_C(62),UINT8_C(63)};
 static uint8_t const t3[64] = {UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63),UINT8_C(63)};
 uintptr_t t1;
 bool t0;
 uint8_t t4[64];
 uint8_t t5;
 nav__7908_38 t6;
 t0 = false;
 for (t1 = (uintptr_t)0ul; t1 < (uintptr_t)64ul; t1 += (uintptr_t)1ul) {
  t0 |= a0.array[t1];
 }
 t0 = !t0;
 if (t0) {
  return (nav__7908_38){true,UINT8_C(0x2a)};
 }
 goto zig_block_0;

 zig_block_0:;
 for (t1 = (uintptr_t)0ul; t1 < (uintptr_t)64ul; t1 += (uintptr_t)1ul) {
  t4[t1] = a0.array[t1] ? t2[t1] : t3[t1];
 }
 t5 = UINT8_C(63);
 for (t1 = (uintptr_t)0ul; t1 < (uintptr_t)64ul; t1 += (uintptr_t)1ul) {
  t5 = t5 < t4[t1] ? t5 : t4[t1];
 }
 t6.is_null = false;
 t6.payload = t5;
 return t6;
}

static uintptr_t mem_alignForward__anon_6582__7909(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 bool t0;
 t0 = mem_isValidAlignGeneric__anon_6671__7910(a1);
 debug_assert__180(t0);
 t1 = a1 - (uintptr_t)1ul;
 t1 = a0 + t1;
 t1 = mem_alignBackward__anon_6672__7911(t1, a1);
 return t1;
}

static bool mem_isAligned__935(uintptr_t const a0, uintptr_t const a1) {
 uint64_t t0;
 uint64_t t1;
 bool t2;
 t0 = a0;
 t1 = a1;
 t2 = mem_isAlignedGeneric__anon_6676__7912(t0, t1);
 return t2;
}

static bool mem_isValidAlignGeneric__anon_6671__7910(uintptr_t const a0) {
 uint64_t t0;
 bool t1;
 bool t2;
 t0 = a0;
 t1 = t0 > UINT64_C(0);
 if (t1) {
  t1 = math_isPowerOfTwo__anon_3820__7347(a0);
  t2 = t1;
  goto zig_block_0;
 }
 t2 = false;
 goto zig_block_0;

 zig_block_0:;
 return t2;
}

static uintptr_t mem_alignBackward__anon_6672__7911(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 bool t0;
 t0 = mem_isValidAlignGeneric__anon_6671__7910(a1);
 debug_assert__180(t0);
 t1 = a1 - (uintptr_t)1ul;
 t1 = zig_not_u64(t1, UINT8_C(64));
 t1 = a0 & t1;
 return t1;
}

static bool mem_isAlignedGeneric__anon_6676__7912(uint64_t const a0, uint64_t const a1) {
 uint64_t t0;
 bool t1;
 t0 = mem_alignBackward__anon_6678__7913(a0, a1);
 t1 = t0 == a0;
 return t1;
}

static uint64_t mem_alignBackward__anon_6678__7913(uint64_t const a0, uint64_t const a1) {
 uint64_t t1;
 bool t0;
 t0 = mem_isValidAlignGeneric__anon_6680__7914(a1);
 debug_assert__180(t0);
 t1 = a1 - UINT64_C(1);
 t1 = zig_not_u64(t1, UINT8_C(64));
 t1 = a0 & t1;
 return t1;
}

static bool mem_isValidAlignGeneric__anon_6680__7914(uint64_t const a0) {
 bool t0;
 bool t1;
 t0 = a0 > UINT64_C(0);
 if (t0) {
  t0 = math_isPowerOfTwo__anon_6681__7915(a0);
  t1 = t0;
  goto zig_block_0;
 }
 t1 = false;
 goto zig_block_0;

 zig_block_0:;
 return t1;
}

static bool math_isPowerOfTwo__anon_6681__7915(uint64_t const a0) {
 uint64_t t1;
 bool t0;
 t0 = a0 > UINT64_C(0);
 debug_assert__180(t0);
 t1 = a0 - UINT64_C(1);
 t1 = a0 & t1;
 t0 = t1 == UINT64_C(0);
 return t0;
}

uint8_t c_instrument_hooks_executed_benchmark__243(struct instruments_root_InstrumentHooks__547 *const a0, uint32_t const a1, uint8_t const *const a2) {
 uint8_t t0;
 t0 = c_instrument_hooks_set_executed_benchmark__242(a0, a1, a2);
 return t0;
}

static uint16_t runner_fifo_RunnerFifo_set_integration__1090(struct runner_fifo_RunnerFifo__621 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct runner_fifo_RunnerFifo__621 *const *t1;
 struct runner_fifo_RunnerFifo__621 *t2;
 struct runner_fifo_RunnerFifo__621 *t0;
 struct fifo_UnixPipe_Writer__628 *t3;
 nav__1090_61 t4;
 nav__1090_61 t5;
 nav__1090_61 t6;
 struct shared_Command__struct_2042__2042 t7;
 struct shared_Command__2038 t8;
 struct fifo_UnixPipe_Reader__630 *t10;
 uint16_t t9;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__621 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_UnixPipe_Writer__628 *)&t2->writer;
 t4 = mem_span__anon_6487__7898(a1);
 t5 = mem_span__anon_6487__7898(a2);
 memcpy(&t6, &t4, sizeof(nav__1090_61));
 memcpy(&t4, &t5, sizeof(nav__1090_61));
 t7.name = t6;
 t7.version = t4;
 t8.tag = UINT8_C(5);
 t8.payload.SetIntegration = t7;
 t9 = fifo_UnixPipe_Writer_sendCmd__1121(t3, t8);
 if (t9) {
  return t9;
 }
 t2 = (*t1);
 t10 = (struct fifo_UnixPipe_Reader__630 *)&t2->reader;
 t9 = fifo_UnixPipe_Reader_waitForAck__1128(t10, (nav__1090_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t9) {
  return t9;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_set_integration__766(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_integration__1090(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_integration__779(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_integration__1090(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_integration__244(struct instruments_root_InstrumentHooks__547 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct instruments_root_InstrumentHooks__547 *t1;
 struct instruments_root_InstrumentHooks__547 *t2;
 struct instruments_root_InstrumentHooks__547 *t5;
 struct instruments_root_InstrumentHooks__547 *const *t3;
 struct instruments_root_InstrumentHooks__547 t6;
 struct instruments_valgrind_ValgrindInstrument__554 *t8;
 struct instruments_valgrind_ValgrindInstrument__554 t9;
 struct instruments_valgrind_ValgrindInstrument__554 t11;
 struct instruments_valgrind_ValgrindInstrument__554 const *t12;
 struct mem_Allocator__589 t13;
 nav__244_71 t14;
 nav__244_73 t15;
 nav__244_55 t17;
 uint8_t *t18;
 uint8_t const *t19;
 struct mem_Allocator__589 const *t20;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t21;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t22;
 uint16_t t4;
 uint16_t t10;
 uint16_t t16;
 bool t0;
 uint8_t t7;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t2;
  t1 = (*t3);
  t5 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t5;
  t6 = (*t1);
  t7 = t6.tag;
  switch (t7) {
   case UINT8_C(0): {
    t1 = (*t3);
    t8 = (struct instruments_valgrind_ValgrindInstrument__554 *)&t1->payload.valgrind;
    t9 = (*t8);
    t11 = t9;
    t12 = (struct instruments_valgrind_ValgrindInstrument__554 const *)&t11;
    t13 = t9.allocator;
    t14.f0 = a1;
    t14.f1 = a2;
    t15 = fmt_allocPrintZ__anon_6716__7916(t13, t14);
    if (t15.error) {
     t16 = t15.error;
     t10 = t16;
     goto zig_block_4;
    }
    t17 = t15.payload;
    t18 = t17.ptr;
    t19 = (uint8_t const *)t18;
    callgrind_dump_stats_at(t19);
    t20 = (struct mem_Allocator__589 const *)&t12->allocator;
    t13 = (*t20);
    mem_Allocator_free__anon_6718__7917(t13, t17);
    t10 = 0;
    goto zig_block_4;

    zig_block_4:;
    memcpy(&t16, &t10, sizeof(uint16_t));
    if (t16) {
     t4 = t16;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(1): {
    t1 = (*t3);
    t21 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *)&t1->payload.perf;
    t16 = instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_set_integration__766(t21, a1, a2);
    if (t16) {
     t4 = t16;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(2): {
    t1 = (*t3);
    t22 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *)&t1->payload.analysis;
    t16 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_integration__779(t22, a1, a2);
    if (t16) {
     t4 = t16;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(3): {
    goto zig_block_3;
   }
   default: zig_unreachable();
  }

  zig_block_3:;
  t4 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t16, &t4, sizeof(uint16_t));
  t0 = t16 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static nav__7916_40 fmt_allocPrintZ__anon_6716__7916(struct mem_Allocator__589 const a0, nav__7916_43 const a1) {
 nav__7916_40 t0;
 nav__7916_39 t2;
 nav__7916_39 t3;
 nav__7916_39 const *t4;
 uintptr_t t5;
 uint8_t *t6;
 uint16_t t1;
 t0 = fmt_allocPrint__anon_6753__7918(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (nav__7916_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t3 = t2;
 t4 = (nav__7916_39 const *)&t3;
 t5 = t2.len;
 t5 = t5 - (uintptr_t)1ul;
 t2 = (*t4);
 t6 = t2.ptr;
 t6 = (uint8_t *)(((uintptr_t)t6) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t6;
 t2.len = t5;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static void mem_Allocator_free__anon_6718__7917(struct mem_Allocator__589 const a0, nav__7917_40 const a1) {
 struct mem_Allocator__589 const *t1;
 nav__7917_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__589 t10;
 struct mem_Allocator__589 t0;
 struct mem_Allocator__589 t12;
 struct mem_Allocator_VTable__592 const *const *t13;
 struct mem_Allocator_VTable__592 const *t14;
 void (*const *t15)(void *, nav__7917_40, uint8_t, uintptr_t);
 void (*t16)(void *, nav__7917_40, uint8_t, uintptr_t);
 void *t17;
 bool t5;
 uint8_t t11;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t2 = mem_sliceAsBytes__anon_6760__7919(a1);
 t3 = t2.len;
 t3 = t3 + (uintptr_t)1ul;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t11 = mem_Alignment_fromByteUnits__1066((uintptr_t)1ul);
 t3 = (uintptr_t)zig_return_address();
 t12 = t10;
 t1 = (struct mem_Allocator__589 const *)&t12;
 t13 = (struct mem_Allocator_VTable__592 const *const *)&t1->vtable;
 t14 = (*t13);
 t15 = (void (*const *)(void *, nav__7917_40, uint8_t, uintptr_t))&t14->free;
 t16 = (*t15);
 t17 = t10.ptr;
 t16(t17, t2, t11, t3);
 return;
}

static nav__7918_40 fmt_allocPrint__anon_6753__7918(struct mem_Allocator__589 const a0, nav__7918_43 const a1) {
 struct mem_Allocator__589 const *t1;
 uintptr_t t2;
 uintptr_t t6;
 uint64_t t3;
 nav__7918_54 t4;
 struct mem_Allocator__589 t7;
 struct mem_Allocator__589 t0;
 nav__7918_40 t8;
 nav__7918_40 t11;
 nav__7918_39 t10;
 uint16_t t9;
 bool t5;
 t0 = a0;
 t1 = (struct mem_Allocator__589 const *)&t0;
 t3 = fmt_count__anon_6767__7920(a1);
 t4 = math_cast__anon_6769__7921(t3);
 t5 = t4.is_null != true;
 if (t5) {
  t6 = t4.payload;
  t2 = t6;
  goto zig_block_0;
 }
 return (nav__7918_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_OutOfMemory};

 zig_block_0:;
 t7 = (*t1);
 t8 = mem_Allocator_alloc__anon_2542__4255(t7, t2);
 if (t8.error) {
  t9 = t8.error;
  t8.payload = (nav__7918_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t9;
  return t8;
 }
 t10 = t8.payload;
 t8 = fmt_bufPrint__anon_6772__7922(t10, a1);
 t5 = t8.error == UINT16_C(0);
 if (t5) {
  t10 = t8.payload;
  t8.payload = t10;
  t8.error = UINT16_C(0);
  t11 = t8;
  goto zig_block_1;
 }
 t9 = t8.error;
 switch (t9) {
  case zig_error_NoSpaceLeft: {
   zig_unreachable();
  }
  default: zig_unreachable();
 }

 zig_block_1:;
 return t11;
}

static nav__7919_39 mem_sliceAsBytes__anon_6760__7919(nav__7919_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__7919_39 t7;
 bool t2;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = a0.ptr;
 t4 = (uint8_t *)t3;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uint16_t fmt_format__anon_6823__7961(struct io_Writer__4171 const a0, nav__7961_40 const a1) {
 struct io_Writer__4171 const *t1;
 struct io_Writer__4171 t2;
 struct io_Writer__4171 t0;
 uint8_t const *t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__4171 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7961_44){(uint8_t const *)&__anon_6848,(uintptr_t)10ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_6857__7962(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7961_44){(uint8_t const *)&__anon_6863,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f1;
 t3 = fmt_formatType__anon_6857__7962(t4, (struct fmt_FormatOptions__4577){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__7467(t2, (nav__7961_44){(uint8_t const *)&__anon_6876,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint64_t fmt_count__anon_6767__7920(nav__7920_39 const a0) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 t1;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 t0;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 t2;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 t3;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 const *t4;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 const *t5;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 const *const *t6;
 void const **t8;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *const *t9;
 void const *t10;
 nav__7920_53 (**t11)(void const *, nav__7920_54);
 struct io_Writer__4171 t12;
 struct io_Writer__4171 t7;
 uint64_t t15;
 uint16_t t13;
 bool t14;
 t1 = io_counting_writer_countingWriter__anon_6793__7947();
 t0 = t1;
 t2 = io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_writer__7946(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7920_53 (**)(void const *, nav__7920_54))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWr__7960;
 t12 = t7;
 t13 = fmt_format__anon_6823__7961(t12, a0);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 zig_unreachable();

 zig_block_0:;
 t1 = t0;
 t15 = t1.bytes_written;
 return t15;
}

static nav__7921_38 math_cast__anon_6769__7921(uint64_t const a0) {
 uintptr_t t0;
 nav__7921_38 t1;
 t0 = a0;
 t1.is_null = false;
 t1.payload = t0;
 return t1;
}

static nav__7922_40 fmt_bufPrint__anon_6772__7922(nav__7922_39 const a0, nav__7922_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *t9;
 void const *t10;
 nav__7922_57 (**t11)(void const *, nav__7922_58);
 struct io_Writer__4171 t12;
 struct io_Writer__4171 t7;
 nav__7922_39 t15;
 nav__7922_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2242__4174(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4164(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__4161 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2239 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7922_57 (**)(void const *, nav__7922_58))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__7459;
 t12 = t7;
 t13 = fmt_format__anon_6823__7961(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7922_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4172(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 io_counting_writer_countingWriter__anon_6793__7947(void) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 t0;
 t0 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790){UINT64_C(0)};
 return t0;
}

static struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_writer__7946(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *const a0) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 **t1;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__6804 t0;
 t1 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static nav__7960_38 io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWr__7960(void const *const a0, nav__7960_41 const a1) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *const *t0;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *t1;
 nav__7960_38 t2;
 nav__7960_38 t3;
 t0 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *const *)a0;
 t1 = (*t0);
 t2 = io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_write__7945(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__7960_38));
 return t3;
}

static uint16_t fmt_formatType__anon_6857__7962(uint8_t const *const a0, struct fmt_FormatOptions__4577 const a1, struct io_Writer__4171 const a2, uintptr_t const a3) {
 nav__7962_47 t2;
 nav__7962_47 t3;
 uint8_t const *t0;
 struct io_Writer__4171 t1;
 uint16_t t4;
 uint16_t t5;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = mem_span__anon_6487__7898(a0);
 memcpy(&t3, &t2, sizeof(nav__7962_47));
 t4 = fmt_formatBuf__anon_5399__7863(t3, a1, a2);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static nav__7945_38 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_write__7945(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *const a0, nav__7945_42 const a1) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *const *t1;
 nav__7945_38 t2;
 uintptr_t t3;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *t4;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *t0;
 uint64_t *t5;
 uint64_t t6;
 uint64_t t7;
 t0 = a0;
 t1 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__6790 *const *)&t0;
 t2 = io_dummyWrite__4078(a1);
 t3 = t2.payload;
 t4 = (*t1);
 t5 = (uint64_t *)&t4->bytes_written;
 t6 = (*t5);
 t7 = t3;
 t7 = t6 + t7;
 (*t5) = t7;
 t2.payload = t3;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__4078_38 io_dummyWrite__4078(nav__4078_40 const a0) {
 uintptr_t t0;
 nav__4078_38 t1;
 t0 = a0.len;
 t1.payload = t0;
 t1.error = UINT16_C(0);
 return t1;
}

static uint16_t runner_fifo_RunnerFifo_add_marker__1091(struct runner_fifo_RunnerFifo__621 *const a0, uint32_t const a1, struct shared_MarkerType__2045 const a2) {
 struct runner_fifo_RunnerFifo__621 *const *t1;
 struct runner_fifo_RunnerFifo__621 *t2;
 struct runner_fifo_RunnerFifo__621 *t0;
 struct fifo_UnixPipe_Writer__628 *t3;
 struct shared_Command__struct_2043__2043 t4;
 struct shared_Command__2038 t5;
 struct fifo_UnixPipe_Reader__630 *t7;
 uint16_t t6;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__621 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_UnixPipe_Writer__628 *)&t2->writer;
 t4.marker = a2;
 t4.pid = a1;
 t5.tag = UINT8_C(7);
 t5.payload.AddMarker = t4;
 t6 = fifo_UnixPipe_Writer_sendCmd__1121(t3, t5);
 if (t6) {
  return t6;
 }
 t2 = (*t1);
 t7 = (struct fifo_UnixPipe_Reader__630 *)&t2->reader;
 t6 = fifo_UnixPipe_Reader_waitForAck__1128(t7, (nav__1091_77){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t6) {
  return t6;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_add_marker__767(struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const a0, uint32_t const a1, struct shared_MarkerType__2045 const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_add_marker__1091(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_add_marker__780(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const a0, uint32_t const a1, struct shared_MarkerType__2045 const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t0;
 struct runner_fifo_RunnerFifo__621 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__621 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_add_marker__1091(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_add_marker__249(struct instruments_root_InstrumentHooks__547 *const a0, uint32_t const a1, uint8_t const a2, uint64_t const a3) {
 struct instruments_root_InstrumentHooks__547 *t1;
 struct instruments_root_InstrumentHooks__547 *t2;
 struct instruments_root_InstrumentHooks__547 *t7;
 struct instruments_root_InstrumentHooks__547 *const *t3;
 struct shared_MarkerType__2045 t4;
 struct shared_MarkerType__2045 t5;
 struct instruments_root_InstrumentHooks__547 t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *t10;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *t13;
 uint16_t t6;
 uint16_t t11;
 uint16_t t12;
 bool t0;
 uint8_t t9;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t2;
  switch (a2) {
   case UINT8_C(0): {
    t5.tag = UINT8_C(0);
    t5.payload.SampleStart = a3;
    t4 = t5;
    goto zig_block_1;
   }
   case UINT8_C(1): {
    t5.tag = UINT8_C(1);
    t5.payload.SampleEnd = a3;
    t4 = t5;
    goto zig_block_1;
   }
   case UINT8_C(2): {
    t5.tag = UINT8_C(2);
    t5.payload.BenchmarkStart = a3;
    t4 = t5;
    goto zig_block_1;
   }
   case UINT8_C(3): {
    t5.tag = UINT8_C(3);
    t5.payload.BenchmarkEnd = a3;
    t4 = t5;
    goto zig_block_1;
   }
   default: {
    return UINT8_C(2);
   }
  }

  zig_block_1:;
  t1 = (*t3);
  t7 = t1;
  t3 = (struct instruments_root_InstrumentHooks__547 *const *)&t7;
  t8 = (*t1);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(1);
  if (t0) {
   t1 = (*t3);
   t10 = (struct instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29__572 *)&t1->payload.perf;
   t11 = instruments_fifo_instrument_FifoInstrument_28_Perf_2cerror_7bModeError_7d_29_add_marker__767(t10, a1, t4);
   memcpy(&t12, &t11, sizeof(uint16_t));
   t6 = t12;
   goto zig_block_3;
  }
  t8 = (*t1);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(2);
  if (t0) {
   t1 = (*t3);
   t13 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__581 *)&t1->payload.analysis;
   t12 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_add_marker__780(t13, a1, t4);
   memcpy(&t11, &t12, sizeof(uint16_t));
   t6 = t11;
   goto zig_block_3;
  }
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_4;

  zig_block_4:;
  t6 = 0;
  goto zig_block_3;

  zig_block_3:;
  memcpy(&t12, &t6, sizeof(uint16_t));
  t0 = t12 == UINT16_C(0);
  if (t0) {
   goto zig_block_2;
  }
  return UINT8_C(1);

  zig_block_2:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

uint64_t c_instrument_hooks_current_timestamp__250(void) {
 uint64_t t0;
 t0 = utils_clock_gettime_monotonic__3971();
 return t0;
}

static uint64_t utils_clock_gettime_monotonic__3971(void) {
 struct os_linux_timespec__struct_3100__3100 t0;
 struct os_linux_timespec__struct_3100__3100 t3;
 nav__3971_41 t1;
 intptr_t t4;
 uint64_t t5;
 uint64_t t6;
 bool t2;
 t1 = posix_clock_gettime__1773(UINT32_C(1));
 t2 = t1.error == UINT16_C(0);
 if (t2) {
  t3 = t1.payload;
  t0 = t3;
  goto zig_block_0;
 }
 zig_unreachable();

 zig_block_0:;
 t4 = t0.sec;
 t5 = (uint64_t)t4;
 t5 = t5 * UINT64_C(1000000000);
 t4 = t0.nsec;
 t6 = (uint64_t)t4;
 t6 = t5 + t6;
 return t6;
}

static uint64_t const builtin_zig_backend__255 = UINT64_C(3);

static bool const start_simplified_logic__109 = false;

static uint8_t const builtin_output_mode__256 = UINT8_C(1);

static uint8_t const builtin_link_mode__257 = UINT8_C(0);

static uint64_t features_features__322 = ((uint64_t)UINT64_C(0));

static uintptr_t const bit_set_IntegerBitSet_2864_29_bit_length__353 = 64ul;

static bool const builtin_link_libc__267 = true;

static bool const posix_use_libc__1425 = true;

static struct Target_Os__665 const builtin_os__263 = {{ .linux = {{{6ul,18ul,5ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},{6ul,18ul,5ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}}},{2ul,39ul,0ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},UINT32_C(14)} },UINT8_C(9)};

static uint8_t const c_native_os__1867 = UINT8_C(9);

static struct Target_DynamicLinker__844 const Target_DynamicLinker_none__3589 = {"\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252",UINT8_C(0)};

static bool const builtin_is_test__259 = false;

static uint8_t *heap_CAllocator_alloc__3713(void *const a0, uintptr_t const a1, uint8_t const a2, uintptr_t const a3) {
 uint64_t t0;
 uint8_t *t2;
 bool t1;
 (void)a0;
 (void)a3;
 t0 = a1;
 t1 = t0 > UINT64_C(0);
 debug_assert__180(t1);
 t2 = heap_CAllocator_alignedAlloc__3710(a1, a2);
 return t2;
}

static bool heap_CAllocator_resize__3714(void *const a0, nav__3714_40 const a1, uint8_t const a2, uintptr_t const a3, uintptr_t const a4) {
 uintptr_t t0;
 uint64_t t1;
 uint64_t t2;
 uint8_t *t4;
 bool t3;
 (void)a0;
 (void)a2;
 (void)a4;
 t0 = a1.len;
 t1 = a3;
 t2 = t0;
 t3 = t1 <= t2;
 if (t3) {
  return true;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a1.ptr;
 t0 = heap_CAllocator_alignedAllocSize__3712(t4);
 t2 = a3;
 t1 = t0;
 t3 = t2 <= t1;
 if (t3) {
  return true;
 }
 goto zig_block_1;

 zig_block_1:;
 return false;
}

static uint8_t *heap_CAllocator_remap__3715(void *const a0, nav__3715_40 const a1, uint8_t const a2, uintptr_t const a3, uintptr_t const a4) {
 uint8_t *t0;
 uint8_t *t2;
 uint8_t *t3;
 bool t1;
 t1 = heap_CAllocator_resize__3714(a0, a1, a2, a3, a4);
 if (t1) {
  t2 = a1.ptr;
  t3 = (uint8_t *)t2;
  t0 = t3;
  goto zig_block_0;
 }
 t0 = NULL;
 goto zig_block_0;

 zig_block_0:;
 return t0;
}

static void heap_CAllocator_free__3716(void *const a0, nav__3716_40 const a1, uint8_t const a2, uintptr_t const a3) {
 uint8_t *t0;
 (void)a0;
 (void)a2;
 (void)a3;
 t0 = a1.ptr;
 heap_CAllocator_alignedFree__3711(t0);
 return;
}

static uint8_t *heap_CAllocator_alignedAlloc__3710(uintptr_t const a0, uint8_t const a1) {
 uint8_t const *t1;
 uintptr_t t3;
 uint64_t t4;
 void *t9;
 void *t5;
 uint8_t *t10;
 int t6;
 int32_t t7;
 uint8_t t2;
 uint8_t t0;
 bool t8;
 t0 = a1;
 t1 = (uint8_t const *)&t0;
 t2 = (*t1);
 t3 = mem_Alignment_toByteUnits__1065(t2);
 t3 = ((uintptr_t)8ul > t3) ? (uintptr_t)8ul : t3;
 t4 = t3;
 t3 = t4;
 t6 = posix_memalign(&t5, t3, a0);
 t7 = t6;
 t8 = t7 != INT32_C(0);
 if (t8) {
  return NULL;
 }
 goto zig_block_0;

 zig_block_0:;
 t9 = t5;
 t10 = (uint8_t *)t9;
 return t10;
}

static uintptr_t heap_CAllocator_alignedAllocSize__3712(uint8_t *const a0) {
 void const *t0;
 void const *t1;
 uintptr_t t2;
 t0 = (void const *)a0;
 t1 = t0;
 t2 = malloc_usable_size(t1);
 return t2;
}

static void heap_CAllocator_alignedFree__3711(uint8_t *const a0) {
 void *t0;
 void *t1;
 t0 = (void *)a0;
 t1 = t0;
 free(t1);
 return;
}

static uintptr_t mem_Alignment_toByteUnits__1065(uint8_t const a0) {
 uintptr_t t1;
 uint8_t t0;
 t0 = a0;
 t1 = zig_shlw_u64((uintptr_t)1ul, t0, UINT8_C(64));
 return t1;
}

static struct mem_Allocator__589 const heap_c_allocator__3678 = {((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)&heap_CAllocator_vtable__3705)};

static struct mem_Allocator__589 const c_allocator__235 = {((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__592 const *)&heap_CAllocator_vtable__3705)};

static struct Target_Cpu_Feature_Set__861 const Target_Cpu_Feature_Set_empty__3634 = {{0ul,0ul,0ul,0ul,0ul}};

static struct Target_Cpu__829 const builtin_cpu__262 = {((struct Target_Cpu_Model__856 const *)&Target_x86_cpu_znver4__3833),{{449166724097319152ul,238734216392818280ul,15402363330914092936ul,3ul,0ul}},UINT8_C(42)};

static uint8_t const builtin_abi__261 = UINT8_C(1);

static uint8_t const builtin_object_format__265 = UINT8_C(0);

static struct Target__663 const builtin_target__264 = {{((struct Target_Cpu_Model__856 const *)&Target_x86_cpu_znver4__3833),{{449166724097319152ul,238734216392818280ul,15402363330914092936ul,3ul,0ul}},UINT8_C(42)},{{ .linux = {{{6ul,18ul,5ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},{6ul,18ul,5ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}}},{2ul,39ul,0ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},UINT32_C(14)} },UINT8_C(9)},UINT8_C(1),UINT8_C(0),{"/nix/store/zdpby3l6azi78sl83cpad2qjpfj25aqx-glibc-2.40-66/lib/ld-linux-x86-64.so.2\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252",UINT8_C(82)}};

static struct builtin_CallingConvention__266 const builtin_CallingConvention_c__463 = {{ .x86_64_sysv = {{UINT64_C(0xaaaaaaaaaaaaaaaa),true}} },UINT8_C(4)};

static uint8_t const (*const shared_RUNNER_CTL_FIFO__754)[21] = &__anon_1941;

static uint8_t const (*const shared_RUNNER_ACK_FIFO__755)[21] = &__anon_2012;

static uint64_t const runner_fifo_PROTOCOL_VERSION__1078 = UINT64_C(2);

static uint8_t const mem_native_endian__789 = UINT8_C(1);

static int const cimport_F_GETFL__5193 = 3;

static int const cimport_O_NONBLOCK__5169 = 2048;

static int const cimport_F_SETFL__5194 = 4;

static uint8_t const os_linux_native_arch__2773 = UINT8_C(42);

static int const cimport_EINTR__6896 = 4;

static uint8_t const fs_path_native_os__4295 = UINT8_C(9);

static uint8_t const fs_native_os__1138 = UINT8_C(9);

static uint8_t const fs_Dir_native_os__3960 = UINT8_C(9);

static bool const fs_File_is_windows__1314 = false;

static uint8_t const posix_native_os__1423 = UINT8_C(9);

static uint8_t logger_max_level__4271 = UINT8_C(0);

static uint8_t const builtin_mode__266 = UINT8_C(3);

static bool const debug_runtime_safety__159 = false;

static bool const fs_Dir_have_flock__3961 = true;

static uint8_t const c_native_abi__1865 = UINT8_C(1);

static bool const posix_unexpected_error_tracing__1854 = false;

static uint16_t const fmt_max_format_args__7205 = UINT16_C(32);

static bool const debug_default_enable_segfault_handler__205 = false;

static uint8_t const log_default_level__7777 = UINT8_C(0);

static struct std_Options__4590 const std_options__97 = {{0xaaaaaaaaaaaaaaaaul,true},{0xaaaaaaaaaaaaaaaaul,true},3ul,false,UINT8_C(0),false,true,false,false,false,UINT8_C(2)};

static bool const posix_lfs64_abi__1853 = true;

static uint8_t const (*const fmt_ANY__7208)[4] = &__anon_5243;

static uint8_t const unicode_native_endian__7485 = UINT8_C(1);

static uint32_t const unicode_replacement_character__7486 = UINT32_C(65533);

static bool const mem_backend_supports_vectors__817 = true;

static bool const builtin_valgrind_support__270 = false;

static nav__3694_38 const heap_page_size_min_default__3694 = {4096ul,false};

static uintptr_t const heap_page_size_min__3673 = 4096ul;

static uint8_t const c_MARKER_TYPE_SAMPLE_START__245 = UINT8_C(0);

static uint8_t const c_MARKER_TYPE_SAMPLE_END__246 = UINT8_C(1);

static uint8_t const c_MARKER_TYPE_BENCHMARK_START__247 = UINT8_C(2);

static uint8_t const c_MARKER_TYPE_BENCHMARK_END__248 = UINT8_C(3);

static struct mem_Allocator_VTable__592 const heap_CAllocator_vtable__3705 = {&heap_CAllocator_alloc__3713,&heap_CAllocator_resize__3714,&heap_CAllocator_remap__3715,&heap_CAllocator_free__3716};

static bool const heap_CAllocator_supports_malloc_size__3706 = true;

static bool const heap_CAllocator_supports_posix_memalign__3708 = true;

static struct Target_Cpu_Model__856 const Target_x86_cpu_znver4__3833 = {{(uint8_t const *)&__anon_6988,6ul},{(uint8_t const *)&__anon_6988,6ul},{{149677348868194480ul,220719817882287720ul,10790633606899500936ul,3ul,0ul}}};

#endif
#ifndef WRAPPER_H
#define WRAPPER_H

#include <stdint.h>

#ifndef _WIN32
#include "callgrind.h"
#include "valgrind.h"

uint8_t running_on_valgrind() { return RUNNING_ON_VALGRIND > 0; }

void callgrind_dump_stats() { CALLGRIND_DUMP_STATS; }

void callgrind_dump_stats_at(uint8_t const* pos_str) {
  CALLGRIND_DUMP_STATS_AT(pos_str);
}

void callgrind_zero_stats() { CALLGRIND_ZERO_STATS; }

void callgrind_start_instrumentation() { CALLGRIND_START_INSTRUMENTATION; }

void callgrind_stop_instrumentation() { CALLGRIND_STOP_INSTRUMENTATION; }

#else
// Windows and other platforms - provide no-op implementations
uint8_t running_on_valgrind() { return 0; }

void callgrind_dump_stats() {}

void callgrind_dump_stats_at(uint8_t const* pos_str) {}

void callgrind_zero_stats() {}

void callgrind_start_instrumentation() {}

void callgrind_stop_instrumentation() {}
#endif

#endif

